//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// CmisAgent.h - CMIS Agent object declaration

// By George Pavlou, February 1991


#ifndef _CMIS_AGENT_
#define _CMIS_AGENT_


#ifndef _BOOLEAN_
#include "Boolean.h"
#endif

#ifndef _LIST_ITERATOR_
#include "ListIterator.h"
#endif

#ifndef _GENERIC_KS_
#include "GenericKS.h"
#endif

#ifndef _module_SMI_defined_
#include "SMI-types.h"
#endif

#ifndef _AVA_
#include "AVA.h"
#endif

#if	defined(SECURITY) && !defined(_ASSOCIATION_MANAGER_)
#include "AssocMgr.h"
#endif

#if	defined(ACCESS) && !defined(_ADF_)
#include "ADF.h"
#endif


class MO;

struct AssocInfo
{
    int		msd;
    DN		aet;

    static const int MsdCmp;
    static const int AetCmp;
    static int cmpmode;

    static int msd_cmp (AssocInfo*, AssocInfo*);
    static int aet_cmp (AssocInfo*, AssocInfo*);

};


class AssocInfoList : public List
{
protected:
    int		compare (void*, void*);
public:
    void	deleteElem (void*);
    AssocInfo*	find (AssocInfo*);
    AssocInfo*	remove (AssocInfo*);
};


class CMISAgent : public KS
{
    friend class AssociationMgr;

#define AGENT_FUNUNITS	0x17		/* CMIS agent capabilities:
					   scope, filter, cancelGet */
#define DFLT_MAXASSOCS	64		/* default maximum CMIS associations */
#define CANCELGETBLOCK	5		/* to check for M-Cancel-Get */
#define	CONTEXT		"management"	/* the application context */

#define	GARBAGE_COLLECTION_INTVL 600	/* every 10 minutes */

#ifdef	ACCESS_TEST			/* for testing access control */
#define	TEST_INITIATOR	"c=GB@o=UCL@ou=CS@cn=MGR"
#endif

protected:
    static CMISAgent* _instance;
    DN		 _emptyDN;
    OID		 _processingFailureOID;
    char*	 _operationClass;
    int		 _psapEndpoint;		// listen for new associations
    fd_set	 _psapMask;
    int		 _maxAssocs;		// max cmis associations
    int		 _curAssocs;		// current cmis associations
    int		 _thisAssoc;		// association being served
    char*	 _aei;			// application entity title
    List	 _requests;
    AssocInfoList _assocInfoList;
    ListIterator _iterator;
    int		 _moClassIndex;
    Bool	 _createOperation;
    int		 _invokeId;
    timeval	 _prevTime;
#ifdef ACCESS
    ADF	 	 _adf;
#endif
#ifdef	ACCESS_TEST
    DN		 _initr;
#endif
#ifdef USE_X500
    char*	_agentService;
    Bool	_singleInstanceAgent;	// for true location transparency
    Bool	_ltFailed;		// in case the DSA is down 
    /* a class e.g. CMISappProcess is needed, this is a temporary hack */
    Bool	_SMAPrunning;		// if multiple SMAEs add the SMAP once
#endif

    void	doTime (char* = NULLCP);
    void	init ();
    CMISErrors	checkParms (MID, MN, MN,
			CMISScope*, CMISFilter*, CMISSync);
    int		serviceCmisRequest (int, MSAPIndication*, PE);
    int		checkForCancelGet (int, int);
#ifdef ACCESS
    int		checkMOAccess(DN, MO*, CMISScope*, CMISFilter*, int);
#endif

    int		Get (int, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, int, MID);
    int		CancelGet (int, int, int, int);

    int		Set (int, Bool, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, int, CMISSetAttr*);
    int		Action (int, Bool, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync, CMISParam*);
    int		Create (int, int, MID, MN, int, MN,
			External*, int, CMISParam*);
    int		Delete (int, int, MID, MN, CMISScope*, CMISFilter*,
			External*, CMISSync);
    int		MError (int, int, int, int, CMISErrors,
			MID = NULLMID, MN = NULLMN,
			CMISScope* = NULLMSCOPE, CMISFilter* = NULLMFILTER,
			CMISSync = NULLMSYNC, MIDListVal* = NULL,
			PE = NULLPE, AVA*& = NULLAVAREF);

public:
    int		initialise (int, char**, char*, TSAPdisconnect*, Bool = False);
    int		associationRequest (int, char**);
    int		readCommEndpoint (int);	// read association's data
    int		readCommEndpointAux (int); // ..
    int		shutdown (int);		// shutdown association
    int		wakeUp (char*);		// for "garbage" collection
    int		sendEventReport (MO*, MID, MN, MID, PE, char*, AET*, int = -1);
    void	setMaxAssocs (int);
    int		getThisAssoc ();
    char*	getAEI ();
    DN		mn2localdn (MN);
    DN		getEmptyDN ();
#ifdef	ACCESS /* MLS */
    DN		getInitiator (int = -1);
    Bool	verifyInitiator (DN);
#endif

    static int associationInit (int, char**);
    static int associationWork (int);
    static CMISAgent* getInstance ();

    CMISAgent (int, char**, char*, Bool = False);
    CMISAgent ();
    ~CMISAgent ();
};


inline int  AssocInfo::msd_cmp (AssocInfo* v1, AssocInfo* v2)
{ return ((AssocInfo*) v1)->msd == ((AssocInfo*) v2)->msd ? OK : NOTOK; }

inline int  AssocInfo::aet_cmp (AssocInfo* v1, AssocInfo* v2)
{ return v1->aet && v2->aet &&
  dn_cmp(((AssocInfo*) v1)->aet, ((AssocInfo*) v2)->aet)
  == 0 ? OK : NOTOK; }

inline int  AssocInfoList::compare (void* v1, void* v2)
{ return AssocInfo::cmpmode == AssocInfo::MsdCmp ?
  AssocInfo::msd_cmp((AssocInfo*) v1, (AssocInfo*) v2) :
  (AssocInfo::cmpmode == AssocInfo::AetCmp ?
  AssocInfo::aet_cmp((AssocInfo*) v1, (AssocInfo*) v2) : NOTOK); }

inline AssocInfo*  AssocInfoList::find (AssocInfo* v)
{ AssocInfo::cmpmode = AssocInfo::AetCmp; return (AssocInfo*) List::find(v); }

inline AssocInfo*  AssocInfoList::remove (AssocInfo* v)
{ AssocInfo::cmpmode = AssocInfo::MsdCmp; return (AssocInfo*) List::remove(v); }

inline void  AssocInfoList::deleteElem (void* v)
{ if (((AssocInfo*)v)->aet) dn_free(((AssocInfo*) v)->aet); delete v; }

inline void	CMISAgent::setMaxAssocs (int assocs)	{ _maxAssocs = assocs; }
inline int	CMISAgent::getThisAssoc ()	{ return _thisAssoc; }
inline char*	CMISAgent::getAEI ()		{ return _aei; }
inline CMISAgent* CMISAgent::getInstance ()	{ return _instance; }
inline DN	CMISAgent::getEmptyDN ()	{ return _emptyDN ? _emptyDN :
					    (_emptyDN = dn_comp_new(NULLRDN)); }
#ifdef	ACCESS /* MLS */
inline DN	CMISAgent::getInitiator (int msd) { return _initr; }
inline Bool	CMISAgent::verifyInitiator (DN intr) { return True; }
#endif
inline		CMISAgent::CMISAgent ()		{ init(); }


#endif
