//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// MOClassInfo.h - generic meta managed object class

// By George Pavlou, February 1991


#ifndef _MO_CLASS_INFO_
#define _MO_CLASS_INFO_


#ifndef _BOOLEAN_
#include "Boolean.h"
#endif

#ifndef _GENERIC_ATTR_
#include "GenericAttr.h"
#endif


// forward references

class MO;			// in GenericMO.h
class MOClassArray;		// in Create.h


// MO creation (factory) methods

typedef MO* (*MCREATE_FNX) (RDN, MO*, MO*, int, CMISParam*,
                                CMISErrors*, CMISErrorInfo*, OID*);
typedef MO* (*CREATE_FNX) (RDN, MO*);

#define NULLMCREATE     ((MCREATE_FNX) 0)
#define NULLCREATE      ((CREATE_FNX) 0)



typedef enum {
    cm_noCreate,
    cm_createWithName,
    cm_createWithAutoInstanceName
} CreateMode;


typedef enum {
    dm_noDelete,
    dm_deleteOnlyIfNoContainedObjs,
    dm_deleteContainedObjs
} DeleteMode;


typedef enum {
    sm_getOnly,
    sm_set,
    sm_setByCreate
} SetMode;


class NameBinding
{
    OID		_nameBinding;
    OID		_superiorClass;
    OID		_namingAttr;

    CreateMode	_createMode;
    Bool	_referenceObjMode;
    DeleteMode	_deleteMode;

public:
    OID	getNameBinding ();
    OID	getSuperiorClass ();
    OID	getNamingAttr ();

    void setCreateMode (CreateMode, Bool = False);
    void setDeleteMode (DeleteMode);

    CreateMode getCreateMode ();
    Bool       getReferenceObjMode ();
    DeleteMode getDeleteMode ();

    NameBinding (OID, OID, OID);
    ~NameBinding ();
};

inline OID  NameBinding::getNameBinding ()
{ return _nameBinding; }

inline OID  NameBinding::getSuperiorClass ()
{ return _superiorClass; }

inline OID  NameBinding::getNamingAttr ()
{ return _namingAttr; }

inline void  NameBinding::setCreateMode (CreateMode createMode, Bool refObjMode)
{ _createMode = createMode; _referenceObjMode = refObjMode; }

inline void  NameBinding::setDeleteMode (DeleteMode deleteMode)
{ _deleteMode = deleteMode; }

inline CreateMode  NameBinding::getCreateMode ()
{ return _createMode; }

inline Bool  NameBinding::getReferenceObjMode ()
{ return _referenceObjMode; }

inline DeleteMode  NameBinding::getDeleteMode ()
{ return _deleteMode; }




class MOClassInfo			/* class specific information */
{
friend class MO;

    OID		_moClass;			// mo class identifier
    char*	_moClassName;			// mo class name
    int		_nbindings;
    NameBinding** _nameBindings;		// name bindings
    int		_nattrs;
    OID*	_attrOids;			// attribute object identifiers
    char**	_attrNames;			// attribute names
    int*	_attrIds;			// attribute identifiers
    SetMode*	_settable;
    Attr**	_defaults;
    int		_ngroups;
    int*	_ngroupattrs;
    OID**	_attrGroups;			// attr groups
    char**	_groupNames;			// group names
    int*	_groupIds;			// group identifiers
    int		_nevents;
    OID*	_eventOids;			// event object identifiers
    char**	_eventNames;			// event names
    int*	_eventIds;			// event identifiers
    int		_nactions;
    OID*	_actionOids;			// action object identifiers
    char**	_actionNames;			// action names
    int*	_actionIds;			// action ids
    Attr**	_actionInfoTmpls;		// action templates
    Attr**	_actionReplyTmpls;		//  ..
    int		_npackages;
    OID*	_packageOids;			// conditional packages
    char**	_packageNames;
    int*	_packageIds;

    MOClassInfo* _parent;			// parent in inheritance tree
    MO*		_first;				// points to the first instance

						// factory methods for creation
    MCREATE_FNX	_mcreate;			// CMIS create
    CREATE_FNX	_create;			// local create

    static MOClassArray* _classArray;		// meta-class information

public:

    Bool	_gdmoCmplGenerated;		// for compatibility with
						// fully hand-written MOs

    int		setClass (char*, int, int, int, int, int, int = 0);
					// initialises the class
    int		setNameBinding (char*, char*, char*,
				CreateMode = cm_noCreate, Bool = False,
				DeleteMode = dm_noDelete);
    NameBinding** getNameBindings (int*);
    NameBinding*  checkNameBinding (MO*, RDN);
					// checks all inheritance branch
    NameBinding*  checkCmisCreate ();	//  ..

    int		setAttr (char*, int, SetMode = sm_getOnly, Attr* = NULLATTR);
					// initialises an attribute
    int		setGroupAttr (char*, int, int);
					// initialises a group attribute
    int		addGroupAttr (int, int);// adds a group attribute
    int		setEvent (char*, int);
					// initialises an event
    int		setAction (char*, int, Attr*, Attr*);
    int		setAction (int, Attr*, Attr*);
					// initialise an action
    int		setPackage (char*, int);// initialises a package
    int 	getAttrId (OID, SetMode* = NULL, Attr** = NULL);
					// gets the attribute identifier
    int 	getAttrId (char*);	// 	   ..
    OID 	getAttrOid (int);	// 	   ..
    OID* 	getGroupAttrs (OID, int*);
					// gets the group attrs
    int		getSetInfo (int, SetMode*, Attr**);
					// gets the settable and default info
    Attr*	getDefault (int);
    int		setDefault (int, Attr*);
    SetMode	getSettable (int);
    int 	getEventId (OID);
					// gets the event identifier
    OID 	getEventOid (int);
					// gets the event object identifier
    OID 	getEventOid (char*);	//	   ..
    int 	getActionId (OID, Attr** = NULL, Attr** = NULL);
					// gets the action identifier
    Attr* 	getActionInfoTemplate (int);
    Attr* 	getActionReplyTemplate (int);
    int 	getPackageId (OID);	// gets the cnd package identifier
    OID 	getPackageOid (int);	// gets the cnd package oid

    MO*		getMO (RDN);		// finds object with rdn among
					// objects of the same class
    OID		getClass ();
    char*	getClassName ();
    MO*		getFirst ();
    void	setFirst (MO*);
    int		getNAttrs();
    char	*getAttrName(int);
    Bool	initialised ();
    MOClassInfo* getParent ();
    MOClassInfo* getParent (char*);	// checks all the inheritance branch
    void	setParent (MOClassInfo*);
    MCREATE_FNX	getMCreate ();
    CREATE_FNX	getCreate ();
    void	setMCreate (MCREATE_FNX);
    void	setCreate (CREATE_FNX);
    void	clear ();		// releases allocated memory
    MOClassInfo ();
    ~MOClassInfo ();
    static MOClassArray* getClassArray ();
};


inline OID  MOClassInfo::getClass ()
{ return _moClass; }

inline char*  MOClassInfo::getClassName ()
{ return _moClassName; }

inline MO*  MOClassInfo::getFirst ()
{ return _first; }

inline void  MOClassInfo::setFirst (MO* first)
{ _first = first; }

inline int  MOClassInfo::getNAttrs()
{ return(_nattrs); }

inline char* MOClassInfo::getAttrName(int i)
{ return((i>=0) && (i<_nattrs) ? _attrNames[i] : (char*) NULL); }

inline Attr* MOClassInfo::getDefault (int attrId)
{ SetMode s; Attr* d; getSetInfo(attrId, &s, &d); return d; }

inline SetMode  MOClassInfo::getSettable (int attrId)
{ SetMode s; Attr* d; getSetInfo(attrId, &s, &d); return s; }

inline Attr*  MOClassInfo::getActionInfoTemplate (int actionId)
{ return actionId < _nactions ? _actionInfoTmpls[actionId] : NULLATTR; }

inline Attr*  MOClassInfo::getActionReplyTemplate (int actionId)
{ return actionId < _nactions ? _actionReplyTmpls[actionId] : NULLATTR; }

inline MOClassInfo*  MOClassInfo::getParent ()
{ return _parent; }

inline void  MOClassInfo::setParent (MOClassInfo* parent)
{ _parent = parent; }

inline MCREATE_FNX  MOClassInfo::getMCreate ()
{ return _mcreate; }

inline CREATE_FNX  MOClassInfo::getCreate ()
{ return _create; }

inline void  MOClassInfo::setMCreate (MCREATE_FNX mcreate)
{ _mcreate = mcreate; }

inline void  MOClassInfo::setCreate (CREATE_FNX create)
{ _create = create; }

inline Bool  MOClassInfo::initialised ()
{ return _moClass ? True:False;}

inline MOClassInfo::~MOClassInfo ()
{ clear(); }

inline MOClassArray*  MOClassInfo::getClassArray ()
{ return MOClassInfo::_classArray; }


#endif
