//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Cache of log records

// Saleem N. Bhatti
// March 1994

#if !defined(_RecordInfo_cc_)
#define _RecordInfo_cc_

#include <stdlib.h>
#if defined(__GNUG__)
#include <sys/stat.h>
#elif defined(__cplusplus)
#include <sysent.h>
#endif

#include "RecordInfo.h"
#include "eventLogRecord.h"
#include "ObjectClass.h"
#include "ManagedObjectInstance.h"
#include "EventType.h"
#include "Error.h"

static int findClass(const char *moclass)
{
    for (register int i = 0; i < NCLASSES; ++i)
        if (strcmp(moclass, moclasses[i].className) == 0)
            return i;   // found it

    return -1;
}

extern char *etcdir; /* In msap/syntax.c */

/*
** Only create a new logRecord type when it is needed.
*/

int RecordInfo::_count = 0;

RecordInfo::RecordInfo(const MOClass *recordInfo,
                       const NotificationMapping *mappingInfo)
{
    _recordInfo = recordInfo;
    _mappingInfo = mappingInfo;
    _record = (logRecord *) 0;

    OID id = name2oid((char *) mappingInfo->notification);

    oid_table_attr *oidInfo;

    if ((oidInfo = oid2attr(id)) == (oid_table_attr *) 0) {
        fprintf(stderr, "RecordInfo: No attribute table for %s!\n",
                mappingInfo->notification);
	oid_free(id);
        exit(0); /* drastic but necessary! */
    }
    oid_free(id);

    if ((_syntax = get_syntax_table(oidInfo->oa_syntax)) == (sntx_table *) 0) {
        fprintf(stderr, "RecordInfo: No syntax table for %s!\n",
                mappingInfo->notification);
        exit(0); /* drastic but necessary! */
    }

    /* No syntax routines registered! */
    if ((IFP) _syntax->s_encode == (IFP) pe_cpy) {
        fprintf(stderr, "RecordInfo: No syntax rotuines for %s\n",
                mappingInfo->notification);
        exit(0); /* drastic but necessary! */
    }
}

RecordInfo::~RecordInfo()
{
    if (_record) {
        _record->deleteRR();
        delete _record;
    }
}

logRecord *
RecordInfo::getRecord()
{
    if (_record == (logRecord *) 0) {
        char name[64];
        RDN rdn;

        sprintf(name, "logRecordId=%d", --_count);
        rdn = str2rdn(name);
        _record = (logRecord *) _recordInfo->getClassInfo()->
						getCreate()(rdn, (MO *) 0);
	_record->createRR();

        /* Will this hold in the face of variant record types? */
        ((eventLogRecord *) _record)->managedObjectClass()->set((void *) 0);
        ((eventLogRecord *) _record)->managedObjectInstance()->set((void *) 0);
        ((eventLogRecord *) _record)->eventType()->set((void *) 0);
    }

    return _record;
}

NotificationMapper *notificationMapper = (NotificationMapper *) 0;

RecordInfo **
NotificationMapper::_recordInfoTable = (RecordInfo **) 0;

RecordInfo *
NotificationMapper::getRecordInfo(const char *notificationName)
{
    if (_recordInfoTable == (RecordInfo **) 0)
        return (RecordInfo *) 0;

    /* search for the name in the table */
    register
    RecordInfo **r;

    for(r = _recordInfoTable; *r != (RecordInfo *) 0; ++r)
        if (strcmp(notificationName, (*r)->_mappingInfo->notification) == 0)
            break;

    return *r;
}

NotificationMapper::NotificationMapper()
{
    /* Only need to do initialisation once */
    if (_recordInfoTable != (RecordInfo **) 0)
        return;

    _recordInfoTable = new RecordInfo *[NNOTIFICATIONS + 1];
    _recordInfoTable[NNOTIFICATIONS] = (RecordInfo *) 0;

    register
    int j, k, flag = 0;

    /* Get the table mapping notifications to record types */
    register
    NotificationMapping *nm;

    fprintf(stderr, "setting up notification to event record mappings...\n");
    for(j = 0, nm = notificationMappingTable; j < NNOTIFICATIONS; ++j, ++nm) {

        /* Does this notification have a known syntax? */
        if (str2AttrT((char *) nm->notification) == (AttributeType) 0) {
            fprintf(stderr, "NotficationMapper: Error - line %d - Unknown syntax name %s\n", j, nm->notification);
            flag = 1;
        }

        /* Does this class exist? */
        else if ((k = findClass(nm->record)) == NOTOK) {
            fprintf(stderr, "NotficationMapper: Error - line %d - unknown record type - %s\n", j, nm->record);
            flag = 1;
        }

        /* Can it be created by the GMS? */
        else if (moclasses[k].getClassInfo()->getCreate() == (CREATE_FNX) 0) {
            fprintf(stderr, "NotficationMapper: Error - line %d - class %s has no create(RDN *, MO *) method registered\n", j, nm->record);
            flag = 1;
        }

        /* Check for duplicate entries */
        else {
            _recordInfoTable[j] = (RecordInfo *) 0; /* for getRecordInfo() */
            if (getRecordInfo(nm->notification) != (RecordInfo *) 0) {
                fprintf(stderr, "NotficationMapper: Error - line %d - duplicate entry for %s", j, nm->notification);
                flag = 1;
            }
        }

        /* Error -- quit */
        if (flag) {
            delete this;
            error((char *) 0, "NotificationMapper: Initialisation failed!");
        }

        _recordInfoTable[j] = new RecordInfo(&(moclasses[k]), nm);

#if defined(DEBUG)
        printf("%-40s %s\n", nm->notification, nm->record);
#endif

    }

}

NotificationMapper::~NotificationMapper()
{
    for(RecordInfo **r = _recordInfoTable; *r != (RecordInfo *) 0; ++r)
        delete *r;
    delete _recordInfoTable;
}

logRecord *
NotificationMapper::getRecord(const char *notificationName)
{
    register
    RecordInfo *r = getRecordInfo(notificationName);

    return (r != (RecordInfo *) 0 ? r->_record : (logRecord *) 0);
}

logRecord *
NotificationMapper::mapNotification(const char *name)
{
    register
    RecordInfo *ri = getRecordInfo(name);

    if (ri == (RecordInfo *) 0)
        return (logRecord *) 0;

    return ri->getRecord();
}

PE
NotificationMapper::encodeNotification(const char *name, const void *notification)
{
    register
    RecordInfo *ri = getRecordInfo(name);

    if (ri == (RecordInfo *) 0) {
        fprintf(stderr, "NotificationMapper: No RecordInfo for %s\n", name);
        return (PE) 0; /* Shouldn't happen */
    }

    /* This had better exist or else ... */
    return (PE) (ri->_syntax->s_encode)((void *) notification);
}

void
NotificationMapper::freeNotification(const char *name, const void *notification)
{
    register
    RecordInfo *ri = getRecordInfo(name);

    if (ri == (RecordInfo *) 0) {
        fprintf(stderr, "NotificationMapper: No RecordInfo for %s\n", name);
        return ; /* Shouldn't happen */
    }

    /* This had better exist or else ... */
    /* (ri->_syntax->s_free)((void *) notification); GP-940616 */
}

#endif /* _RecordInfo_cc_ */
