//
// NAME
//   movingAverageMeanMonitor.inc.cc
//
// DESCRIPTION
//   for movingAverageMeanMonitor MOC, via OSIMIS GDMO compiler
//   (based on ISO/IEC 10164-11 Monitor Objects MIB)
//
// AUTHOR
//   Henryka Siejka-Jormakka
//   Technical Research Centre of Finland
//   icm.group@tel.vtt.fi
//


#include "SmiSntx.h"		/* for QosAlarmInfo + other notifications */
#include "MonitorSntx.h"	/* for TimePeriod */

#include "Bool.h"		/* for monitorMetric::counterOrGaugeDifference*/
#include "AdministrativeState.h"/* for scanner::administrativeState */


#define  APPROXIMATING_UP	1
	/* Is rounding up of meanEstimate is desired (1) or not (0) ? */


int movingAverageMeanMonitor::createRR(AVA*& err, void* pv, int) {
  static Bool boolClassInitialised = False;

  // correct OO style: initialize parent class stuff
  if (monitorMetric::createRR(err, pv) == NOTOK)
    return NOTOK;

  estimateOfMean()->setType(this->gaugeType());
    // _gaugeType was set in monitorMetric::createRR()

  double MEAN_ESTIMATE;
  if (this->gaugeType() == OV_INTEGER) {
    MEAN_ESTIMATE = 
      ((GaugeInt*)monitorMetric::derivedGauge())->getint();
  } else {
    MEAN_ESTIMATE = 
      ((GaugeInt*)monitorMetric::derivedGauge())->getreal();
  }

  estimateOfMean()->setreal(MEAN_ESTIMATE); 
   

  type_Monitor_TimePeriod* tp = (type_Monitor_TimePeriod*)
    scanner::granularityPeriod()->get();

  int iPollInterval = TimePeriod2seconds(tp);

  
  type_Monitor_TimePeriod* mtp = (type_Monitor_TimePeriod*)
    movingTimePeriod()->get();


  int mTimePeriod = TimePeriod2seconds(mtp);

  if (iPollInterval+mTimePeriod != 0) {
    _smoothingConst = 
      (double)(2*iPollInterval)/(double)(iPollInterval+mTimePeriod);
  } else {
    _smoothingConst = 0.0;
  }

  _oldMeanEstimate = MEAN_ESTIMATE;

  update();
  
  //
  // the estimateOfMeanSeverityIndicatingGaugeThreshold attribute
  // set type for GaugeThreshold attributes because GDMO compiler can't know it:
  //   OV_INTEGER or OV_REAL
  // the type is the same as derivedGauge
  //
  estimateOfMeanSeverityIndicatingGaugeThreshold()->setType(this->gaugeType());
  //   
  // set the initial value (GDMO compiler leaves it empty)
  // only if it has not already be set through CMIS
  //
  if (!estimateOfMeanSeverityIndicatingGaugeThreshold()->get()) {
      estimateOfMeanSeverityIndicatingGaugeThreshold()
       ->add(EMSIGAUGETHLDLOW, Off, EMSIGAUGETHLDHIGH, Off); 
  }

  //
  // associate the threshold to the gauge
  //
  estimateOfMean()->associateThreshold (
      estimateOfMeanSeverityIndicatingGaugeThreshold() );

  //
  // set the default value/mode only once for the class.
  //
  if (!boolClassInitialised) {
    if (this->gaugeType() == OV_INTEGER) {
      _classInfo->setDefault (I_estimateOfMeanSeverityIndicatingGaugeThreshold,
        new GaugeThresholdInt (
        EMSIGAUGETHLDLOW, Off, EMSIGAUGETHLDHIGH, Off));

    } else {
      _classInfo->setDefault (I_estimateOfMeanSeverityIndicatingGaugeThreshold,
        new GaugeThresholdReal (
        EMSIGAUGETHLDLOW, Off, EMSIGAUGETHLDHIGH, Off));
    }

  }


  // the estimateOfMeanSeverityIndicatingTideMarkMax attribute
  // need to set the type first to TM_MAX as the GDMO compiler does not "know"
  estimateOfMeanSeverityIndicatingTideMarkMax() -> setType(TM_MAX);

  // the estimateOfMeanSeverityIndicatingTideMarkMin attribute
  // need to set the type first to TM_MAX as the GDMO compiler does not "know"
  estimateOfMeanSeverityIndicatingTideMarkMin() -> setType(TM_MIN);


  // associate the tide-marks to the gauge
  // unless the observed attribute it is a non-difference counter 

  if (_observedAttrType != OA_COUNTER ||
      monitorMetric::counterOrGaugeDifference() -> getbool() != False) {

    estimateOfMean() -> associateTideMarkMax(
	estimateOfMeanSeverityIndicatingTideMarkMax() );
    estimateOfMean() -> associateTideMarkMin(
	estimateOfMeanSeverityIndicatingTideMarkMin() );
  }

  if (!boolClassInitialised) { boolClassInitialised = True; }

  return OK;

} //movingAverageMeanMonitor::createRR 


void movingAverageMeanMonitor::update()
{
#if DEBUG
  advise(NULL, "movingAverageMeanMonitor::update()");
#endif

  AdminStateVal admState = *(AdminStateVal*)
    scanner::administrativeState()->get();
  if (admState == as_unlocked) {
  
    monitorMetric::updateDerivedGauge ();
      //update monitorMetric::derivedGauge attribute
  
  
    type_Monitor_TimePeriod* tp = (type_Monitor_TimePeriod*)
      scanner::granularityPeriod()->get();
  
    int iPollInterval = TimePeriod2seconds(tp);
  
    type_Monitor_TimePeriod* mtp = (type_Monitor_TimePeriod*)
      movingTimePeriod()->get();
    int mTimePeriod = TimePeriod2seconds(mtp);
  
  
    int admState = *(int*) scanner::administrativeState()->get();   
  
  
    if (iPollInterval+mTimePeriod != 0) {
      _smoothingConst = 
        (double)(2*iPollInterval)/(double)(iPollInterval+mTimePeriod);
    } else {
      _smoothingConst = 0.0;
    }
  
    double currentGaugeVal;
  
    if (this->gaugeType() == OV_INTEGER) {
      currentGaugeVal = 
       ((GaugeInt*)monitorMetric::derivedGauge())->getint(); 
    } else {
      currentGaugeVal = 
       ((GaugeInt*)monitorMetric::derivedGauge())->getreal(); 
    }
  
    double meanEstimate = 
      (_smoothingConst * currentGaugeVal) + 
      ((1.0 - _smoothingConst) * _oldMeanEstimate);
  
    _oldMeanEstimate = meanEstimate;
  
#if DEBUG
    fprintf(
      stdout,
      "movingAverageMeanMonitor::update() meanEstimate=%.3f\n",
      meanEstimate
    );
#endif
  
#if APPROXIMATING_UP 
    if ((meanEstimate -(int)meanEstimate > 0.5) && 
        ((this->gaugeType()) == OV_INTEGER)
    ) {
      estimateOfMean()->setreal(meanEstimate + 1.0); 
    } else {
      estimateOfMean()->setreal(meanEstimate); 
    }
#else
    estimateOfMean()->set(meanEstimate); 
#endif            
  
    if (estimateOfMean()->triggeredThreshold() == On) {
      _qosAlarmGauge   = estimateOfMean();
      _qosAlarmThldOid = _classInfo ->
		getAttrOid(I_estimateOfMeanSeverityIndicatingGaugeThreshold);
      triggerEvent(I_qualityofServiceAlarm, monitorMetric::_level);
    }
  
  }

} //movingAverageMeanMonitor::update


int  movingAverageMeanMonitor::set (
  CMISModifyOp op, int attrId, int classLevel, void* newval, AVA*& err,
  Bool checkOnly, int
) {

    if (classLevel != _level) {
        return monitorMetric::set(op, attrId, classLevel, newval, err, checkOnly);
    } 

#if DEBUG
    advise(NULL, "!!! movingAverageMeanMonitor::set");
#endif

    AdminStateVal admState = *(AdminStateVal*)
      scanner::administrativeState()->get();
    switch (attrId) {
    case I_estimateOfMean:
      if (admState == as_unlocked) {
        if (checkOnly)
	  return NOTOK;
        err = new AVA(m_processingFailure, "adminStateUnlocked");
        return NOTOK;
      } else {
        if (checkOnly)
	  return OK;

        ObservedValue* ov = (ObservedValue*)newval;
        switch (this->gaugeType()) {
        case OV_INTEGER:
	  estimateOfMean()->setreal((double)ov->ov_integer);
	  break;
        case OV_REAL:
	  estimateOfMean()->setreal((double)ov->ov_real);
	  break;
        }
#if DEBUG
        advise (NULL,
	  "movingAverageMeanMonitor::set typ=%d val=%d",
          ov->ov_type, ov->ov_integer);
#endif
      }
      break;
        
    case I_estimateOfMeanSeverityIndicatingGaugeThreshold:
    case I_movingTimePeriod:
      if (admState==as_unlocked) {
        if (checkOnly)
	  return NOTOK;
        err = new AVA(m_invalidOperation);
        return NOTOK;
      }
      break;

    default:
      break;
    } /* end switch */

    return OK;

} //movingAverageMeanMonitor::set
