//
// NAME
//   sumstat.cc
//
// DESCRIPTION
//   member functions of the class for the statistics
//   of the Monitoring Support objects
//
// IMPLEMENTORS
//   George Mykoniatis
//   NTUA
//   mykoniat@ektor.ntua.gr
//

#include "msap.h"

#include "sumstat.h"
// ------------------------------------------------------------------
double sumstats::get_last_value()
{
  if (index)
    return( array[index-1] );
  else
	return( array[no_of_samples-1] );
} //sumstats::get_last_value

// ------------------------------------------------------------------
int sumstats::initialize(int win, int prd, int alg, double iv)
{
  // Store the parameters
  algorithm = alg;
  if (alg == SUMSTAT_NTHNG)	// Keep only the current value
    window = prd;               // if in pass through mode
  else
    window = win;
  period = prd;

  index = 0;		        // point to first entry;
  buffer_flag = 0;		// Set to false

  // Check if invalid window or period
  if (prd == 0)
  {
    fprintf( stderr, "sumstat :: Invalid period %d\n", period);
    return NOTOK;
  }
  if (win < prd)
  {
    fprintf( stderr, "sumstat :: window < period %d<%d\n", win, period);
    return NOTOK;
  }
  // --------------------------
  // Calculate size of array
  if (alg == SUMSTAT_NTHNG)
	no_of_samples = 1;
  else
  {
    no_of_samples = win/prd;
    if (win%prd)      		// In case of not exact multiple...
      no_of_samples++;
  }	
  // Allocate buffer
  array = new double[no_of_samples];
  if (array==NULL)
  {
    fprintf( stderr, "sumstat :: Failed to allocate memory\n", win, period);
    return NOTOK;
  }
  // Initialize the buffer
  for (int i = 0; i<no_of_samples ; i++)
    array[i] = iv;

  // Set previous result to initial value (?).
  // Might by dangerous for the algorithms that require a previous result !!
  prev_val = iv;

  // Initialise the rest of the variables
  result = apply_algorithm();	// Initialise result

  return OK;
}  //sumstats::initialize

// ------------------------------------------------------------------
sumstats::~sumstats()
{
#ifdef DEBUG
  fprintf(stderr, "~sumstats: Executing..\n");
#endif

  if (array)	// release array
    delete array;
  return;
} //sumstats::~sumstats

// ------------------------------------------------------------------
int sumstats::append_value(double val)
{
  array[index] = val;
  index++;
  if (index == no_of_samples) 	// If buffer full ...
  {
    index = 0;  		// reset pointer
	if (buffer_flag == 0)
	{
	  double dtmp=0;
	  for(int i=0; i < no_of_samples; i++)
		dtmp += array[i];
	  prev_val = dtmp / no_of_samples;
  
      buffer_flag = 1;    // Set buffer full flag
    }
  }
  result = apply_algorithm();	// Update result
  return(index);
} //sumstats::append_value

// ------------------------------------------------------------------
double sumstats::apply_algorithm()
{
  double dtmp= 0.0;
  register int i=0; /*if not defined here, some switch cases bypass initialization (SUNCC4)*/

  switch (algorithm)
  {
    case SUMSTAT_NTHNG:
      dtmp = array[0];
      break;
    case SUMSTAT_UWMA:       // Uniform weighted moving average
      for(i=0; i<no_of_samples ; i++)
	    dtmp += array[i];
      dtmp /= no_of_samples;
      break;

    case SUMSTAT_EWMA:	     // Exponential weighted moving average
      switch (index)
      {
			case 0 :
			 {
			  dtmp = prev_val +
				 ( array[no_of_samples - 1] -
				   array[no_of_samples - 2]
				 ) / no_of_samples ;
			 }; break;

			case 1 :
			 {
			  dtmp = prev_val +
				 ( array[0] -
				   array[no_of_samples - 1]
				 ) / no_of_samples ;
			 }; break;
			default:
			 {
			  dtmp = prev_val +
				 ( array[index-1] -
				   array[index-2]
				 ) / no_of_samples ;
			 }; break;
      } // Switch (index)
      prev_val = dtmp;
      break;

  case SUMSTAT_WMA:     		// Weighted moving average
    register int j=index;

    i = 1;
    while (i <= no_of_samples)
    {
      dtmp += i * array[j];
      ( j == (no_of_samples-1)) ? j = 0 : j++;
      i++;
    }
    dtmp /= (no_of_samples*(no_of_samples +1)/2);
    break;
  } // Switch algorithm

  result = dtmp;
  return(dtmp);
}  //sumstats::apply_algorithm

// ------------------------------------------------------------------
void sumstats::print_sumstats()
{
  fprintf(stdout, "\n-------sumstats--------\n");
  fprintf(stdout, "Window : %d, Period : %d, Algorithm : %d, Samples : %d\n",
		   window, period, algorithm, no_of_samples);
  for (int i = 0; i<no_of_samples; i++)
    fprintf(stdout, "\tSample no %d : %lf\n", i, array[i]);
  fprintf(stdout, "\t\t*** Result : %lf\tFlag : %d\n", result, buffer_flag);
  fprintf(stdout, "-----------------------\n\n");
  return;
} //sumstats::print_sumstats

