/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
        execlex.l

        Lexical Analyser for GDMO script Interpreter

        Author: J.Cowan         (j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

/* %C       target language is C */

%{
static  int     EofCmd;
typedef struct  Reserved
        {
        char    *_word;
        int     _token;
        };
static  Reserved        *LookupReserved(char *);
#define DEBUGLEX(x)
/*
#define DEBUGLEX(x)     printf(x)
*/
#define MAXSTRING       1000
#undef  YY_INPUT
#define YY_INPUT(buf, result, max_size) \
        { int c; if (!EofCmd) c = fgetc(yyin); else c = EOF; \
          if (c=='$') EofCmd=1; result = (c==EOF) ? YY_NULL : (buf[0] = c, 1);}
#undef  YY_DECL
#define YY_DECL int InterPreter::Lex()
%}

%START  NEWFILE
%%
[ \t]                   { ; }

        /* Strings */

["]     {
        StringSymbol *ssym;
        char    str[MAXSTRING];
        int     count, ch;
        int     startLine = ExecLine;

        for (count=0; ((ch=yyinput()) != '\0') && count < MAXSTRING; count++)
        {
                switch (ch)
		{
		case '"':
                        if (count>0)
                                if (str[count-1] == '\\')
                                {
                                        str[count] = ch;
                                        continue;
                                }
                        str[count] = '\0';
                        /* printf("\nString %s", str);*/
                        ssym = (NEW StringSymbol(str, "STRING"));
                        AddSymbol(ssym, False, True);
                        yylval.sym = ssym;
                        DEBUGLEX("\nSTRING");
                        return STRING;

		case '$':
			/* horrid hack to allow $ in strings */
			EofCmd = 0;
			break;

		case '\n':
			ExecLine++;
			break;
                }
                str[count] = ch;
        }

        printf("\nError: unterminated string starting in line %d : ",
                startLine);
        for (int i=0; i<count; i++)
                printf("%c", str[i]);
        exit(1);
        }

        /* Numbers */

[0-9]*                  {
                        IntegerSymbol *isym;

                        sscanf(yytext, "%d", &yylval.intVal);
                        /*printf("\nInteger %d", yylval.intVal);*/
                        isym = (NEW IntegerSymbol(yylval.intVal, "INTEGER"));
                        AddSymbol(isym, False, True);
                        yylval.sym = (Symbol *)isym;
                        DEBUGLEX("\nNUMBER");
                        return NUMBER;
                        }

        /* Variables */

[a-zA-Z][a-zA-Z0-9]*    {
                        StringSymbol    *ssym;
                        BUILTIN         *entry;
                        Reserved        *reserved;

                        /*printf("\nText %s", yytext);*/
                        ssym = (NEW StringSymbol(yytext, "TEXT"));
                        AddSymbol(ssym, False, True);
                        yylval.text = ssym->GetValue();
                        if ((entry=BUILTINNode::Lookup(yytext)) !=
                                (BUILTIN *) 0)
                        {
                                DEBUGLEX("\nBUILTINFN");
                                if (entry->_fn)
                                        return BUILTINFN;
                                else
                                        return BUILTINPROC;
                        }
                        if ((reserved=LookupReserved(yytext)) !=
                                        (Reserved *) 0)
                                return reserved->_token;

                        DEBUGLEX("\nTEXT");
                        return TEXT;
                        }

        /* Comments */

"//"[^\n]*              ;

        /* Other lexemes */

"."                     { DEBUGLEX("\nDOT"); return DOT;}
"="                     { DEBUGLEX("\nASSIGN"); return ASSIGN;}
">="                    { DEBUGLEX("\nGE"); return GE;}
">"                     { DEBUGLEX("\nGT"); return GT;}
"<"                     { DEBUGLEX("\nLT"); return LT;}
"<="                    { DEBUGLEX("\nLE"); return LE;}
"=="                    { DEBUGLEX("\nEQ"); return EQ;}
"!="                    { DEBUGLEX("\nNEQ"); return NEQ;}
"+"                     { DEBUGLEX("\nPLUS"); return PLUS;}
"-"                     { DEBUGLEX("\nMINUS"); return MINUS;}
"*"                     { DEBUGLEX("\nMULTIPLY"); return MULTIPLY;}
"/"                     { DEBUGLEX("\nDIVIDE"); return DIVIDE;}
","                     { DEBUGLEX("\nCOMMA"); return COMMA;}
";"                     { DEBUGLEX("\nSEMI"); return SEMI;}
"{"                     { DEBUGLEX("\nLBRACE"); return LBRACE;}
"}"                     { DEBUGLEX("\nRBRACE"); return RBRACE;}
"("                     { DEBUGLEX("\nLBRACKET"); return LBRACKET;}
")"                     { DEBUGLEX("\nRBRACKET"); return RBRACKET;}
"&"                     { DEBUGLEX("\nBITAND"); return BITAND; }
"&&"                    { DEBUGLEX("\nAND"); return AND; }
"|"                     { DEBUGLEX("\nBITOR"); return BITOR; }
"||"                    { DEBUGLEX("\nOR"); return OR; }
"!"                     { DEBUGLEX("\nNOT"); return NOT; }

"$"                     { BEGIN NEWFILE ; DEBUGLEX("\nEOF$"); return 0; }
[\n]                    { ExecLine++;}

<NEWFILE><<EOF>>        { EofCmd = 0;DEBUGLEX("\nSTART"); YY_NEW_FILE; BEGIN(0); }

%%

/* Reserved words */

static  Reserved        reservedWords[] =
{
        {"FOREACH",     FOREACH},
        {"IN",          IN},
        {"IF",          IF},
        {"ELSE",        ELSE},
        {"WHILE",       WHILE},
        {"GLOBAL",      GLOBAL},
        {"STRING",      STRINGTYPE},
        {"INTEGER",     INTEGERTYPE},
        {"OBJECT",      OBJECTTYPE},
        {"AND",         AND},
        {"OR",          OR},
        {"BITAND",      BITAND},
        {"BITOR",       BITOR},
        {"NOT",         NOT},
	{"CLASS",	CLASS},
	{"END",		END},
	{"EXTERN",	EXTERNTYPE},
        {"LIST",  	LISTTYPE},
	{"OF",		OF}
};

static  Reserved *LookupReserved(char *text)
{
int     count;
int     lgth = sizeof(reservedWords) / sizeof(Reserved);
Reserved *entry;

        for (count=0; count<lgth; count++)
        {
                entry = &reservedWords[count];
#ifdef  __MSDOS__
                if (!stricmp(text, entry->_word))
#else
                if (!strcasecmp(text, entry->_word))
#endif
                {
                        DEBUGLEX("\n");
                        DEBUGLEX(entry->_word);
                        return entry;
                }
        }

        return (Reserved *) 0;

} /* LookupReserved */

/****************************************************************/
/* End of file */
/****************************************************************/




