/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
        ExecYacc.y

        YACC Parser for GDMO script interpreter

        Author: J.Cowan         (j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

/* DEFINITIONS */
%{
static  char    *yytext;
static	Bool	GlobalVariable;
#define yylex   Lex
/*
#define YYDEBUG   1
*/
%}

/* Yacc Stack Typedef */

%union
        {
        Symbol	*sym;
        char	*text;
        Node	*node;
        int	intVal;
	SymType	type;
        }

/* Start symbol of the grammar */

%start begin

%token  SEMI FOREACH IN IF ELSE LBRACE RBRACE LBRACKET RBRACKET COMMA
%token  STRINGTYPE INTEGERTYPE OBJECTTYPE ASSIGN WHILE DOT GLOBAL
%token	CLASS END STRINGLISTTYPE EXTERNTYPE LISTTYPE OF 
%token  <sym>   STRING NUMBER
%token  <text>  TEXT BUILTINFN BUILTINPROC
%type   <node>  statements statement exprlist expr var
%left   OR
%left   AND
%left   BITOR
%left   BITAND
%left   GT GE LT LE EQ NEQ
%left   PLUS MINUS
%left   MULTIPLY DIVIDE
%left   NOT

%% /* grammar rules */

begin:          statements
                {
                        FirstNode($1);
                }
		|
                {
                        FirstNode((Node *) 0);
                }
                ;

statements:     statement
                {
                        $$ = $1;
                }
                |
                statements statement
                {
                SeqNode *seqNode = (NEW SeqNode($1, $2, this));

                        $$ = seqNode;
                }
                ;

statement:      SEMI    /* ; statement */
                {
                NullNode *nullNode = (NEW NullNode(this));

                        $$ = nullNode;
                }
		|
		CLASS TEXT /* DECLARE CLASS */
		{
			ClassDictionary::AddClass($2);
			// printf("\nDECLARE CLASS %s", $2);
		}
		class_members END
                {
                NullNode *nullNode = (NEW NullNode(this));

                        $$ = nullNode;
		}
		|
		GLOBAL
		{
			// nasty hack to flag global variables

			GlobalVariable = True; 
		}
		var
		{
			GlobalVariable = False; 
			$$ = $3;
		}
		|
		var
		{
			$$ = $1;
		}
                |
                TEXT ASSIGN expr
                {
                Symbol		*sym = Lookup($1);
                AssignNode	*node = (NEW AssignNode(sym, $3, this));

                        if (sym == (Symbol *) 0)
                                yynerrs++;

                        $$ = node;
                }
                |
                FOREACH TEXT IN expr statement
                {
                Symbol	*var = Lookup($2);
                ForNode	*forNode = (NEW ForNode(var, $4, $5, this));

                        if (var == (Symbol *) 0)
                                yynerrs++;
                        else
                                if (!forNode->Initialise())
                                        yynerrs++;
                        $$ = forNode;
                }
                |
                IF LBRACKET expr RBRACKET statement
                {
                IfNode *ifNode = (NEW IfNode($3, $5, this));

                        $$ = ifNode;
                }
                |
                IF LBRACKET expr RBRACKET statement ELSE statement
                {
                IfElseNode *ifElseNode = (NEW IfElseNode($3, $5, $7, this));

                        $$ = ifElseNode;
                }
                |
                WHILE LBRACKET expr RBRACKET statement
                {
                WhileNode *node = (NEW WhileNode($3, $5, this));

                        $$ = node;
                }
                |
                BUILTINPROC LBRACKET exprlist RBRACKET
                {
                BUILTINNode   *node =
                        (NEW BUILTINNode($1, (ExpNode *)$3, this));

                        $$ = node;
                }
                |
                LBRACE statements RBRACE
                {
                        $$ = $2;
                }
                ;

exprlist:       expr
                {
                        $$ = $1;
                }
                |
                exprlist COMMA expr
                {
                ExprListNode    *node =
                (NEW ExprListNode((ExpNode *)$1, (ExpNode *)$3, this));

                        $$ = node;
                }
                ;

expr:           TEXT
                {
                Symbol	*sym = Lookup($1);
                VarNode	*varNode = (NEW VarNode(sym, this));

                        if (sym == (Symbol *) 0)
                                yynerrs++;
                        $$ = varNode;
                }
                |
                TEXT DOT TEXT
                {
                Symbol		*sym1 = Lookup($1);
                VarMemberNode	*node = (NEW VarMemberNode(sym1, $3, this));

                        if (sym1 == (Symbol *) 0)
                                yynerrs++;
			else
                        	if (!node->Initialise())
                                	yynerrs++;

                        $$ = node;
                }
                |
                STRING
                {
                VarNode *strNode = (NEW VarNode($1, this));

                        $$ = strNode;
                }
                |
                NUMBER
                {
                VarNode *node = (NEW VarNode($1, this));

                        $$ = node;
                }
                |
                LBRACKET expr RBRACKET
                {
                        $$ = $2;
                }
                |
                expr PLUS expr
                {
                PlusNode        *node = (NEW PlusNode($1, $3, this));

                        $$ = node;
                }
                |
                expr MINUS expr
                {
                MinusNode       *node = (NEW MinusNode($1, $3, this));

                        $$ = node;
                }
                |
                expr MULTIPLY expr
                {
                MultiplyNode    *node = (NEW MultiplyNode($1, $3, this));

                        $$ = node;
                }
                |
                expr DIVIDE expr
                {
                DivideNode      *node = (NEW DivideNode($1, $3, this));

                        $$ = node;
                }
                |
                expr GT expr
                {
                GTNode  *node = (NEW GTNode($1, $3, this));

                        $$ = node;
                }
                |
                expr GE expr
                {
                GENode  *node = (NEW GENode($1, $3, this));

                        $$ = node;
                }
                |
                expr LT expr
                {
                LTNode  *node = (NEW LTNode($1, $3, this));

                        $$ = node;
                }
                |
                expr LE expr
                {
                LENode  *node = (NEW LENode($1, $3, this));

                        $$ = node;
                }
                |
                expr EQ expr
                {
                EQNode  *node = (NEW EQNode($1, $3, this));

                        $$ = node;
                }
                |
                expr NEQ expr
                {
                NEQNode *node = (NEW NEQNode($1, $3, this));

                        $$ = node;
                }
                |
                expr BITAND expr
                {
                BITANDNode *node = (NEW BITANDNode($1, $3, this));

                        $$ = node;
                }
                |
                expr BITOR expr
                {
                BITORNode *node = (NEW BITORNode($1, $3, this));

                        $$ = node;
                }
                |
                expr AND expr
                {
                ANDNode *node = (NEW ANDNode($1, $3, this));

                        $$ = node;
                }
                |
                expr OR expr
                {
                ORNode *node = (NEW ORNode($1, $3, this));

                        $$ = node;
                }
                |
                NOT expr
                {
                NOTNode *node = (NEW NOTNode($2, this));

                        $$ = node;
                }
                |
                BUILTINFN LBRACKET exprlist RBRACKET
                {
                BUILTINNode   *node =
                        (NEW BUILTINNode($1, (ExpNode *)$3, this));

                        $$ = node;
                }
                ;

var:		STRINGTYPE TEXT  /* Variable Declaration */
                                /* Uses Null node so no execution */
                {
                StringSymbol	*ssym = (NEW StringSymbol("", $2));
                NullNode        *node = (NEW NullNode(this));

                        if (!AddSymbol(ssym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE ssym;
                        }
                        $$ = node;
                }
                |
                INTEGERTYPE TEXT
                {
                IntegerSymbol   *isym = (NEW IntegerSymbol(0, $2));
                NullNode        *node = (NEW NullNode(this));

                        if (!AddSymbol(isym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE isym;
                        }
                        $$ = node;
                }
                |
                OBJECTTYPE TEXT
                {
                ObjectRefSymbol *sym = (NEW ObjectRefSymbol($2));
                NullNode        *node = (NEW NullNode(this));

                        if (!AddSymbol(sym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE sym;
                        }
                        $$ = node;
                }
		|
		LISTTYPE OF list_type
		{
                NullNode        *node = (NEW NullNode(this));

			$$ = node;
		}
                ;

list_type:	STRINGTYPE	TEXT
                {
                Symbol *sym = (NEW StringSymbolList(True, $2));

                        if (!AddSymbol(sym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE sym;
                        }
		}
		|
		INTEGERTYPE	TEXT
                {
                Symbol *sym = (NEW StringSymbolList(True, $2));

                        if (!AddSymbol(sym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE sym;
                        }
		}
		|
		OBJECTTYPE	TEXT
                {
                Symbol *sym = (NEW ObjectRefSymbolList(True, $2));

                        if (!AddSymbol(sym, True, GlobalVariable))
                        {
                                yynerrs++;
                                DELETE sym;
                        }
		}
		;

class_members:	class_member
		|
		class_members class_member
		;

class_member:	TEXT
		{
			// printf("\nCLASS MEMBER %s", $1);
			ClassDictionary::AddClassMember($1);
		}
		member_extern
		;

member_extern:	EXTERNTYPE member_type
		{
			ClassDictionary::SetClassMemberExternal(True);
		}
		|
		member_type
		{
			ClassDictionary::SetClassMemberExternal(False);
		}
		;

member_type:	STRINGTYPE
		{
			ClassDictionary::SetClassMemberType(StringTypeDef);
		}
		|
		INTEGERTYPE
		{
			ClassDictionary::SetClassMemberType(IntegerTypeDef);
		}
		|
		OBJECTTYPE
		{
			ClassDictionary::SetClassMemberType(ObjectRefTypeDef);
		}
		|
		LISTTYPE OF STRINGTYPE
		{
			ClassDictionary::SetClassMemberType(StringListTypeDef);
		}
		|
		LISTTYPE OF INTEGERTYPE
		{
			ClassDictionary::SetClassMemberType(IntegerListTypeDef);
		}
		|
		LISTTYPE OF OBJECTTYPE
		{
			ClassDictionary::SetClassMemberType(ObjectRefListType);
		}
		;

/****************************************************************/
%% /* functions */

static void yyerror(char *s)
{
        printf ("\nError: %s in line %d at token <%s>\n",
                s, ExecLine, yytext);
}

/****************************************************************/

int ExecParseErrors()
{
        return yynerrs;

} /* ExecParseErrors */

/****************************************************************/
/* End of Module */
/****************************************************************/


