/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMO.h

	Class that represents various GDMO services

	Author:	J.Cowan		(j.cowan@.cs.ucl.ac.uk)

*/
/****************************************************************/

#include "Boolean.h"

#ifndef	_SYMBOL_
#include "Symbol.h"
#endif

#ifndef	_GENERIC_LIST_
#include "GenList.h"
#endif

#ifndef	_GENERIC_STRING_
#include "GenStr.h"
#endif

#ifndef	_GDMOGEN_
#include "GDMOGen.h"
#endif

#ifndef _IIMCH_
#include "IIMC.h"
#endif

/****************************************************************/

#define	MAXLEXBUFFER	5000

/****************************************************************/

class GDMO : public MetaClass
{
	StringSymbol	_gdmoDirectory;	// Home directory
	GDMOList	_moClassList;	// Mo class list
	GDMOPackageList	_packageList;	// Package list
	GDMOAttrList	_attrList;	// Attribute list
	GDMOActionList	_actionList;	// Action list
	GDMONotificationList		// notification list
			_notificationList;
	GDMONameBindingList
			_nameBindingList;// Name Bindings  list
	GDMOAttrGroupList
			_attrGroupList;	// Attribute group list
	GDMOBehaviourList
			_behaviourList;	// Behaviour list

	GDMOClass	*_curMoClass;	// Current Mo Class
	GDMOPackage	*_curPackage;	// Current Package
	GDMOAttribute	*_curAttribute;	// Current Attribute
	GDMOAttribute	*_curPkgAttribute; // Current package Attribute
	GDMONameBinding	*_curNameBinding;// Current name Binding
	GDMOAction	*_curAction;	// Current Action
	GDMONotification *_curNotification;// Current Notification
	GDMOAttributeGroup
			*_curAttrGroup;	// Current Attribute group
	GDMOBehaviour	*_curBehaviour;	// Current behaviour
	GDMOOid		_curOid;	// Current Oid
	Bool		_error;
	StringSymbolList
			_behavRefList;	// Behaviour reference list

	SymbolTable     _lexSymTable;
	char		_lexBuffer[MAXLEXBUFFER];

	void	Initialise();
	Bool	UpdateRefs();
	void	PrintParents();
	void	PrintChildren(char *, int);
	void	PrintSubordinates(GDMOClass *, int);
	void	PrintError(char	*msg)
	{
		printf("\nError: %s", msg);
		_error = True;
	}
	int	ParseGDMO();

	// Lexical functions

	int	Lex();
	char	*readToQuote(int);
	char	*getText();
	char	*getDocId();
	char	*getDelimitedText();
	char	*HeapAlloc(char *);

	Bool	ParseIIMC(MetaClass *metaClass, IIMCBeginState state);

	// Internal assertion checks

	void	checkCurrentMoClass(int);
	void	checkCurrentPackage(int);
	void	checkCurrentAttribute(int);
	void	checkCurrentNameBinding(int);
	void	checkCurrentAction(int);
	void	checkCurrentNotification(int);
	void	checkCurrentAttributeGroup(int);
	void	checkCurrentBehaviour(int);

	void	NewMoClass(char *);
	void	AddMoParent(char *);
	void	AddMoPackage(char *);
	void	NewPackage(char *);
	void	AddPackageAttribute(char *);
	void	UpdatePackageAttribute(char *, char *);
	void	UpdatePackageAttribute(char *, int, UpdateMode);
	void	AddPackageAction(char *);
	void	AddPackageNotification(char *);
	void	AddPackageAttributeGroup(char *);
	void	UpdatePackageAttrGroup(char *, Symbol *);
	void	NewAttribute(char *);
	void	AddAttributeDerivation(char *);
	void	AddAttributeSyntax(char *);
	void	UpdateAttribute(char *, char *);
	void	UpdateAttribute(char *, int, UpdateMode);
	void	NewAction(char *);
	void	UpdateAction(char *, char *);
	void	UpdateAction(char *, int, UpdateMode);
	void	AddActionInfoSyntax(char *);
	void	AddActionReplySyntax(char *);
	void	NewNotification(char *);
	void	UpdateNotification(char *, char *);
	void	UpdateNotification(char *, int, UpdateMode);
	void	AddNotificationInfoSyntax(char *);
	void	AddNotificationReplySyntax(char *);
	void	NewNameBinding(char *);
	void	AddNameBinding(char *, char *, char *);
	void	AddCreateModifier(int);
	void	AddDeleteModifier(int);
	void	NewAttributeGroup(char *);
	void	UpdateAttributeGroup(char *, int, UpdateMode);
	void	UpdateAttributeGroup(char *, Symbol *);
	void	NewBehaviour(char *);
	void	UpdateBehaviour(char *, char *);

	void	ClearRegistration();
	void	AddRegistration(char *);
	void	AddMoRegistration();
	void	AddPackageRegistration();
	void	AddNameBindingRegistration();
	void	AddAttributeRegistration();
	void	AddAttrGroupRegistration();
	void	AddActionRegistration();
	void	AddNotificationRegistration();
public:
	GDMO(char *gdmoDir) : 
		MetaClass("GDMO"),
		_moClassList("ManagedObjectList"),
		_packageList("PackageList"),
		_attrList("AttributeList"),
		_actionList("ActionList"),
		_notificationList("NotificationList"),
		_nameBindingList("NamebindingsList"),
		_attrGroupList("AttributeGroupList"),
		_behaviourList("BehaviourList"),
		_gdmoDirectory(gdmoDir, "GDMODirectory"),
		_behavRefList("BehaviourReferenceList"),
		_curOid("curOid"),
		_lexSymTable(True)
	{
		Initialise();

		AddSymbol(&_moClassList);
		AddSymbol(&_packageList);
		AddSymbol(&_attrList);
		AddSymbol(&_actionList);
		AddSymbol(&_notificationList);
		AddSymbol(&_nameBindingList);
		AddSymbol(&_attrGroupList);
		AddSymbol(&_behaviourList);
		AddSymbol(&_behavRefList);
		AddSymbol(&_gdmoDirectory);
	}
	Bool	Parse(char *);
	Bool	ParseIIMCBehaviour();

	void	PrintMoList();
	void	PrintPackageList();
	void	PrintAttrList();
	void	PrintInheritanceHierarchy();
	void	PrintContainmentHierarchy();
	void	ReadMoDatabase(char *);
	void	CodeGenerate(char *);
};

/****************************************************************/



