/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMOClass.h

	Class that represents a	GDMO Managed Object Class definition

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#ifndef	_SYMBOL
#include "Symbol.h"
#endif

#ifndef	_GENERIC_LIST_
#include "GenList.h"
#endif

#ifndef	_GENERIC_STRING_
#include "GenStr.h"
#endif

/****************************************************************/
// Forward references

class	GDMOClass;

class	GDMONameBinding;

/****************************************************************/

class	GDMOList : public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	GDMOList(Bool del, char	*name)
	: ListSymbol(name, "GDMOClass", ObjectTypeDef, del) {}
	GDMOList(char *name)
	: ListSymbol(name, "GDMOClass", ObjectTypeDef) {}

	~GDMOList() { clear();}

	GDMOClass	*FindMo(char *);
};

/****************************************************************/

class GDMONameBindingList : public ListSymbol
{
protected:
	void	deleteElem (void*);
public:
	GDMONameBindingList(Bool del, char *name)
	: ListSymbol(name, "GDMONameBinding", ObjectTypeDef, del) {}
	GDMONameBindingList(char *name)
	: ListSymbol(name, "GDMONameBinding", ObjectTypeDef) {}
	~GDMONameBindingList() { clear();}
};

/****************************************************************/

class	GDMORefList : public List
{
protected:
	void	deleteElem (void*);
public:
	~GDMORefList() { clear();}
};

/****************************************************************/

class GDMOClass	: public MetaClass
{
	StringSymbol	_name, _className, _headerFileName;
	IntegerSymbol	_externalMO;

	GDMOPackageRefList	_packageRefList;
	GDMORefList		_parentRefList;
	GDMOList		_childList, _parentList;
	GDMOAttrList		_attrList;
	GDMOAttrGroupList	_attrGroupList;
	GDMONameBindingList	_nameBindingList;
	GDMOActionList		_actionList;
	GDMONotificationList	_notificationList;
	GDMOOid			_oid;

	Bool	UpdateMoRefs(GDMORefList *, GDMOList *);
	Bool	UpdateMoPackageRefs(GDMOPackageRefList *,
				    GDMOPackageList *);
	Bool	UpdateChildRefs(GDMOList *);
	void	UpdateFromPackage();
	void	UpdateParentList();

	void	PrintError(char *msg)
	{
		printf("\nError: %s", msg);
	}
public:
	GDMOClass(char *name, char *className, char *headerFileName,
		  int externalMO) :
			MetaClass("GDMOClass"),
			_name(name, "MOName"),
			_className(className, "ClassName"),
			_headerFileName(headerFileName, "HeaderFileName"),
			_childList(False, "ChildList"),
			_parentList(False, "ParentList"),
			_attrList(False, "AttributeList"),
			_attrGroupList(False, "AttributeGroupList"),
			_nameBindingList(False, "NamebindingsList"),
			_actionList(False, "ActionList"),
			_notificationList(False, "NotificationList"),
			_oid("Oid"),
			_externalMO(externalMO, "External")
	{
		AddSymbol(&_name);
		AddSymbol(&_className);
		AddSymbol(&_headerFileName);
		AddSymbol(&_childList);
		AddSymbol(&_parentList);
		AddSymbol(&_attrList);
		AddSymbol(&_attrGroupList);
		AddSymbol(&_nameBindingList);
		AddSymbol(&_actionList);
		AddSymbol(&_notificationList);
		AddSymbol(&_oid);
		AddSymbol(&_externalMO);
	}

	// return name of managed object

	char	*GetName()
	{
		return _name.print();
	}

	// Get list of attributes

	GDMOAttrList *GetAttributes()
	{
		return(&_attrList);
	}

	// Get list of children

	GDMOList *GetChildren()
	{
		return(&_childList);
	}

	// Get list of parents

	GDMOList *GetParents()
	{
		return(&_parentList);
	}

	// Package reference functions

	void	AddPackageRef(char *);

	// parent reference functions

	void	AddParentRef(char *);

	// child class

	void	AddChild(GDMOClass *);

	// Add to name binding list

	void	AddNameBinding(GDMONameBinding *);

	GDMONameBindingList *GetNameBindings() { return &_nameBindingList;}

	// Update all references

	Bool	UpdateRefs(GDMOList *, GDMOPackageList *);

	// Find attribute

	GDMOAttribute *FindAttr(char *);

	GDMOOid	*GetOid() { return &_oid;}
};

/****************************************************************/

class GDMOClassRef
{
	GString		_name;
	GDMOClass	*_mo;
public:
	GDMOClassRef(char *name)
	{
		_name = name;
		_mo = (GDMOClass *) 0;
	}

	void	UpdateRef(GDMOClass *);
	char	*GetName()
	{
		return (char *)_name;
	}
	GDMOClass *GetMO()
	{
		if (!_mo)
			printf("\nMO Class %s not updated",
				(char *) _name);
		return _mo;
	}
};

/****************************************************************/

typedef	enum
	{Create = 1, CreateWithReference = 2, CreateWithAutoInstance = 4}
	CREATEMODIFIER;

typedef	enum
	{Delete = 1, DeleteIfNoContained = 2, DeleteIfContained = 4}
	DELETEMODIFIER;

class GDMONameBinding : public MetaClass
{
	StringSymbol	_name, _superiorName,
			_subordinateName, _attributeName;
	IntegerSymbol	_create, _delete;
	GDMOOid		_oid;
	GDMOClass	*_superiorMo, *_subordinateMo;
	GDMOAttribute	*_attribute;
public:
	GDMONameBinding(char *name) :
			MetaClass("GDMONameBinding"),
			_name(name, "BindingName"),
			_superiorName("", "Superior"),
			_subordinateName("", "Subordinate"),
			_attributeName("", "NamingAttribute"),
			_create(False, "Create"),
			_delete(False, "Delete"),
			_oid("Oid")
	{
		_superiorMo = _subordinateMo = (GDMOClass *) 0;
		_attribute = (GDMOAttribute *) 0;
		AddSymbol(&_name);
		AddSymbol(&_superiorName);
		AddSymbol(&_subordinateName);
		AddSymbol(&_attributeName);
		AddSymbol(&_create);
		AddSymbol(&_delete);
		AddSymbol(&_oid);
	}

	void	AddNameBinding(char *subordinate, char *superior,
			       char *attribute)
	{
		_subordinateName.StrReplace(subordinate);
		_superiorName.StrReplace(superior);
		_attributeName.StrReplace(attribute);
	}

	// Get Name Binding name

	char	*GetName()		{return (char *) _name;}

	// Get superior MO

	GDMOClass *GetSuperiorMo()	{return _superiorMo;}

	// Get subordinate MO

	GDMOClass *GetSubordinateMo()	{return _subordinateMo;}

	// Get subordinate MO

	GDMOAttribute *GetAttribute()	{return _attribute;}

	void	AddCreateModifier(int);

	void	AddDeleteModifier(int);

	// Get Name Binding List

	// Update References

	Bool	UpdateRefs(GDMOList *);

	GDMOOid	*GetOid() { return &_oid;}
};

/****************************************************************/
// inline methods for GDMOList

inline void  GDMOList::deleteElem (void* val)
{  if (del) delete (GDMOClass *) val; }

/****************************************************************/
// inline methods for GDMORefList

inline void  GDMORefList::deleteElem (void* val)
{  if (del) delete (GDMOClassRef *) val; }

/****************************************************************/
// inline methods for GDMONameBindingList

inline void  GDMONameBindingList::deleteElem (void* val)
{  if (del) delete (GDMONameBinding *) val; }

/****************************************************************/





