/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/*******************************************************/
/*
	GDMOMain.cc

	main program for GDMO

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/*******************************************************/

#ifdef	__MSDOS__
#include <io.h>
#else
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "GDMOGen.h"
#include "GDMOOid.h"
#include "GDMOAttr.h"
#include "GDMOAct.h"
#include "GDMOBeh.h"
#include "GDMONot.h"
#include "GDMOPkg.h"
#include "GDMOClas.h"
#include "GDMO.h"
#include "SymExec.h"

/*******************************************************/

#define	DEBUGFILE	"gdmo.dbg"
#define	ATTRDATABASE	"attr.dat"
#define	SYNTAXDATABASE	"syntax.dat"

/*******************************************************/
/* Global Variables */

char	*TopMo = DEFAULTTOPMO;
FILE	*DebugFile;

static	char	*AttrDatabase, *UserSyntaxDatabase, *ScriptName,
		*MoDatabase;

static	char	*InitScriptName = "init.gen";

/*******************************************************/

static	void	checkArg(char *arg, int &count, int argc)
{

	if ((count + 1) > argc)
	{
		printf("\nOption %s requires a further argument",
			arg);
		exit(1);
	}

} /* checkArg */

/*******************************************************/

static	void	GdmoArgs(int argc, char *argv[],
			 char **fileName, Bool *debugFlag)
{
	for (int count=1; count<argc; count++)
	{
	char	*arg = argv[count];

		if (*arg == '-')
		{
			arg++;
			switch (*arg)
			{
			default:
				printf("\nUnknown option %c", *arg);
				exit(1);

			case 'd':
			case 'D':
				*debugFlag = True;
				break;

			case 't':
			case 'T':
				checkArg(arg, count, argc);
				TopMo = argv[++count];
				break;

			case 'm':
			case 'M':
				checkArg(arg, count, argc);
				MoDatabase = argv[++count];
				break;

			case 'a':
			case 'A':
				checkArg(arg, count, argc);
				AttrDatabase = argv[++count];
				break;

			case 's':
			case 'S':
				checkArg(arg, count, argc);
				UserSyntaxDatabase = argv[++count];
				break;

			case 'x':
			case 'X':
				checkArg(arg, count, argc);
				ScriptName = argv[++count];
				break;

			case 'i':
			case 'I':
				checkArg(arg, count, argc);
				InitScriptName = argv[++count];
				break;
			}
		}
		else
		{
			if (*fileName == (char *) 0)
				*fileName = arg;
			else
			{
				printf("\nMore than one filename: %s, %s",
					*fileName, arg);
				exit(1);
			}
		}
	}

	if (argc < 2 || (*fileName == (char *)0))
	{
		printf("\nUsage is %s input file options.\n",
			argv[0]);
		printf("\nOptions are:\n-d (debug)");
		printf("\n-t TopMO (default top)");
		printf("\n-m MOdatabaseName (default gdmodir/mo.dat) ");
		printf("\n-a AttributeDatabaseName (default gdmodir/attr.dat)");
		printf("\n-s SyntaxDatabaseName (default gdmodir/syntax.dat)");
		printf("\n-x ScriptName (default gdmodir/start.gen)\n\n");
		printf("\n-i Initial ScriptName (default gdmodir/init.gen)\n\n");
		exit(1);
	}

} /* getArgs */

/*******************************************************/

static	void	debugGDMO(GDMO *CurGDMO)
{
int	done = False;
int	display = True;

	while (!done)
	{
	int	ch;

		if (display)
			printf("\n\n1) Generate Code\
				\n\n2) Display managed objects\
				\n\n3) Display inheritance hierarchy\
				\n\n4) Display containment hierarchy\
				\n\n5) Display Symbol Table Structure\
				\n\n6) Quit\n\nSelect option: ");
		ch = getchar();
		display = True;
		switch (ch)
		{
		case '1':
			CurGDMO->CodeGenerate(ScriptName);
			break;
		case '2':
			CurGDMO->PrintMoList();
			break;
		case '3':
			CurGDMO->PrintInheritanceHierarchy();
			break;
		case '4':
			CurGDMO->PrintContainmentHierarchy();
			break;
		case '5':
			CurGDMO->DumpSymTableClasses();
			break;
/*
		case '6':
			CurGDMO->DumpSymTableContents();
			break;
*/
		case '6':
			done = True;
			break;
		default:
			display = False;
		}
	}

} /* debugGDMO */

/*******************************************************/
// Initialise Meta Class information

static	void	InitMetaClass(char *gdmodir)
{
InterPreter     *interPreter;
SymbolTable	symtable(True);

	// See if STARTCODE in local directory

	if (access(InitScriptName, 0) == 0)
	{
		printf("\nExecuting %s\n", InitScriptName);
		interPreter = (NEW InterPreter(InitScriptName, CODELOG,
			       &symtable));
	}
	else
	{
	GString tmp = GString(gdmodir) + GString(InitScriptName);

		if (access((char *)tmp, 0) == 0)
		{
			printf("\nExecuting %s", (char *) tmp);
			interPreter = (NEW InterPreter((char *) tmp, CODELOG,
						       &symtable));
		}
		else
		{
			printf("\nCannot find file %s", (char *)tmp);
			exit(1);
		}
	}

	if (interPreter->Execute() == False)
	{
		printf("\nCannot initialise");
		exit(1);
	}

	DELETE interPreter;

} /* InitMetaClass */

/*******************************************************/
/* Main function */
/*******************************************************/

main(int argc, char *argv[])
{
GDMO	*CurGDMO;
char	*fileName = (char *) 0;
Bool	debugFlag = False;
int	exitResult = 0;
GString	gdmoDir;
char	*envGdmoDir;

	// Get GDMODIR environment variable

	if ((envGdmoDir=getenv("GDMODIR")) == (char *) 0)
	{
#ifdef	__MSDOS__
		printf("\nWarning: no GDMODIR environment variable");
		gdmoDir = "";
#else
		// gdmoDir = GDMOPWD;
		printf("\nGDMODIR environment variable must be defined");
		exit(1);
#endif
	}

	gdmoDir = envGdmoDir;
#ifndef	__MSDOS__
	gdmoDir += "/";		// Add slash
#endif

	GString	moDbase(gdmoDir), attrDbase(gdmoDir),
		syntaxDbase(gdmoDir);

	// Set up default global database paths
	// can be changed by command line arguments

	moDbase += MODATABASE;
	MoDatabase = (char *) moDbase;
	attrDbase += ATTRDATABASE;
	AttrDatabase = (char *) attrDbase;
	syntaxDbase += SYNTAXDATABASE;
	ScriptName = STARTCODE;

	GdmoArgs(argc, argv, &fileName, &debugFlag);

	// Initialise Meta Class information

	InitMetaClass((char *)gdmoDir);

	GDMOAttribute::AddAttributeDatabase(AttrDatabase);

	// Until the attribute compiler is finished - we
	// use two syntax databases - one global and
	// one user supplied

	GDMOAttribute::AddSyntaxDatabase((char *) syntaxDbase);

	// Now read the user database if any

	if (UserSyntaxDatabase != (char *) 0)
		GDMOAttribute::AddSyntaxDatabase(UserSyntaxDatabase);

	CurGDMO = (NEW GDMO((char *)gdmoDir));

	// Read external MO database

	CurGDMO->ReadMoDatabase(MoDatabase);

	MetaClass::AddGlobalSymbol(CurGDMO);

	/* Open Debug file */

	if (access(DEBUGFILE,0) == 0)
		DebugFile = fopen(DEBUGFILE, "w");

#ifdef	DEBUG_MEMORY
	init_mem_trace();
#endif

	if (CurGDMO->Parse(fileName))
	{
/*
		CurGDMO->PrintMoList(); getchar();
		CurGDMO->PrintPackageList(); getchar();
		CurGDMO->PrintAttrList(); getchar();
*/
		if (CurGDMO->ParseIIMCBehaviour())
		{
			if (debugFlag)
				debugGDMO(CurGDMO);
			else
				CurGDMO->CodeGenerate(ScriptName);
		}
	}
	else
		exitResult = 1;

	DELETE CurGDMO;

	// Delete static Class Dictionary

	ClassDictionary::DeleteClassDictionary();

#ifdef	DEBUG_MEMORY
	print_mem_trace();
#endif
	exit(exitResult);
	return 0;
}

/*******************************************************/
/* End of program */
/*******************************************************/

