/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	GDMOPackage.cc

	Methods of a GDMO Package Class

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "DebugMem.h"
#include "GDMOOid.h"
#include "GDMOAttr.h"
#include "GDMOAct.h"
#include "GDMONot.h"
#include "GDMOPkg.h"

/****************************************************************/

GDMOAttribute	*GDMOPackage::AddAttribute(char *attribute)
{
GDMOAttribute	*attr =	(NEW GDMOAttribute(attribute));

	_attrRefList.append(attr);
	return attr;

}

/****************************************************************/

void GDMOPackage::AddAction(char *action)
{
	_actionRefList.append((NEW GDMOActionRef(action)));
}

/****************************************************************/

void GDMOPackage::AddNotification(char *notification)
{
	_notificationRefList.append((NEW GDMONotificationRef(notification)));
}

/****************************************************************/
/* Update references */

Bool GDMOPackage::UpdateRefs(GDMOAttrList &attrList,
			     GDMOActionList &actionList,
			     GDMONotificationList &notificationList,
			     GDMOAttrGroupList &attrGroupList)
{
	if (!UpdateAttrRefs(attrList))
		return False;

	if (!UpdateAttrGroupRefs(attrList, attrGroupList))
		return False;

	if (!UpdateActionRefs(actionList))
		return False;

	if (!UpdateNotificationRefs(notificationList))
		return False;

	return True;

} /* UpdateRefs */

/****************************************************************/
/* Update attribute references */

Bool GDMOPackage::UpdateAttrRefs(GDMOAttrList &attrList)
{
ListIterator	attrIterator(&_attrRefList);
GDMOAttribute	*attrRef;

	while ((attrRef=(GDMOAttribute *)attrIterator.getNext()) !=
		(GDMOAttribute *) 0)
	{
	GDMOAttribute	*attr;
	ConvertEntry	*convEntry;
	char	*attrName = attrRef->GetName();

		if ((attr=attrList.FindAttr(attrName)) != (GDMOAttribute *) 0)
		{
			attrRef->UpdatePkgAttrRef(attr);
			continue;
		}

		if ((convEntry=GDMOAttribute::LookupAttributeDatabase(attrName))
		    != (ConvertEntry *) 0)
		{
			attrRef->UpdatePkgAttrRef(convEntry);
			continue;
		}

		GString	msg = GString("Cannot find attribute ") +
			      GString(attrRef->GetName()) +
			      GString(" in package ") + GString(GetName());

		printf("\n%s", (char *) msg);
		return False;
	}

	return True;

} /* UpdateAttrRefs */

/****************************************************************/
/* Update attribute group refs */

Bool GDMOPackage::UpdateAttrGroupRefs(GDMOAttrList &attrList,
				      GDMOAttrGroupList &attrGroupList)
{
ListIterator		attrIterator(&_attributeGroupList);
GDMOAttributeGroup	*attrGroupRef;

	while ((attrGroupRef=(GDMOAttributeGroup *)attrIterator.getNext()) !=
		(GDMOAttributeGroup *) 0)
	{
	GDMOAttributeGroup	*attrGroup;
	char	*name = attrGroupRef->GetName();

		if ((attrGroup=attrGroupList.FindAttrGroup(name))
			 == (GDMOAttributeGroup *) 0)
		{
		GString	msg = GString("Cannot find attribute group ") +
			      GString(name) +
			      GString(" in package ") + GString(GetName());

			printf("\n%s", (char *) msg);
			return False;
		}
/*
		printf("\nCheck Group %s GroupRef %s",
			attrGroup->GetName(), attrGroupRef->GetName());
*/
		// Check references in attribute group

		if (!attrGroup->UpdateRefs(attrList))
			return False;

		// Copy attribute group template fields

		static	char	*attrNames[] = {"Elements", "Fixed", "Oid"};
		int	lgth = sizeof(attrNames) / sizeof(char *);

		for (int count=0; count<lgth; count++)
		{
		char	*name = attrNames[count];

			Copy(attrGroupRef->FindSymbol(name),
			     attrGroup->FindSymbol(name));
		}

		// Check references in package attribute group reference

		if (!attrGroupRef->UpdateRefs(attrList))
			return False;
	}

	return True;

} /* UpdateAttrGroupRefs */

/****************************************************************/
/* Update action references */

Bool GDMOPackage::UpdateActionRefs(GDMOActionList &actionList)
{
ListIterator	actionIterator(&_actionRefList);
GDMOActionRef	*actionRef;

	while ((actionRef=(GDMOActionRef *)actionIterator.getNext()) !=
		(GDMOActionRef *) 0)
	{
	GDMOAction	*action;

		if ((action=actionList.FindAction(actionRef->GetName())) !=
			(GDMOAction *) 0)

			actionRef->UpdateRef(action);
		else
		{
		GString	msg = GString("Cannot find action ") +
			      GString(actionRef->GetName()) +
			      GString(" in package ") + GString(GetName());

			printf("\n%s", (char *) msg);
			return False;
		}
	}

	CopyActions(&_actionList);

	return True;

} /* UpdateActionRefs */

/****************************************************************/
/* Update notification references */

Bool GDMOPackage::UpdateNotificationRefs(GDMONotificationList &notificationList)
{
ListIterator		notifyIterator(&_notificationRefList);
GDMONotificationRef	*notifyRef;

	while ((notifyRef=(GDMONotificationRef *)notifyIterator.getNext()) !=
		(GDMONotificationRef *) 0)
	{
	GDMONotification	*notify;

		if ((notify=notificationList.FindNotification(notifyRef->GetName())) !=
			(GDMONotification *) 0)

			notifyRef->UpdateRef(notify);
		else
		{
		GString	msg = GString("Cannot find notification ") +
			      GString(notifyRef->GetName()) +
			      GString(" in package ") + GString(GetName());

			printf("\n%s", (char *) msg);
			return False;
		}
	}

	CopyNotifications(&_notificationList);

	return True;

} /* UpdateActionRefs */

/****************************************************************/

void GDMOPackage::CopyPackage(GDMOAttrList *attrList,
			      GDMOActionList *actionList,
			      GDMONotificationList *notificationList,
			      GDMOAttrGroupList *attrGroupList)
{
	CopyAttributes(attrList);
	CopyActions(actionList);
	CopyNotifications(notificationList);
	CopyAttributeGroups(attrGroupList);

} /* CopyPackage */

/****************************************************************/

void GDMOPackage::CopyAttributes(GDMOAttrList *attrList)
{
ListIterator	attrIterator(&_attrRefList);
GDMOAttribute	*attr;

	while ((attr=(GDMOAttribute *)attrIterator.getNext()) !=
		(GDMOAttribute *) 0)

		attrList->append(attr);

} /* CopyAttributes */

/****************************************************************/

void GDMOPackage::CopyAttributeGroups(GDMOAttrGroupList *attrGroupList)
{
ListIterator		groupIterator(&_attributeGroupList);
GDMOAttributeGroup	*group;

	while ((group=(GDMOAttributeGroup *)groupIterator.getNext()) !=
		(GDMOAttributeGroup *) 0)

		attrGroupList->append(group);

} /* CopyAttributeGroups */

/****************************************************************/

void GDMOPackage::CopyActions(GDMOActionList *actionList)
{
ListIterator	actionIterator(&_actionRefList);
GDMOActionRef	*actionRef;

	while ((actionRef=(GDMOActionRef *)actionIterator.getNext()) !=
		(GDMOActionRef *) 0)

		actionList->append(actionRef->GetAction());

} /* CopyActions */

/****************************************************************/

void GDMOPackage::CopyNotifications(GDMONotificationList *notificationList)
{
ListIterator		notificationIterator(&_notificationRefList);
GDMONotificationRef	*notificationRef;

	while ((notificationRef=(GDMONotificationRef *)
				notificationIterator.getNext()) !=
				(GDMONotificationRef *) 0)

		notificationList->append(notificationRef->GetNotification());

} /* CopyNotifications */

/****************************************************************/
/* Package List methods */
/****************************************************************/

// Find Package by name

GDMOPackage *GDMOPackageList::FindPackage(char *packageName)
{
GDMOPackage	*package;
ListIterator	pkgIterator(this);

	while ((package=(GDMOPackage *)pkgIterator.getNext()) !=
		(GDMOPackage *) 0)

		if (!strcmp(package->GetName(), packageName))
			break;

	return package;

} /* FindPackage */

/****************************************************************/
/* End of file */
/****************************************************************/

