/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	Node.cc

	Parse tree Node methods

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#ifdef	__MSDOS__
#include <io.h>
#else
#include <unistd.h>
#endif
#include <stdio.h>
#if defined(__GNUG__)
#include <std.h>
#endif
#include <stdlib.h>
#ifndef	__MSDOS__
#ifdef	__GNUG__
#include <string.h>
#else
#include <strings.h>
#endif
#endif
#include "DebugMem.h"
#include "GDMOGen.h"
#include "Symbol.h"
#include "SymExec.h"
#include "Node.h"

/****************************************************************/
/* Node methods */
/****************************************************************/

void	Node::PrintLFsSpaces(int nls, int spaces)
{
int	count;

	for (count=0; count<nls; count++)
		if (DebugFile) fprintf(DebugFile, "\n");

	for (count=0; count<spaces; count++)
		if (DebugFile) fprintf(DebugFile, " ");

} /* PrintLFsSpaces */

/****************************************************************/

void	Node::Error(int type)
{
	switch (type)
	{
	default:
		printf("\nUnknown Error");
		break;

	case ERRORILLEGALTYPE:
		printf("\nIllegal data type");
		break;

	case ERRORNOTSAMETYPE:
		printf("\nData types in expression must match");
		break;

	case ERRORSTRINGMINUS:
		printf("\nCannot subtract strings");
		break;
	}

	_interPreter->Stop();

} /* Error */

/****************************************************************/
/* Seq Node Methods */
/****************************************************************/

void	SeqNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"Sequence: ");
	_node1->PrintTree(1, offset+2);
	_node2->PrintTree(1, offset+2);

} /* PrintTree */

/****************************************************************/

void	SeqNode::Evaluate()
{
	_node2->Evaluate();
	_node1->Evaluate();

} /* Evaluate */

/****************************************************************/
/* Null Node */
/****************************************************************/

void	NullNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"Null (;) ");

} /* PrintTree */

/****************************************************************/
/* Var Node Methods */
/****************************************************************/

void	VarNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);

	if (DebugFile) fprintf(DebugFile, "Var node: %s, ", _sym->GetName());

	switch (_sym->GetSymType())
	{
	case StringTypeDef:
		{
		StringSymbol *ssym = _sym->GetStringSymbol();

			if (DebugFile)
				fprintf(DebugFile, "%s", ssym->GetValue());
			break;
		}

	case IntegerTypeDef:
		{
		IntegerSymbol *isym = _sym->GetIntegerSymbol();

			if (DebugFile)
				fprintf(DebugFile, "%d", isym->GetValue());
			break;
		}
	default:
		break;
	}

} /* PrintTree */

/****************************************************************/

void	VarNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);

} /* Evaluate */

/****************************************************************/

void	VarNode::Do()
{
	switch (_sym->GetSymType())
	{
	case StringTypeDef:
		{
		StringSymbol *ssym = _sym->GetStringSymbol();
		_interPreter->PushValue(ssym->GetValue());
		break;
		}

	case IntegerTypeDef:
		{
		IntegerSymbol *isym = _sym->GetIntegerSymbol();
		_interPreter->PushValue(isym->GetValue());
		break;
		}

	default:
		_interPreter->PushValue(_sym);
		break;
	}

} /* Do */

/****************************************************************/
/* Var Member Node Methods */
/****************************************************************/

void	VarMemberNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);

	if (DebugFile)
		fprintf(DebugFile, "Variable member %s.%s, ",
			_sym->GetName(), _memberName);

} /* PrintTree */

/****************************************************************/

Bool	VarMemberNode::Initialise()
{
	switch (_sym->GetSymType())
	{
	default:
		printf("\nVariable %s must be of type object",
			_sym->GetName());
		return False;

	case ObjectRefTypeDef:
	case ObjectTypeDef:
		break;
	}

	return True;

} /* Initialise */

/****************************************************************/

void	VarMemberNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);

} /* Evaluate */

/****************************************************************/

void	VarMemberNode::Do()
{
SymbolTable	*symTable;
ObjectSymbol	*object;
Symbol		*refSym;

	switch (_sym->GetSymType())
	{
	case ObjectRefTypeDef:
		{
		ObjectRefSymbol	*objRef = _sym->GetObjectRefSymbol();

		object = objRef->GetValue();

		if (object == (ObjectSymbol *) 0)
		{
			printf("\nVariable %s not initialised",
				_sym->GetName());
			_interPreter->Stop();
			return;
		}
		break;
		}

	case ObjectTypeDef:
		object = _sym->GetObjectSymbol();
		break;

	default:
		printf("\nIllegal Variable reference");
		_interPreter->Stop();
		break;
	}

	symTable = object->GetSymbolTable();

	refSym = symTable->Lookup(_memberName);

	if (refSym == (Symbol *) 0)
	{
		printf("\n%s not a member of %s",
			_memberName, _sym->GetName());

		_interPreter->Stop();
		return;
	}

	// Now push value on stack

	switch (refSym->GetSymType())
	{
	case StringTypeDef:
		{
		StringSymbol *ssym = refSym->GetStringSymbol();
		_interPreter->PushValue(ssym->GetValue());
		break;
		}

	case IntegerTypeDef:
		{
		IntegerSymbol *isym = refSym->GetIntegerSymbol();
		_interPreter->PushValue(isym->GetValue());
		break;
		}

	default:
		/* Nothing else to do except push symbol on stack */
		_interPreter->PushValue(refSym);
		break;
	}

} /* Do */

/****************************************************************/
/* AssignNode Methods */
/****************************************************************/

void	AssignNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile," %s Assign ", _sym->GetName());
	_node->PrintTree(0, 0);

} /* PrintTree */

/****************************************************************/

void	AssignNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_node->Evaluate();

} /* Evaluate */

/****************************************************************/

void	AssignNode::Do()
{
ValueStackVal	*val = val=_interPreter->PopValue();

	switch (val->_type)
	{
	case STK_CHAR:
		if (_sym->GetSymType() == StringTypeDef)
		{
		StringSymbol	*ssym = _sym->GetStringSymbol();

			ssym->Update(val->_data._strValue);
		}
		else
			Error(ERRORNOTSAMETYPE);
		break;

	case STK_INT:
		if (_sym->GetSymType() == IntegerTypeDef)
		{
		IntegerSymbol	*ssym = _sym->GetIntegerSymbol();

			ssym->Update(val->_data._intValue);
		}
		else
			Error(ERRORNOTSAMETYPE);
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* Do */

/****************************************************************/
/* IF node */
/****************************************************************/

void	IfNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"IF ");
	_exprNode->PrintTree(0, 0);
	if (DebugFile)
		fprintf(DebugFile," Statements: ");
	_statements->PrintTree(1, offset+2);

} /* PrintTree */

/****************************************************************/

void	IfNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_exprNode->Evaluate();

} /* Evaluate */

/****************************************************************/

void	IfNode::Do()
{
ValueStackVal	*val = val=_interPreter->PopValue();

	switch (val->_type)
	{
	case STK_CHAR:
		if (val->_data._strValue != (char *) 0)
			_statements->Evaluate();
		break;

	case STK_INT:
		if (val->_data._intValue)
			_statements->Evaluate();
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* Do	*/

/****************************************************************/
/* IFElse Node */
/****************************************************************/

void	IfElseNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"IFELSE ");
	_exprNode->PrintTree(0, 0);
	if (DebugFile)
		fprintf(DebugFile," Statements: ");
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"ELSE ");
	_elseNode->PrintTree(1, offset+2);

} /* PrintTree */

/****************************************************************/

void	IfElseNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_exprNode->Evaluate();

} /* Evaluate */

/****************************************************************/

void	IfElseNode::Do()
{
ValueStackVal	*val = val=_interPreter->PopValue();

	switch (val->_type)
	{
	case STK_CHAR:
		if (val->_data._strValue != (char *) 0)
			_statements->Evaluate();
		else
			_elseNode->Evaluate();
		break;

	case STK_INT:
		if (val->_data._intValue)
			_statements->Evaluate();
		else
			_elseNode->Evaluate();
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* Do	*/

/****************************************************************/
/* While Node */
/****************************************************************/

void	WhileNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"While ");
	_exprNode->PrintTree(0, 0);
	if (DebugFile)
		fprintf(DebugFile," Statements: ");
	_statements->PrintTree(1, offset+2);

} /* PrintTree */

/****************************************************************/

void	WhileNode::Evaluate()
{
        FnPtr bpf = &Node::Do;          /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_exprNode->Evaluate();

} /* Evaluate */

/****************************************************************/

void	WhileNode::Do()
{
ValueStackVal	*val = val=_interPreter->PopValue();

	switch (val->_type)
	{
	case STK_CHAR:
		if (val->_data._strValue != (char *) 0)
		{
			FnPtr bpf = &Node::Evaluate;    /* Workaround for g++ 2.5.x bug. */
			_interPreter->PushControl(this, bpf);
			_statements->Evaluate();
		}
		break;

	case STK_INT:
		if (val->_data._intValue)
		{
			FnPtr bpf = &Node::Evaluate;    /* Workaround for g++ 2.5.x bug. */
			_interPreter->PushControl(this, bpf);
			_statements->Evaluate();
		}
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* Do	*/

/****************************************************************/
/* For Node */
/****************************************************************/

void	ForNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"For var: %s", _varSym->GetName());
	if (DebugFile)
		fprintf(DebugFile," List: ");
	_exprNode->PrintTree(1, offset);
	if (DebugFile)
		fprintf(DebugFile," Statements: ");
	_statements->PrintTree(1, offset+2);

} /* PrintTree */

/****************************************************************/

Bool	ForNode::Initialise()
{
	if (!strcmp(_exprNode->Name(), "VARNODE") ||
	    !strcmp(_exprNode->Name(), "VARMEMBER"))
		return True;

	printf("\nError: FOREACH variable IN list variable");
	_interPreter->Stop();
	return False;

} /* Initialise */

/****************************************************************/

void	ForNode::Evaluate()
{
        FnPtr bpf = &Node::InitialiseList;  /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_exprNode->Evaluate();

} /* Evaluate */

/****************************************************************/

void	ForNode::InitialiseList()
{
ValueStackVal	*val = _interPreter->PopValue();
SymType		memberType, varType;
ListSymbol	*listSym;

	if (val->_type != STK_SYM)
	{
		printf("\nError in FOREACH");
		Error(ERRORILLEGALTYPE);
		return;
	}

	listSym = val->_data._sym->GetListSymbol();

	if (listSym->GetSymType() != ListTypeDef)
	{
		printf("\nFOREACH expects variable %s to be a list",
			listSym->GetName());
		_interPreter->Stop();
		return;
	}

	memberType = listSym->GetMemberType();
	varType = _varSym->GetSymType();

	if (varType != memberType &&
	   (!(varType == ObjectRefTypeDef && memberType == ObjectTypeDef)))
	{
		printf("\nFOREACH variable %s must match list %s member type",
			_varSym->GetName(), listSym->GetName());
		printf("\nTypes are %s %s", _varSym->GetSymTypeName(),
			listSym->GetMemberTypeName());
		_interPreter->Stop();
		return;
	}

	_iterator.setList(listSym);

	// Now execute the FOR loop

        FnPtr bpf = &Node::Do;            /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);

} /* InitialiseList */

/****************************************************************/

void	ForNode::Do()
{
Symbol	*sym = (Symbol *) _iterator.getNext();

	if (sym == (Symbol *) 0)
		return;

        FnPtr bpf = &Node::Do;               /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);

	// Update variable with contents of list element

	switch (_varSym->GetSymType())
	{
	case StringTypeDef:
		{
		StringSymbol *lsym = _varSym->GetStringSymbol();
		StringSymbol *rsym = sym->GetStringSymbol();

		lsym->Update(rsym->GetValue());
		break;
		}

	case IntegerTypeDef:
		{
		IntegerSymbol *lsym = _varSym->GetIntegerSymbol();
		IntegerSymbol *rsym = sym->GetIntegerSymbol();

		lsym->Update(rsym->GetValue());
		break;
		}

	case ObjectRefTypeDef:
		{
		ObjectRefSymbol	*lsym = _varSym->GetObjectRefSymbol();
		ObjectSymbol *rsym = sym->GetObjectSymbol();

		lsym->Update(rsym->GetValue());
		break;
		}

	default:
		printf("\nIllegal Variable reference");
		_interPreter->Stop();
		break;
	}


	_statements->Evaluate();

} /* Do	*/

/****************************************************************/
/* ExprListNode */
/****************************************************************/

void	ExprListNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile," Expression: ");
	_node2->PrintTree(0, 0);
	_node1->PrintTree(0, 0);

} /* PrintTree */

/****************************************************************/

int	ExprListNode::Count()
{
	 return(_node1->Count() + _node2->Count());

} /* Count */

/****************************************************************/

void	ExprListNode::Evaluate()
{
	_node1->Evaluate();
	_node2->Evaluate();

} /* Evaluate */

/****************************************************************/
/* BinaryNode */
/****************************************************************/

void	BinaryNode::PrintExpTree(int nls, int offset, char *name,
			      Node *node1, Node *node2)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile, " %s ", name);
	node2->PrintTree(0, 0);
	node1->PrintTree(0, 0);

} /* PrintExpTree */

/****************************************************************/

void	BinaryNode::Evaluate()
{
        FnPtr bpf = &Node::Do;              /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_node2->Evaluate();
	_node1->Evaluate();

} /* Evaluate */

/****************************************************************/

void	BinaryNode::Do()
{
ValueStackVal	*val2 = _interPreter->PopValue();
ValueStackVal	*val1 = _interPreter->PopValue();

	if (val1->_type == val2->_type)
	{
		switch (val1->_type)
		{
		case STK_INT:
			DoInt(val1->_data._intValue, val2->_data._intValue);
			break;

		case STK_CHAR:
			DoChar(val1->_data._strValue, val2->_data._strValue);
			break;

		default:
			Error(ERRORILLEGALTYPE);
			break;
		}
	}
	else
		Error(ERRORNOTSAMETYPE);

} /* Do	*/

/****************************************************************/
/* Plus Node */
/****************************************************************/

void	PlusNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 + val2);

} /* DoInt */

/****************************************************************/

void	PlusNode::DoChar(char *val1, char *val2)
{
GString	tmp1(val1);
GString	tmp2(val2);
GString	res = tmp1 + tmp2;
StringSymbol *ssym;

	ssym = (NEW StringSymbol(res.print(), ""));
	_interPreter->AddSymbol(ssym, False, False);
	_interPreter->PushValue(ssym->GetValue());

} /* DoChar */

/****************************************************************/
/* Minus Node */
/****************************************************************/

void	MinusNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 - val2);

} /* DoInt */

/****************************************************************/

void	MinusNode::DoChar(char *, char *)
{
	Error(ERRORSTRINGMINUS);

} /* DoChar */

/****************************************************************/
/* Multiply Node */
/****************************************************************/

void	MultiplyNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 * val2);

} /* DoInt */

/****************************************************************/

void	MultiplyNode::DoChar(char *, char *)
{
	Error(ERRORSTRINGMINUS);

} /* DoChar */

/****************************************************************/
/* Divide Node */
/****************************************************************/

void	DivideNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 / val2);

} /* DoInt */

/****************************************************************/

void	DivideNode::DoChar(char *, char *)
{
	Error(ERRORSTRINGMINUS);

} /* DoChar */

/****************************************************************/
/* GT Node */
/****************************************************************/

void	GTNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 > val2);

} /* DoInt */

/****************************************************************/

void	GTNode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res > 0);

} /* DoChar */

/****************************************************************/
/* GE Node */
/****************************************************************/

void	GENode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 >= val2);

} /* DoInt */

/****************************************************************/

void	GENode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res >= 0);

} /* DoChar */

/****************************************************************/
/* LT Node */
/****************************************************************/

void	LTNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 < val2);

} /* DoInt */

/****************************************************************/

void	LTNode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res < 0);

} /* DoChar */

/****************************************************************/
/* LE Node */
/****************************************************************/

void	LENode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 <= val2);

} /* DoInt */

/****************************************************************/

void	LENode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res <= 0);

} /* DoChar */

/****************************************************************/
/* EQ Node */
/****************************************************************/

void	EQNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 == val2);

} /* DoInt */

/****************************************************************/

void	EQNode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res == 0);

} /* DoChar */

/****************************************************************/
/* NEQ Node */
/****************************************************************/

void	NEQNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 != val2);

} /* DoInt */

/****************************************************************/

void	NEQNode::DoChar(char *val1, char *val2)
{
int	res = strcmp(val1, val2);

	_interPreter->PushValue(res != 0);

} /* DoChar */

/****************************************************************/
/* AND Node */
/****************************************************************/

void	ANDNode::Evaluate()
{
        FnPtr bpf = &Node::DoAnd;           /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_node1->Evaluate();

} /* ANDNode */

/****************************************************************/

void	ANDNode::DoAnd()
{
ValueStackVal	*val = _interPreter->PopValue();

	switch (val->_type)
	{
	case STK_INT:
		// If False no need to evaluate second node
		if (val->_data._intValue == 0)
			_interPreter->PushValue(0);
		else
			// Result of AND is result of node 2 evaluation
			_node2->Evaluate();
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* DoAnd */

/****************************************************************/
/* OR Node */
/****************************************************************/

void	ORNode::Evaluate()
{
        FnPtr bpf = &Node::DoOr;            /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_node1->Evaluate();

} /* ORNode */

/****************************************************************/

void	ORNode::DoOr()
{
ValueStackVal	*val = _interPreter->PopValue();

	switch (val->_type)
	{
	case STK_INT:
		// If True no need to evaluate second node
		if (val->_data._intValue != 0)
			_interPreter->PushValue(1);
		else
			// Result of OR is result of node 2 evaluation
			_node2->Evaluate();
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* DoOr */

/****************************************************************/
/* BITAND Node */
/****************************************************************/

void	BITANDNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 & val2);

} /* DoInt */

/****************************************************************/

void	BITANDNode::DoChar(char *, char *)
{
	Error(ERRORSTRINGMINUS);

} /* DoChar */

/****************************************************************/
/* BITOR Node */
/****************************************************************/

void	BITORNode::DoInt(int val1, int val2)
{
	_interPreter->PushValue(val1 | val2);

} /* DoInt */

/****************************************************************/

void	BITORNode::DoChar(char *, char *)
{
	Error(ERRORSTRINGMINUS);

} /* DoChar */

/****************************************************************/
/* NOT Node */
/****************************************************************/

void	NOTNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"NOT : %s");
       _node->PrintTree(0, 0);

} /* PrintTree */

/****************************************************************/

void	NOTNode::Evaluate()
{
        FnPtr bpf = &Node::Do;              /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);
	_node->Evaluate();

} /* NOTNode */

/****************************************************************/

void	NOTNode::Do()
{
ValueStackVal	*val = _interPreter->PopValue();
int	res;

	switch (val->_type)
	{
	case STK_INT:
		// If True no need to evaluate second node
		res = (val->_data._intValue == 0);
		_interPreter->PushValue(res);
		break;

	default:
		Error(ERRORILLEGALTYPE);
		break;
	}

} /* Do */

/****************************************************************/
/* BUILTIN Node */
/****************************************************************/

void	BUILTINNode::PrintTree(int nls, int offset)
{
	PrintLFsSpaces(nls, offset);
	if (DebugFile)
		fprintf(DebugFile,"BUILTIN: %s", _name);
	_node->PrintTree(0, 0);

} /* PrintTree */

/****************************************************************/

BUILTIN	BUILTINNode::_table[] =
	{
		{ListCountType,		"LISTCOUNT",	True,	1},
		{ExistFileType,		"EXISTFILE",	True,	1},
		{ExecuteType,		"EXECUTE",	False,	3},
		{PrintType,		"PRINT",	False,	0},
		{DisplayType,		"DISPLAY",	False,	0},
		{IncludeType,		"INCLUDE",	False,	1},
		{GetEnvType,		"GETENV",	True,	1},
	};

BUILTIN	*BUILTINNode::Lookup(char *name)
{
int	count;
int	lgth = sizeof(_table) / sizeof(BUILTIN);
BUILTIN	*entry;

	for (count=0; count<lgth; count++)
	{
		entry = &_table[count];
#ifdef	__MSDOS__
		if (!stricmp(name, entry->_name))
#else
		if (!strcasecmp(name, entry->_name))
#endif
			return entry;
	}

	return (BUILTIN *) 0;

} /* Lookup */

/****************************************************************/

void	BUILTINNode::Initialise()
{
BUILTIN	*entry;

	if ((entry=Lookup(_name)) == (BUILTIN *) 0)
	{
		printf("\nCannot find type %s", _name);
		exit(1);
	}
	_type = entry->_type;
	_argCount = _node->Count();  // BUG iN MSDOS !!!
	if (DebugFile)
		fprintf(DebugFile, "\nBUILTIN %s ArgCount %d",
			_table[_type]._name, _argCount);

} /* Initialise */

/****************************************************************/

void	BUILTINNode::Evaluate()
{
        FnPtr bpf = &Node::Do;              /* Workaround for g++ 2.5.x bug. */
        _interPreter->PushControl(this, bpf);

	// If variable length arguments push null

	if ((_table[_type]._argCount != 0)
		 && (_table[_type]._argCount != _argCount))
	{
		printf("\nBUILTIN %s expects %d arguments (not %d)",
			_table[_type]._name, _table[_type]._argCount,
			_argCount);
		_interPreter->Stop();
		return;
	}

	_node->Evaluate();

} /* BUILTINNode */

/****************************************************************/

void	BUILTINNode::Do()
{
	switch (_type)
	{
	case ListCountType:
		ListCount();
		break;

	case ExistFileType:
		ExistFile();
		break;

	case ExecuteType:
		Execute();
		break;

	case PrintType:
		Print(False);
		break;

	case DisplayType:
		Print(True);
		break;

	case IncludeType:
		Include();
		break;

	case GetEnvType:
		GetEnvSymbol();
		break;
	}

} /* Do	*/

/****************************************************************/

void	BUILTINNode::ListCount()
{
ValueStackVal *val = _interPreter->PopValue();

	if (val->_type == STK_SYM)
	{
	Symbol *sym = val->_data._sym;

		if (sym->GetSymType() != ListTypeDef)
		{
			printf("\nLISTCOUNT: Variable %s is not a list",
				sym->GetName());
			_interPreter->Stop();
		}
		else
		{
		ListSymbol	*lsym = sym->GetListSymbol();

		_interPreter->PushValue(lsym->getCount());
		}
	}
	else
	{
		printf("\nLISTCOUNT usage is: ListCount listvar");
		Error(ERRORILLEGALTYPE);
	}

} /* ListCount */

/****************************************************************/

void	BUILTINNode::ExistFile()
{
ValueStackVal *val = _interPreter->PopValue();

	if (val->_type == STK_CHAR)
	{
	char	*fName = val->_data._strValue;
	int	res = (access(fName, 0) == 0);

		_interPreter->PushValue(res);
	}
	else
	{
		printf("\nEXISTFILE usage is: ExistFile filename");
		Error(ERRORILLEGALTYPE);
	}

} /* ExistFile */

/****************************************************************/

void	BUILTINNode::Execute()
{
ValueStackVal	*val1 = _interPreter->PopValue();
ValueStackVal	*val2 = _interPreter->PopValue();
ValueStackVal	*val3 = _interPreter->PopValue();

	if (!((val1->_type == STK_CHAR) && (val2->_type == STK_CHAR)
		&& (val3->_type == STK_SYM)))
	{
		printf("\nEXECUTE usage is: Execute filename filename object");
		Error(ERRORILLEGALTYPE);
		return;
	}

	if (val3->_data._sym->GetSymType() != ObjectRefTypeDef)
	{
		printf("\nExecute Variable %s not an object",
			val3->_data._sym->GetName());
		Error(ERRORILLEGALTYPE);
		return;
	}
	ObjectRefSymbol	*objRef = val3->_data._sym->GetObjectRefSymbol();
	ObjectSymbol	*obj = objRef->GetValue();

	if (obj == (ObjectSymbol *) 0)
	{
		printf("\nObject %s uninitialised", objRef->GetName());
		Error(ERRORILLEGALTYPE);
		return;
	}
	InterPreter	*interPreter;

	// Hack if output file is "" use current output file

	if (!strcmp(val2->_data._strValue, ""))
		interPreter = (NEW InterPreter(val1->_data._strValue,
					       _interPreter->GetOutputFile(),
					       obj->GetSymbolTable()));
	else
		interPreter = (NEW InterPreter(val1->_data._strValue,
					       val2->_data._strValue,
					       obj->GetSymbolTable()));

	interPreter->Execute();

	DELETE interPreter;

} /* Execute */

/****************************************************************/

void	BUILTINNode::Print(Bool display)
{
ValueStackVal	*val;
int		count = _argCount;

	while (count-- && ((val=_interPreter->PopValue()) != (ValueStackVal *) 0))
	{
		switch (val->_type)
		{
		case STK_CHAR:
			if (display)
				_interPreter->Display(val->_data._strValue);
			else
				_interPreter->Print(val->_data._strValue);
			break;

		case STK_INT:
			{
			char	tmp[20];

			sprintf(tmp, "%d", val->_data._intValue);
			if (display)
				_interPreter->Display(tmp);
			else
				_interPreter->Print(tmp);
			break;
			}
		default:
			Error(ERRORILLEGALTYPE);
			break;
		}
	}

} /* Print */

/****************************************************************/

void	BUILTINNode::Include()
{
ValueStackVal *val = _interPreter->PopValue();

	if (val->_type == STK_CHAR)
	{
	char	*fName = val->_data._strValue;
	FILE	*fp;

		if ((fp=fopen(fName, "r")) != (FILE *) 0)
		{
		int	ch;

			while ((ch=fgetc(fp)) != EOF)
				_interPreter->Print(ch);

			fclose(fp);
		}
		else
		{
			printf("\nCannot Include %s", fName);
			_interPreter->Stop();
		}
	}
	else
	{
		printf("\nInclude usage is: Include filename");
		Error(ERRORILLEGALTYPE);
	}

} /* Include */

/****************************************************************/

void	BUILTINNode::GetEnvSymbol()
{
ValueStackVal *val = _interPreter->PopValue();

	if (val->_type == STK_CHAR)
	{
	char	*envName = getenv(val->_data._strValue);
	StringSymbol *ssym;

		if (envName == (char *) 0)
			ssym = (NEW StringSymbol("", ""));
		else
			ssym = (NEW StringSymbol(envName, ""));

		_interPreter->AddSymbol(ssym, False, False);
		_interPreter->PushValue(ssym->GetValue());
	}
	else
	{
		printf("\nLISTCOUNT usage is: GetEnvSymbol \"STRING\"");
		Error(ERRORILLEGALTYPE);
	}

} /* GetEnvSymbol */

/****************************************************************/
/* End of module */
/****************************************************************/



