/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	Node.h

	Parse Tree Node	classes

	Author:	J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#ifndef	_NODE_
#define	_NODE_


#ifndef _LIST_ITERATOR_
#include "ListIterator.h"
#endif


/****************************************************************/

#define	ERRORILLEGALTYPE	1
#define	ERRORNOTSAMETYPE	2
#define	ERRORSTRINGMINUS	3

/****************************************************************/
/* Forward references */

class	InterPreter;

extern	FILE *DebugFile;

/****************************************************************/

class	Node
{
protected:
	InterPreter	*_interPreter;

	void		PrintLFsSpaces(int, int);
	void		Error(int);
public:
	Node(InterPreter *interPreter) : _interPreter(interPreter)
	{}

	virtual	~Node()	{}

	virtual	void	PrintTree(int, int) = 0;
	virtual	void	Evaluate() = 0;
	virtual	char	*Name() = 0;

        virtual void    Do() { exit(0); }	/* Workaround g++ 2.5.x bug */
        virtual void    DoOr() { exit(0); }
        virtual void    DoAnd() {exit(0); }
        virtual void    InitialiseList() { exit(0); }


}; /* Node */

typedef	void	(Node::*FnPtr)();

/****************************************************************/
/* Sequence (of statements) Node */

class	SeqNode	: public Node
{
	Node	*_node1, *_node2;
public:
	SeqNode(Node *node1, Node *node2, InterPreter *interPreter) :
		_node1(node1), _node2(node2), Node(interPreter)
	{}

	~SeqNode()
	{
		DELETE _node1;
		DELETE _node2;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	char	*Name() {return	"SeqNode"; }

}; /*  SeqNode */

/****************************************************************/
/* Null Node */

class	NullNode : public Node
{
public:
	NullNode(InterPreter *interPreter) : Node(interPreter)
	{}

	void	PrintTree(int nls, int offset);
	void	Evaluate() {}
	char	*Name()	{return	"NullNode"; }

}; /*  NullNode	*/

/****************************************************************/
/* Assign Node */

class	AssignNode : public Node
{
	Symbol	*_sym;
	Node	*_node;
public:
	AssignNode(Symbol *sym, Node *node, InterPreter *interPreter) :
		_sym(sym), _node(node), Node(interPreter)
	{}

	~AssignNode()
	{
		DELETE _node;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();

	char	*Name() {return "AssignNode"; }

}; /*  AssignNode */

/****************************************************************/
/* For Node */

class	ForNode	: public Node
{
	Symbol		*_varSym;
	Node		*_exprNode;
	Node		*_statements;
	ListIterator	_iterator;
public:
	ForNode(Symbol *varSym, Node *exprNode,
		Node *statements, InterPreter *interPreter) :
		_varSym(varSym), _exprNode(exprNode),
		_statements(statements), Node(interPreter)
	{}

	Bool	Initialise();
	void	InitialiseList();

	~ForNode()
	{
		DELETE _exprNode;
		DELETE _statements;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"ForNode"; }

}; /*  ForNode */

/****************************************************************/
/* If Node */

class	IfNode : public	Node
{
	Node	*_exprNode;
	Node	*_statements;
public:
	IfNode(Node *exprNode, Node *statements, InterPreter *interPreter) :
		_exprNode(exprNode), _statements(statements),
		Node(interPreter)
	{}

	~IfNode()
	{
		DELETE _exprNode;
		DELETE _statements;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"IfNode"; }

}; /*  IfNode */

/****************************************************************/
/* IfElse Node */

class	IfElseNode : public Node
{
	Node	*_exprNode;
	Node	*_statements, *_elseNode;
public:
	IfElseNode(Node	*exprNode, Node	*statements,
		   Node	*elseNode, InterPreter *interPreter) :
			_exprNode(exprNode), _statements(statements),
			_elseNode(elseNode), Node(interPreter)
	{}

	~IfElseNode()
	{
		DELETE _exprNode;
		DELETE _statements;
		DELETE _elseNode;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"IfElseNode"; }

}; /* IfElseNode */

/****************************************************************/
/* While Node */

class	WhileNode : public Node
{
	Node	*_exprNode;
	Node	*_statements;
public:
	WhileNode(Node *exprNode, Node *statements, InterPreter	*interPreter) :
		_exprNode(exprNode), _statements(statements),
		Node(interPreter)
	{}

	~WhileNode()
	{
		DELETE _exprNode;
		DELETE _statements;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"WhileNode"; }

}; /*  WhileNode */

/****************************************************************/
/* Expression Node */

class	ExpNode	: public Node
{
public:
	ExpNode(InterPreter *interPreter) : Node(interPreter)
	{}

	virtual	~ExpNode() {}

	virtual	void	PrintTree(int, int) = 0;
	virtual	void	Evaluate() = 0;
	virtual	char	*Name()	= 0;
	virtual	int	Count()	= 0;

}; /*  ExpNode */

/****************************************************************/
/* Expression List Node	*/

class	ExprListNode : public ExpNode
{
	ExpNode	   *_node1, *_node2;
public:
	ExprListNode(ExpNode *node1, ExpNode *node2,
		     InterPreter *interPreter)
	: _node1(node1), _node2(node2),	ExpNode(interPreter)
	{}

	~ExprListNode()
	{
		DELETE _node1; DELETE _node2;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	char	*Name()	{return	"ExprListNode";	}

	int	Count();

}; /*  ExprListNode */

/****************************************************************/
/* Unary Expression */

class	UnaryNode : public ExpNode
{
public:
	UnaryNode(InterPreter *interPreter) : ExpNode(interPreter)
	{}

	virtual	~UnaryNode() {}

	virtual	void	PrintTree(int, int) = 0;
	virtual	void	Evaluate() = 0;
	virtual	char	*Name()	= 0;
	int	Count()	{return	1;}

}; /*  UnaryNode */

/****************************************************************/
/* Variable Node */

class	VarNode	: public UnaryNode
{
	Symbol	*_sym;
public:
	VarNode(Symbol *sym, InterPreter *interPreter) :
		_sym(sym), UnaryNode(interPreter)
	{}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"VARNODE";}

}; /*  VarNode */

/****************************************************************/
/* Variable member Node	*/

class	VarMemberNode :	public UnaryNode
{
	Symbol	*_sym;
	char	*_memberName;
public:
	VarMemberNode(Symbol *sym, char *memberName,
		      InterPreter *interPreter)
	: _sym(sym), _memberName(memberName), UnaryNode(interPreter)
	{}

	void	PrintTree(int nls, int offset);
	Bool	Initialise();
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"VARMEMBER"; }

}; /*  VarMemberNode */

/****************************************************************/
/* Binary node */

class	BinaryNode : public ExpNode
{
protected:
	Node	*_node1, *_node2;
public:
	BinaryNode(Node	*node1,	Node *node2, InterPreter *interPreter)
	: _node1(node1), _node2(node2),	ExpNode(interPreter)
	{}

	void	PrintExpTree(int nls, int offset, char *name,
			     Node *node1, Node *node2);

	~BinaryNode()
	{
		DELETE _node1; DELETE _node2;
	}
	virtual	void	PrintTree(int, int) = 0;
	virtual	void	Evaluate();
	void	Do();
	virtual	void	DoInt(int, int)	= 0;
	virtual	void	DoChar(char *, char *) = 0;
	virtual	char	*Name()	= 0;
	int	Count()	{return	1;}

}; /*  BinaryNode */

/****************************************************************/
/* Plus	node */

class	PlusNode : public BinaryNode
{
public:
	PlusNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "PLUS", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"PlusNode"; }

}; /*  PlusNode	*/

/****************************************************************/
/* Minus node */

class	MinusNode : public BinaryNode
{
public:
	MinusNode(Node *node1, Node *node2, InterPreter	*interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "MINUS", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"MinusNode"; }

}; /*  MinusNode */

/****************************************************************/
/* Multiply node */

class	MultiplyNode : public BinaryNode
{
public:
	MultiplyNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "MULTIPLY", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"MultiplyNode";	}

}; /*  MultiplyNode */

/****************************************************************/
/* Divide node */

class	DivideNode : public BinaryNode
{
public:
	DivideNode(Node	*node1,	Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "DIVIDE", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"DivideNode"; }

}; /*  DivideNode */

/****************************************************************/
/* GT node */

class	GTNode : public	BinaryNode
{
public:
	GTNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "GT",	_node1,	_node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"GTNode"; }

}; /*  GTNode */

/****************************************************************/
/* GE node */

class	GENode : public	BinaryNode
{
public:
	GENode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "GE",	_node1,	_node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"GENode"; }

}; /*  GENode */

/****************************************************************/
/* LT node */

class	LTNode : public	BinaryNode
{
public:
	LTNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "LT",	_node1,	_node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"LTNode"; }

}; /*  LTNode */

/****************************************************************/
/* LE node */

class	LENode : public	BinaryNode
{
public:
	LENode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "LE",	_node1,	_node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"LENode"; }

}; /*  LENode */

/****************************************************************/
/* EQ node */

class	EQNode : public	BinaryNode
{
public:
	EQNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "EQ",	_node1,	_node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"EQNode"; }

}; /*  EQNode */

/****************************************************************/
/* NEQ node */

class	NEQNode	: public BinaryNode
{
public:
	NEQNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "NEQ", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"NEQNode"; }

}; /*  NEQNode */

/****************************************************************/
/* And node */

class	ANDNode	: public BinaryNode
{
public:
	ANDNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "AND", _node1, _node2);
	}
	void	Evaluate();
	void	DoAnd();
	void	DoInt(int, int)	{}
	void	DoChar(char *, char *) {}
	char	*Name()	{return	"ANDNode"; }

}; /*  ANDNode */

/****************************************************************/
/* OR node */

class	ORNode : public	BinaryNode
{
public:
	ORNode(Node *node1, Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "OR",	_node1,	_node2);
	}
	void	Evaluate();
	void	DoOr();
	void	DoInt(int, int)	{}
	void	DoChar(char *, char *) {}
	char	*Name()	{return	"ORNode"; }

}; /*  ORNode */

/****************************************************************/
/* Bit And node	*/

class	BITANDNode : public BinaryNode
{
public:
	BITANDNode(Node	*node1,	Node *node2, InterPreter *interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "BITAND", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"BITANDNode"; }

}; /*  BITANDNode */

/****************************************************************/
/* Bit OR node */

class	BITORNode : public BinaryNode
{
public:
	BITORNode(Node *node1, Node *node2, InterPreter	*interPreter)
	: BinaryNode(node1, node2, interPreter)
	{}

	void	PrintTree(int nls, int offset)
	{
		PrintExpTree(nls, offset, "BITOR", _node1, _node2);
	}
	void	DoInt(int, int);
	void	DoChar(char *, char *);
	char	*Name()	{return	"BITORNode"; }

}; /*  BITORNode */

/****************************************************************/
/* NOT node */

class	NOTNode	: public UnaryNode
{
	Node	*_node;
public:
	NOTNode(Node *node, InterPreter	*interPreter) :
		_node(node), UnaryNode(interPreter)
	{}

	~NOTNode()
	{
		DELETE _node;
	}
	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"NOTNode"; }

}; /*  NOTNode */

/****************************************************************/
/* BUILTIN NODES */
/****************************************************************/

typedef	enum	{ListCountType = 0, ExistFileType = 1, ExecuteType = 2,
		 PrintType = 3,	DisplayType = 4, IncludeType = 5,
		 GetEnvType = 6} BuiltinTypes;

typedef	struct
	{
	BuiltinTypes	_type;
	char		*_name;
	int		_fn;
	int		_argCount;
	}
	BUILTIN;

class	BUILTINNode : public UnaryNode
{
	void		Initialise();
	ExpNode		*_node;
	char		*_name;
	int		_argCount;
	BuiltinTypes	_type;
	void		ListCount();
	void		ExistFile();
	void		Execute();
	void		Print(Bool);
	void		Include();
	void		GetEnvSymbol();
	static		BUILTIN	_table[];
public:
	BUILTINNode(char *name,	ExpNode	*node, InterPreter *interPreter)
	: _name(name), _node(node), UnaryNode(interPreter)
	{
		Initialise();
	}

	~BUILTINNode()
	{
		DELETE _node;
	}

	void	PrintTree(int nls, int offset);
	void	Evaluate();
	void	Do();
	char	*Name()	{return	"BUILTINNode"; }
	static	BUILTIN	*Lookup(char *);

}; /*  BUILTINNode */

/****************************************************************/

#endif

/****************************************************************/
/* End of module */
/****************************************************************/





