/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	SymExec.cc

	Code generator Interpreter methods

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "DebugMem.h"
#include "GDMOGen.h"
#include "Symbol.h"
#include "SymExec.h"

/****************************************************************/

extern	int	ExecParseErrors(), ExecLine;

/****************************************************************/
/* Initialise */
/****************************************************************/

void	InterPreter::LexInitialise(char *inFileName)
{
	_yyinSave = yyin;	// Save	lex stream

	if ((_inFile=fopen(inFileName, "r")) == (FILE *) 0)
	{
		printf("\nCannot open %s", inFileName);
		exit(1);
	}

	yyin = _inFile;
	ExecLine = 1;

} /* LexInitialise */

/****************************************************************/
/* Initialise */
/****************************************************************/

void	InterPreter::InitialiseOutputFile(char *outFileName)
{
	if ((_outFile=fopen(outFileName, "w")) == (FILE *) 0)
	{
		printf("\nCannot open %s", outFileName);
		exit(1);
	}

} /* InitialiseOutputFile */

/****************************************************************/
/* Execute */

Bool	InterPreter::Execute()
{
int	ch;
Bool	error = False;

	while ((ch=fgetc(_inFile)) != EOF && !error)
	{
		switch (ch)
		{
		case  '$':
			if ((ch=fgetc(_inFile)) != '$')
			{
				/* Execute script until next $ */

				ungetc(ch, _inFile);
				if (Compile())
				{
					if (!Run())
						error = True;
				}
				else
					error = True;

				if (_firstNode)
					DELETE _firstNode;
				break;
			}
			else
				Print(ch);
			break;

		case '\n':
			ExecLine++;

		default:
			Print(ch);
			break;
		}
	}

	fclose(_inFile);
	
	// horrible hack - if outputFilename is 0 do not close the file 
	// we are appending to it

	if (_outFileName != (char *) 0)
		fclose(_outFile);

	return ((error == False) ? True : False);

} /* Execute */

/****************************************************************/
/* Compile */

Bool	InterPreter::Compile()
{
	_firstNode = (Node *) 0;

	_localSymbols.Clear();

	Parse();
	if (!ExecParseErrors())
		return True;
	else
		printf("\nError in script %s", _inFileName);

	return False;

} /* Compile */

/****************************************************************/

Bool	InterPreter::Run()
{
ControlStackVal	*stk;

	if (_firstNode == (Node *) 0)
		return True;

#if	DEBUG
	_firstNode->PrintTree(1, 0);
#endif
	_runError = False;
	_valStkPtr = _ctrlStkPtr = 0;
	_firstNode->Evaluate();

	while (((stk=PopControl()) != (ControlStackVal *) 0) && !_runError)
	{
	Node	*node = stk->_node;
	FnPtr	fnPtr = stk->_fnPtr;

#if	DEBUG
	if (DebugFile)
		fprintf(DebugFile, "\nPop Control %s %ld %ld",
			node->Name() ? node->Name() : "", node, fnPtr);
#endif
		(node->*fnPtr)();
	}

	if (_runError)
		return False;

	return True;

} /* Run */

/****************************************************************/

void	InterPreter::PushControl(Node *node, FnPtr fnPtr)
{
	if (_ctrlStkPtr < MAXSTACK)
	{
		_controlStack[_ctrlStkPtr]._node = node;
		_controlStack[_ctrlStkPtr++]._fnPtr = fnPtr;
	}
	else
	{
		printf("\nOut of (control) stack space");
		Stop();
	}
#if	DEBUG
	if (DebugFile)
		fprintf(DebugFile, "\nPush Control %s %ld %ld",
			node->Name() ? node->Name() : "", node, fnPtr);
#endif

} /* PushControl */

/****************************************************************/

ControlStackVal	*InterPreter::PopControl()
{
ControlStackVal	*res = (ControlStackVal *) 0;

	if (_ctrlStkPtr > 0)
		res = &(_controlStack[--_ctrlStkPtr]);

	return res;

} /* PopControl */

/****************************************************************/

void	InterPreter::PushValue(int intValue)
{
	if (_valStkPtr < MAXSTACK)
	{
		_valueStack[_valStkPtr]._type = STK_INT;
		_valueStack[_valStkPtr++]._data._intValue = intValue;
	}
	else
	{
		printf("\nOut of (value) stack space");
		Stop();
	}

#if	DEBUG
	if (DebugFile)
		fprintf(DebugFile, "\nPush IntValue %d", intValue);
#endif

} /* PushValue */

/****************************************************************/

void	InterPreter::PushValue(char *text)
{
	if (_valStkPtr < MAXSTACK)
	{
		_valueStack[_valStkPtr]._type = STK_CHAR;
		_valueStack[_valStkPtr++]._data._strValue = text;
	}
	else
	{
		printf("\nOut of (value) stack space");
		Stop();
	}

#if	DEBUG
	if (text)
	{
		if (DebugFile)
			fprintf(DebugFile, "\nPush CharValue %s", text);
	}
	else
	{
		if (DebugFile)
			fprintf(DebugFile, "\nPush Null CharValue");
	}
#endif

} /* PushValue */

/****************************************************************/

void	InterPreter::PushValue(Symbol *sym)
{
	if (_valStkPtr < MAXSTACK)
	{
		_valueStack[_valStkPtr]._type = STK_SYM;
		_valueStack[_valStkPtr++]._data._sym= sym;
	}
	else
	{
		printf("\nOut of (value) stack space");
		Stop();
	}

#if	DEBUG
	if (sym)
	{
		if (DebugFile)
			fprintf(DebugFile, "\nPush Sym %s", sym->GetName());
	}
#endif

} /* PushValue */

/****************************************************************/

ValueStackVal	*InterPreter::PopValue()
{
ValueStackVal	*res = (ValueStackVal *) 0;

	if (_valStkPtr > 0)
	{
		res = &(_valueStack[--_valStkPtr]);
		switch (res->_type)
		{
		case STK_CHAR:
			{
			char *text = res->_data._strValue;
			if (text == (char *) 0)
			{
				if (DebugFile)
					fprintf(DebugFile, "\nPop NullChar");
			}
			else
			{
				if (DebugFile)
					fprintf(DebugFile, "\nPop Char %s",
						text);
			}
			break;
			}

		case STK_INT:
			if (DebugFile)
				fprintf(DebugFile, "\nPop Int %d",
					res->_data._intValue);
			break;

		case STK_SYM:
			if (DebugFile)
				fprintf(DebugFile, "\nPop Sym %s",
					res->_data._sym->GetName());
			break;
		}
	}
	else
	{
		printf("\nIllegal value stack operation");
		Stop();
	}
	return res;

} /* PopValue */

/****************************************************************/
/* Output */

void	InterPreter::Output(FILE *file, int ch)
{
	fputc(ch, file);

} /* Output */

/****************************************************************/
/* Output */

void	InterPreter::Output(FILE *file, char *text)
{
static	struct
	{
	char	ch;
	char	trans;
	}
	convert[] =	{
			{'n', '\n'},
			{'t', '\t'},
			{'b', '\b'},
			{'r', '\r'},
			{'f', '\f'}
			};

	while (*text)
		switch (*text)
		{
		case '\\':
			{
			int count;
			char ch = *(++text);

			// Translate if necessary

			for (count=0; count<sizeof(convert); count++)
				if (ch == convert[count].ch)
				{
					fputc(convert[count].trans, file);
					break;
				}

			if (count == sizeof(convert))
				fputc(ch, file);
			text++;
			break;
			}
		default:
			fputc(*text++, file);
			break;
		}

} /* Output */

/****************************************************************/
/* Lookup */

Symbol	*InterPreter::Lookup(char *name)
{
Symbol	*sym;

	if ((sym=_symTable->Lookup(name)) != (Symbol *) 0)
		return sym;

	if ((sym=_localSymbols.Lookup(name)) != (Symbol *) 0)
		return sym;

	if ((sym=_globalSymbols.Lookup(name)) != (Symbol *) 0)
		return sym;

	if ((sym=MetaClass::LookupGlobal(name)) != (Symbol *) 0)
		return sym;

	printf("\nCannot find variable %s", name);

	return (Symbol *) 0;

} /* Lookup */

/****************************************************************/
/* Add symbol */

Bool	InterPreter::AddSymbol(Symbol *newSym, Bool uniqueTest,
			       Bool global)
{
	// Check if symbol is in C++ object's symbol table

	if (uniqueTest)
	{
	char	*name = newSym->GetName();

		if (_symTable->Lookup(name))
		{
			printf("\nVariable %s already declared", name);
			return False;
		}
	}

	// Check global symbol table

	if (global)
		return(_globalSymbols.Add(newSym, uniqueTest));

	// Check Local symbol table

	return(_localSymbols.Add(newSym, uniqueTest));

} /* AddSymbol */

/****************************************************************/
/* End of module */
/****************************************************************/

