/*
 * Copyright (c) 1992 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London,
 * London, England. Neither the name of the University, nor its
 * registered trademark, UCL, may be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/****************************************************************/
/*
	Symbol.cc

	Symbol Table methods

	Author: J.Cowan		(j.cowan@.cs.ucl.ac.uk)
*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#ifndef	__MSDOS__
#ifdef	__GNUG__
#include <string.h>
#else
#include <strings.h>
#endif
#endif

#include "DebugMem.h"
#include "Symbol.h"
#include "ListIterator.h"

/****************************************************************/
/* Symbol Table Class Methods */
/****************************************************************/
/* Lookup Symbol Table Entry */

Symbol *SymbolTable::Lookup(char *name)
{
ListIterator	iterator(&_symList);
Symbol	*sym = (Symbol *) 0;

	while ((sym=(Symbol *)iterator.getNext()) != (Symbol *) 0)
#ifdef	__MSDOS__
		if (!stricmp(sym->GetName(), name))
#else
		if (!strcasecmp(sym->GetName(), name))
#endif
			break;

	return sym;

} /* Lookup */

/****************************************************************/
/* Dump Symbol Table */

void	SymbolTable::DumpContents(int indent)
{
ListIterator	iterator(&_symList);
Symbol	*sym;

	while ((sym=(Symbol *)iterator.getNext()) != (Symbol *)0)

		DumpSymbolContents(sym, indent);

} /* DumpContents */

/****************************************************************/
/* Dump Symbol Table Entry Contents */

void	SymbolTable::DumpSymbolContents(Symbol *sym, int indent)
{
	printf("\n");
	for (int count=0; count<indent; count++)
		printf(" ");
/*
	printf("Symbol: %s. Type: %s.",
		sym->GetName(), sym->GetSymTypeName());
*/
	printf("Symbol: %s. ", sym->GetName());

	switch (sym->GetSymType())
	{
	default:
		break;

	case StringTypeDef:
		{
		StringSymbol *stringSymbol = sym->GetStringSymbol();

		printf(" Contents: %s.", stringSymbol->print());
		}
		break;

	case IntegerTypeDef:
		{
		IntegerSymbol *intSymbol = sym->GetIntegerSymbol();

		printf(" Contents: %d.", intSymbol->GetValue());
		}
		break;

	case ListTypeDef:
		{
		ListSymbol 	*list = sym->GetListSymbol();
		ListIterator	listIterator(list);
		Symbol	*listElement;

		printf(" Member Type: %s. List Contents: ",
			list->GetMemberTypeName());

		while ((listElement=(Symbol *)listIterator.getNext()) !=
				(Symbol *) 0)
			DumpSymbolContents(listElement, indent+1);
		break;
		}

	case ObjectTypeDef:
		ObjectSymbol	*object = sym->GetObjectSymbol();
		SymbolTable	*symTable = object->GetSymbolTable();

		printf(" Object Members: ");
		symTable->DumpContents(indent+2);
		break;
	}

} /* DumpSymbolContents */

/****************************************************************/
/* Dump	Symbol Table Class Members */

void	SymbolTable::DumpMembers(int indent)
{
ListIterator	iterator(&_symList);
Symbol	*sym;

	while ((sym=(Symbol *)iterator.getNext()) != (Symbol *)0)
	{
		printf("\n");
		for (int count=0; count<indent; count++)
			printf(" ");

		printf("Symbol: %s. Type: %s.",
			sym->GetName(), sym->GetSymTypeName());

		if (sym->GetSymType() == ListTypeDef)
		{
		ListSymbol	*list = sym->GetListSymbol();

			printf(" Member Type: %s", list->GetMemberTypeName());
		}
	}

} /* DumpMembers */

/****************************************************************/
/* Class Dictionary Methods */
/****************************************************************/

void	ClassEntry::PrintClassEntry()
{
ListIterator	listIterator(&_memberList);
ClassMember	*classMember;

	printf("\nCLASS %s", (char *)_name);

	while ((classMember=(ClassMember *)listIterator.getNext()) !=
				(ClassMember *) 0)
	{
		printf("\n\t%s\t", classMember->GetName());
/*
		if (classMember->GetExternal() == True)
			printf("EXTERN\t");
*/
		switch (classMember->GetType())
		{
		case StringTypeDef:
			printf("STRING");
			break;

		case IntegerTypeDef:
			printf("INTEGER");
			break;

		case ObjectTypeDef:
		case ObjectRefTypeDef:
			printf("OBJECT");
			break;

		case StringListTypeDef:
			printf("LIST OF STRING");
			break;

		case ListTypeDef:
			printf("LIST OF %s", classMember->GetListMember());
			break;
		}
	}

//	printf("\nEND\n\n");

} /* PrintClassEntry */

/****************************************************************/

ClassMember	*ClassEntry::LookupClassMember(char *name)
{
ListIterator	listIterator(&_memberList);
ClassMember	*classMember;

	while ((classMember=(ClassMember *)listIterator.getNext()) !=
				(ClassMember *) 0)
		if (!strcmp(name, classMember->GetName()))
			return classMember;

	return (ClassMember *) 0;

} /* LookupClassMember */

/****************************************************************/
/* statics */

ClassList ClassDictionary::_classList;

ClassEntry *ClassDictionary::_classEntry;

ClassMember *ClassDictionary::_classMember;

/****************************************************************/

void	ClassDictionary::AddClass(char *name)
{
	if (LookupClassEntry(name) != (ClassEntry *) 0)
	{
		printf("\nDuplicate class %s", name);
		exit(1);
	}

	_classEntry = NEW ClassEntry(name);
	_classList.append(_classEntry);

} /* AddClass */

/****************************************************************/

void	ClassDictionary::AddClassMember(char *name)
{
	if (_classEntry != (ClassEntry *) 0)
		_classMember = _classEntry->AddClassMember(name);
	else
	{
		printf("\nAdding class member %s to null class",
			name);
		exit(1);
	}

} /* AddClassMember */

/****************************************************************/

void	ClassDictionary::ValidClassMember()
{
	if (_classMember == (ClassMember *) 0)
	{
		printf("\nValidClassMember: No Class member");
		exit(1);
	}

} /* ValidClassMember */

/****************************************************************/

void	ClassDictionary::SetClassMemberExternal(Bool external) 
{
	ValidClassMember();
	 _classMember->SetClassMemberExternal(external);

} /* SetClassMemberExternal */

/****************************************************************/

void	ClassDictionary::SetClassMemberType(SymType type)
{
	ValidClassMember();
	 _classMember->SetClassMemberType(type);

} /* SetClassMemberType */

/****************************************************************/

void	ClassDictionary::SetClassListMemberType(char *name)
{
	ValidClassMember();
	 _classMember->SetClassListMemberType(name);

} /* SetClassListMemberType */

/****************************************************************/

ClassEntry	*ClassDictionary::LookupClassEntry(char *name)
{
ListIterator	listIterator(&_classList);
ClassEntry	*classEntry;

	while ((classEntry=(ClassEntry *)listIterator.getNext()) !=
				(ClassEntry *) 0)
		if (!strcmp(name, classEntry->GetName()))
			return classEntry;

	return (ClassEntry *) 0;

} /* LookupClassEntry */

/****************************************************************/

void	ClassDictionary::PrintClassDictionary()
{
ListIterator	listIterator(&_classList);
ClassEntry	*classEntry;

	while ((classEntry=(ClassEntry *)listIterator.getNext()) !=
				(ClassEntry *) 0)
		classEntry->PrintClassEntry();

} /* PrintClassDictionary */

/****************************************************************/
/* MetaClass operations */
/****************************************************************/

SymbolTable	MetaClass::_globalObjectList(False);

/****************************************************************/

void	MetaClass::CreateClassMembers()
{
ListIterator	listIterator(_class->GetMemberList());
ClassMember	*classMember;

	while ((classMember=(ClassMember *)listIterator.getNext()) !=
				(ClassMember *) 0)

	if (classMember->GetExternal() != True)
	{
	Symbol	*symbol = (Symbol *) 0;
	char		*name = classMember->GetName();

		switch (classMember->GetType())
		{
		case StringTypeDef:
			symbol = (NEW StringSymbol("", name));
			break;

		case IntegerTypeDef:
			symbol = (NEW IntegerSymbol(0, name));
			break;

		case ObjectRefTypeDef:
			symbol = (NEW ObjectRefSymbol(name));
			break;

	      	case StringListTypeDef:
			symbol = (NEW StringSymbolList(True, name));
			break;

	      	case IntegerListTypeDef:
			symbol = (NEW IntegerSymbolList(True, name));
			break;

	      	case ObjectRefListType:
			symbol = (NEW ObjectRefSymbolList(True, name));
			break;

		default:
			printf("\nClass %s: Member: %s not valid type",
				_class->GetName(), name);
			exit(1);
		}

		if (symbol != (Symbol *) 0)
			_symTable.Add(symbol, True);
	}

} /* CreateClassMembers */

/****************************************************************/

void	MetaClass::DeleteClassMembers()
{
ListIterator	listIterator(_class->GetMemberList());
ClassMember	*classMember;
SymbolList	deleteList(True);

	// Collect symbol list to be deleted

	while ((classMember=(ClassMember *)listIterator.getNext()) !=
				(ClassMember *) 0)
	{
	Symbol	*symbol = FindSymbol(classMember->GetName());

		if (classMember->GetExternal() == False)
			deleteList.append(symbol);
	}

	// what is appended to deleteList will be deleted

}  /* DeleteClassMembers */

/****************************************************************/

void	MetaClass::CheckClassMembers()
{
ListIterator	listIterator(_class->GetMemberList());
ClassMember	*classMember;

	while ((classMember=(ClassMember *)listIterator.getNext()) !=
				(ClassMember *) 0)
	{
	char	*name = classMember->GetName();

		if (SymLookup(name) == (Symbol *) 0)
		{
		printf("\nClass: %s Member: %s not initialised",
			_class->GetName(), name);
		exit(1);
		}
	}

} /* CheckClassMembers */

/****************************************************************/

void	MetaClass::AddSymbol(Symbol *symbol)
{
char		*symName = symbol->GetName();
ClassMember	*classMember = _class->LookupClassMember(symName);

	if (classMember == (ClassMember *) 0)
	{
		printf("\n%s not a member of class %s",
			symName, _class->GetName());
		exit(1);
	}

	if (classMember->GetExternal() != True)
	{
		printf("\n%s must be defined as EXTERN", symName);
		exit(1);
	}

	_symTable.Add(symbol, True);

} /* AddSymbol */

/****************************************************************/
/* Update an attribute */

void	MetaClass::UpdateSymbol(char *name, char *value)
{
Symbol	*sym = FindSymbol(name);

	switch (sym->GetSymType())
	{
	default:
		printf("\nMetaClass::Update: Cannot update symbol %s", name);
		exit(1);
		break;

	case StringTypeDef:
		{
		StringSymbol	*stringSymbol = sym->GetStringSymbol();

		stringSymbol->Update(value);
		}
		break;

	case IntegerTypeDef:
		{
		IntegerSymbol	 *intSymbol = sym->GetIntegerSymbol();

		intSymbol->Update(atoi(value));
		}
		break;
	}

} /* UpdateSymbol */

/****************************************************************/
/* Update an attribute */

void	MetaClass::UpdateSymbol(char *name, int value, UpdateMode mode)
{
Symbol	*sym = FindSymbol(name);

	if (sym->GetSymType() != IntegerTypeDef)
	{
		printf("\nMetaClass::Update: Cannot update integer symbol %s",
			 name);
		exit(1);
	}

	IntegerSymbol	 *intSymbol = sym->GetIntegerSymbol();

	switch (mode)
	{
	case Assign:
		intSymbol->Update(value);
		break;

	case Or:
		intSymbol->Update(intSymbol->GetValue() | value);
		break;

	case And:
		intSymbol->Update(intSymbol->GetValue() & value);
		break;
	}

} /* UpdateSymbol */

/****************************************************************/
/* Update an attribute list */

void	MetaClass::UpdateSymbolList(char *name, Symbol *element)
{
Symbol	*sym = FindSymbol(name);

	ListSymbol *listSym = sym->GetListSymbol();

	if (listSym->GetMemberType() != element->GetSymType())
	{
		printf("\nMetaClass::UpdateSymboList: Wrong element type %s",
			 name);
		exit(1);
	}

	listSym->append(element);

} /* UpdateSymbolList */

/****************************************************************/
/* StringSymbolList copy */
/****************************************************************/

void	StringSymbolList::Copy(Symbol *sym)
{
ListSymbol	*lsym = sym->GetListSymbol();

	if (lsym->GetMemberType() != StringTypeDef)
	{
		printf("\nStringSymbolList:Copy error");
		exit(1);
	}

	ListIterator	iterator(this);
	StringSymbol	*ssym;

	while ((ssym=(StringSymbol *)iterator.getNext()) !=
			(StringSymbol *) 0)
		lsym->append((NEW StringSymbol(ssym->GetValue(), "")));

} /* Copy */

/****************************************************************/

void	IntegerSymbolList::Copy(Symbol *sym)
{
ListSymbol	*lsym = sym->GetListSymbol();

	if (lsym->GetMemberType() != IntegerTypeDef)
	{
		printf("\nIntegerSymbolList:Copy error");
		exit(1);
	}

	ListIterator	iterator(this);
	IntegerSymbol	*isym;

	while ((isym=(IntegerSymbol *)iterator.getNext()) !=
			(IntegerSymbol *) 0)
		lsym->append((NEW IntegerSymbol(isym->GetValue(), "")));

} /* Copy */

/****************************************************************/

void	ObjectRefSymbolList::Copy(Symbol *sym)
{
ListSymbol	*lsym = sym->GetListSymbol();

	if (lsym->GetMemberType() != ObjectRefTypeDef)
	{
		printf("\nObjectRefSymbolList:Copy error");
		exit(1);
	}

	ListIterator	iterator(this);
	ObjectRefSymbol	*osym;

	while ((osym=(ObjectRefSymbol *)iterator.getNext()) !=
			(ObjectRefSymbol *) 0)
		lsym->append((NEW ObjectRefSymbol(osym->GetValue(), "")));

} /* Copy */

/****************************************************************/
/* End of module */
/****************************************************************/
