//
// Copyright (c) 1993 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// AssociationMgr.h - This class handles all the informations related
//                    to the associations.
//
// By Luca Deri, July 1993


#ifndef _ASSOCIATION_MANAGER_
#define _ASSOCIATION_MANAGER_

#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "Boolean.h"

#define OsimisErr int

class CMISAgent; /* To be defined */

/* Error Codes returned by the AssociationMgr class */
#define DESCRIPTOR_ALREADY_EXISTING  -13000
#define UNKNOWN_DESCRIPTOR           -13001

/* Used for AccessType */
#define ACCESS_DENIED     0
#define ACCESS_PERMITTED  1

/* Denial Response */
#define DENIAL_RESPONSE_GIVEN    0
#define DENIAL_NO_RESPONSE       1
#define DENIAL_ABORT_ASSOCIATION 2
#define DENIAL_FALSE_RESPONSE    3

/* Default access permission */
#define ACTION_OPERATION                   0x001
#define CREATE_OPERATION                   0x002
#define DELETE_OPERATION                   0x004
#define GET_OPERATION                      0x008
#define REPLACE_OPERATION                  0x010
#define ADD_OPERATION                      0x020
#define REMOVE_OPERATION                   0x040
#define REPLACEWITHDEFAULT_OPERATION       0x080
#define MULTIPLEOBJECTSELECTION_OPERATION  0x100
#define FILTER_OPERATION                   0x200
#define ALL_OPERATION_ALLOWED              0x3FF

/* Denial granularity constants */
#define DENIAL_ON_REQUEST   0
#define DENIAL_ON_OBJECT    1
#define DENIAL_ON_ATTRIBUTE 2
/* DENIAL_ON_ATTRIBUTE will be supported with a new OSIMIS version
   that will implement the item rules */   

typedef struct KnownUser_struct {
  int denialResponse;           /* State of the denial response */
  int accessType;               /* Type of the access */
} KnownUser;

typedef struct UnknownUser_struct {
  int defaultAccessPermission;  /* Access permission (default access) */
  int defaultDenialGranularity; /* Denial granularity (default access) */
} UnknownUser;


/* User access restrictions struct definition */
typedef struct UserRights_struct {
 
  short theChoice;
#define KNOWN_USER   0
#define UNKNOWN_USER 1
  
  union {
    KnownUser knownUser;
    UnknownUser unknownUser;
  } theRights;
} UserRights;

typedef int (*InvokeIdGenerator)(const int fd, const int prev);

class AssociationDescriptor 
{
 private:
  int  theFileDescriptor;   /* File descriptor related to the open
			       association */
  char *theInitiatorName;   /* The name of the association initiator
			       contained in the acse token */
  UserRights userRights;    /* User access restrictions */
  Bool toBeUpdated;         /* True is the user rights have to be
			       updated */
  int               _agentId;
  int               _prevAgentId;
  int               _managerId;
  int               _prevManagerId;
  InvokeIdGenerator _invokeIdFunc;

 public:
  AssociationDescriptor();
  ~AssociationDescriptor();
  inline void  SetFileDescriptor(int theFd) { theFileDescriptor = theFd; };
  inline int   GetFileDescriptor() { return(theFileDescriptor); };

  char* GetSessionKey();
  void  SetSessionKey(char* theKey);

  inline void  SetInitiatorName(char* theName) { theInitiatorName = theName;};
  inline char* GetInitiatorName() { return(theInitiatorName); };
  inline UserRights* GetUserRights() { return(&userRights); };
  void   UpdateSecurityDatas();
  inline void SetToBeUpdated() { toBeUpdated = True;  };
  inline Bool IsToBeUpdated()  { return(toBeUpdated); };

  int getAgentId() { return _agentId; }
  int getPrevAgentId() { return _prevAgentId; }
  int makeAgentId();

  int getManagerId() { return _managerId; }
  int getPrevManagerId() { return _prevManagerId; }
  int makeManagerId();
};

class AssociationMgr
{
 private:  
  int numDescriptors;
  AssociationDescriptor **theDescriptor;

  AssociationDescriptor *find(const int fd);

 public:
  AssociationMgr();
  ~AssociationMgr();
  OsimisErr InsertNewAssocDescriptor(int);
  OsimisErr RemoveAssocDescriptor(int);
  char*     GetSessionKey(int);
  OsimisErr StoreSessionInfo(int, char*, char*);
  Bool VerifySessionPermission(int, int, int, int, CMISAgent*);
  void MarkAssocDescriptor(Bool);

  /* InvokeId management */
  int getAgentId(const int fd);
  int getPrevAgentId(const int fd);
  int makeAgentId(const int fd);

  int getManagerId(const int fd);
  int getPrevManagerId(const int fd);
  int makeManagerId(const int fd);
};

extern AssociationMgr *theAssocMgr;

#endif
