//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenericMO.h - generic managed object declaration

// By George Pavlou, February 1991


#ifndef _GENERIC_MO_
#define _GENERIC_MO_


#ifndef _MO_CLASS_INFO_
#include "MOClassInfo.h"
#endif

#ifndef _AVA_
#include "AVA.h"
#endif

#if	defined(PERSISTENCY) && !defined(_dbm_lock_h_)
#include "dbm-lock.h"
#endif


#define MO_MAXDERIVED	10	/* max depth of MO class inheritance tree */
#define MO_MAXATTRS	100	/* max no of attributes per MO class */
#define MO_SCOPEBLOCK	50	/* block for scoping */

#define MO_AgentMibDirectory	"mib"	/* directory for persistent MOs file */
#define MO_AgentMibFile		"mib"	/* ndbm file for persistent MOs */
#define MO_AgentDirectoryMode	0775 	/* creation mode - rwxrwxr-x */
#define MO_AgentMibFileMode	0664	/* creation mode - rw-rw-r-- */

extern char *logDir;         /* The directory for logging */
extern int  logRecordAccess; /* A flag to control scoping of logRecords */


typedef enum
{
    ct_undefined,
    ct_initialisation,
    ct_managementOperation,
    ct_resourceOperation

} CreationType;


class MO				// generic managed object class
{
friend class Top;
friend class CMISAgent;

#define MO_ALLEVELS	-1
#define MO_ALLATTRS	-1
#define MO_NOMOREATTRS	-2

    static  MO*   _root;		// the MIB root
    static  MO**  _scopedMOs;		// the array of scoped MOs
    static  Bool  _wrongnb;		// the last MO created had wrong
					// name binding
#ifdef PERSISTENCY
    static  DbmFile*  _db;		// the dbm file for MO dumps
#endif
    static  char* _agentMibFile;	// the filename for _db;
    static  Bool  _notificationFunction; // the filename for _db;

    // the following three are needed for the CMIS/SNMP proxy

    static  OID   _operationClass;	// class of MO being operated upon
    static  PE _attrSetPE;		// hack to make PEs available to set() -
    static  External* _accessControl;	// hack to make access control info
					// available to createRR() -

    RDN	    _rdn;			// relative distinguished name
    DN	    _dn;			// distinguished name
    MO*     _peer;			// next object at the same level
    MO*     _subordinate;		// first subordinate object
    MO*     _superior;			// immediate ancestor object
    MO*	    _classnext;			// linked list of all class objects
    Bool    _updated;
    Bool    _persistent;

    int	    _nclasses;
    MOClassInfo** _classes;
    int*    _nattrs;
    Attr*** _attrs;
    NameBinding* _activeNB;

    void    buildDN ();
    void    nullifyDN ();
    OID*    getGroup (OID, int*, OID = NULL);	// returns group attrids

    MO*	    getMO_aux (DN, Bool, AVA*&);
    MO*	    getSubordinate_aux (RDN, Bool, AVA*&);
    int     searchMIB (int*, CMISScope*, int, Bool, AVA*&); // Scope_aux
    Bool    Filter_aux (CMISFilter*);

protected:
    // virtual methods (to be possibly provided in derived classes)

    virtual int		get (int, int, AVA*&,
				Bool = False, int = -1);	// CMIS GET
    virtual int		set (CMISModifyOp, int, int, void*, AVA*&,
				Bool = False, int = -1);	// CMIS GET
    virtual int		action (int, int, void*, void*&, Bool&, AVA*&,
				Bool = False, int = -1);	// CMIS ACTION
    virtual int		buildReport (int, int, void*&, Bool&);
								// CMIS EVREPORT
    virtual int		refreshSubordinates (AVA*&, int = -1);
    virtual int		refreshSubordinate (RDN, AVA*&, int = -1);
								// MIT refresh
    virtual CMISErrors	cmisDelete ();				// deprecated

    // other methods

    static OID		getOperationClass ();
    static OID		setOperationClass (OID);
    static PE		getAttrSetPE ();
    static External*	getAccessControl ();

    virtual int	initialiseSpecialAttrs();			// top ones
    int		registerClass(int, Attr**, MOClassInfo*, Bool);
    int		checkNameBinding ();

#ifdef PERSISTENCY
    // Database access

    static datum *getKey(int = DBM_GetNextKey); // DBM_Get[First | Next]Key
    static int   store(datum *, datum *);
    static int   fetch(datum *, datum *);
    static int   remove(datum *);
#endif

    // This method is for use only by log objects for logRecords!
    void setSuperior(MO *);

    // constructors

    MO ();				// for the MIB root
    MO (RDN, MO*);			// for other MOs

public:

    // find information about classes, attributes, actions, notifications

    Attr*	findAttr (OID, char** = NULL);
    Attr*	findAttr (char*, int* = NULL, int* = NULL);
    int		findAttr (OID, int*, int*,
				OID = NULL, Bool* = NULL, Attr** = NULL);
    Bool	isState (int, int);

    int		findAction (OID, int*, int*,
				OID = NULL, Attr** = NULL, Attr** = NULL);

    OID		findEvent (char*);
    OID		findEvent (int, int);
    int		findEvent (OID, int*, int*, char** = NULL);

    // CMISAgent -> MIB Interface

    MO**	Scope (CMISScope*, Bool, AVA*&);
								// CMIS SCOPE
    Bool	Filter (CMISFilter*, AVA*&);
								// CMIS FILTER
    CMISErrors	Get (OID, int, MID, int*, CMISGetAttr*, AVA*&);
								// CMIS GET
    CMISErrors	Get (OID, int*, CMISParam*);			// CMIS GET all

    CMISErrors	Set (Bool, OID, int, CMISSetAttr*, int*, CMISSetAttr*, AVA*&);
								// CMIS SET
    CMISErrors	Action (Bool, OID, CMISParam*, CMISParam*, AVA*&);
								// CMIS ACTION
    static MO*	Create (OID, DN*, RDN, MO*, MO*, External*, int, CMISParam*,
			CMISErrors*, CMISErrorInfo*);		// CMIS CREATE
    CMISErrors  Delete (AVA*&);					// CMIS DELETE

    void	releaseEncodedData ();	// for reducing space requirements


    // virtual methods (to be possibly provided in derived classes)

    virtual int	createRR (AVA*&, void* = NULLVD, int = -1);
								// CREATE
    virtual int	deleteRR (AVA*&, void* = NULLVD, Bool = False, int = -1);
								// DELETE


    // MO, CMISAgent or any other "back-end" object -> MIB Interface

    static int	initialiseMIB (char*);
    static void	terminateMIB ();
    static MO*	getRoot ();

    int		update (AVA*&);
    MO*		getMO (DN, Bool, AVA*&);
    MO*		getMO (char*, Bool, AVA*&);
    MO*		getSubordinate (RDN, Bool, AVA*&);
    MO*		getSubordinate (char*, Bool, AVA*&);
    MO**	getSubordinates (Bool, AVA*&);
    MO**	getWholeSubtree (Bool, AVA*&);

    MO*		getSubordinate ();
    MO*		getSuperior ();
    MO*		getPeer ();

    MOClassInfo* getClassInfo ();
    MO*		getClassNext ();

    OID		getClass ();
    char*	getClassName ();
    int		checkClass (OID, Bool* = NULL);
    int		checkPackage (OID);
    RDN		getRDN ();
    char*	getRDNString ();
    int		setRDN (RDN);
    int		setRDN (char*);
    DN		getDN ();
    char*	getDNString ();
    NameBinding* getNameBinding ();

    Attr*	getUpdatedAttr (OID, AVA*&);	// forces a CMIS GET
    Attr*	getAttr (OID, char** = NULL);
    Attr*	getAttr (char*);
    void*	getAttrVal (char*);
    int		setAttrVal (char*, void*);
    int		replAttrVal (char*, void*);

    char*	getEventName (OID);


    /*
    ** To allow MO to be dumped tp disc, and database access
    ** Still experimental.
    */
#ifdef PERSISTENCY
    int		save();		// save to disc (uses save(char *, int *))
    int		remove();	// remove this MO from disc
    int		restore();	// restore from disc (uses restore(char *))
    DbmFile	*getDbmFile();
#endif
    int		save(char *, int *); // save to (char *) data format
    int		restore(char *);// restore from (char *) data format
    void	makePersistent();
    void	makeNonPersistent();


    // destructor

    virtual ~MO ();			// does delete
};

#define NULLMO ((MO*) 0)


inline int  MO::get (int, int, AVA*&, Bool, int)
{ return OK; }

inline int  MO::set (CMISModifyOp, int, int, void*, AVA*&, Bool, int)
{ return m_noError; }

inline int  MO::action (int, int, void*, void*&, Bool&, AVA*&, Bool, int)
{ return m_processingFailure; }

inline int  MO::createRR (AVA*&, void*, int)
{ return OK; }

inline int  MO::deleteRR (AVA*&, void*, Bool, int)
{ return OK; }

inline CMISErrors  MO::cmisDelete ()
{ return m_processingFailure; }

inline OID  MO::findEvent (int eventId, int classLevel)
{ return _classes[classLevel] -> getEventOid(eventId); }

inline int  MO::buildReport (int, int, void *&eventInfo, Bool& freeFlag)
{ eventInfo = NULLVD; freeFlag = False; return OK; }

inline int  MO::refreshSubordinates (AVA*&, int)
{ return m_noError; }

inline int  MO::refreshSubordinate (RDN, AVA*&, int)
{ return refreshSubordinates(DISCARDAVA); }

inline OID MO::getOperationClass ()
{ return _operationClass; }

inline OID MO::setOperationClass (OID cl)
{ return _operationClass = cl; }

inline PE MO::getAttrSetPE ()
{ return _attrSetPE; }

inline External*  MO::getAccessControl ()
{ return _accessControl; }

inline int  MO::initialiseSpecialAttrs ()
{ return OK; }

inline int  MO::update (AVA*& err)
{ AVA* err_aux = NULLAVA;
  int ret = get(MO_ALLATTRS, MO_ALLEVELS, err == DISCARDAVA ? err_aux : err);
  if (err_aux) delete err_aux; return ret; }

inline MO*  MO::getMO (DN dn, Bool refresh, AVA*& err)
{ AVA* err_aux = NULLAVA; MO* mo;
  if ((mo = getMO_aux(dn, refresh, err == DISCARDAVA ? err_aux : err)) == NULLMO && err_aux )
    delete err_aux;
  return mo; }

inline MO*  MO::getSubordinate (RDN rdn, Bool refresh, AVA*& err)
{ AVA* err_aux = NULLAVA; MO* mo;
  if ((mo = getSubordinate_aux(rdn, refresh, err == DISCARDAVA ? err_aux : err)) == NULLMO
		&& err_aux )
    delete err_aux;
  return mo; }

inline MO*  MO::getMO (char* strDn, Bool refresh, AVA*& err)
{ char* cp = strdup(strDn); DN dn = str2dn(cp); free(cp);
  return dn ? getMO(dn, refresh, err) : NULLMO; }

inline MO*  MO::getSubordinate (char* strRdn, Bool refresh, AVA*& err)
{ char* cp = strdup(strRdn); RDN rdn = str2rdn(cp); free(cp);
  return rdn ? getSubordinate(rdn, refresh, err) : NULLMO; }

inline MO**  MO::getSubordinates (Bool refresh, AVA*& err)
{ CMISScope scope; scope.sc_type = Sc_FirstLevel;
  return Scope(&scope, refresh, err); }

inline MO**  MO::getWholeSubtree (Bool refresh, AVA*& err)
{ MO** mo; CMISScope scope; scope.sc_type = Sc_WholeSubtree;
  mo = Scope(&scope, refresh, err); return mo+1; }

inline OID  MO::getClass ()
{ return _classes[_nclasses-1] -> getClass(); }

inline char*  MO::getClassName ()
{ return _classes[_nclasses-1] -> getClassName(); }

inline int  MO::setRDN (char* strRdn)
{ char* cp = strdup(strRdn); RDN rdn = str2rdn(cp); free(cp);
  return rdn ? setRDN(rdn) : NOTOK; }

inline DN  MO::getDN ()
{ if (!_dn) buildDN(); return _dn; }

inline char*  MO::getDNString ()
{ return dn2str(getDN()); }

inline Attr*  MO::getAttr (OID attrOid, char** attrName)
{ return findAttr(attrOid, attrName); }

inline Attr*  MO::getAttr (char* attr)
{ return findAttr(attr); }

inline void*  MO::getAttrVal (char* attr)
{ Attr* at = getAttr(attr); return at ? at->get() : NULL; }

inline int  MO::setAttrVal (char* attr, void* val)
{ Attr* at = getAttr(attr); if (at) at->set(val); return at ? OK : NOTOK; }

inline int  MO::replAttrVal (char* attr, void* val)
{ Attr* at = getAttr(attr); if (at) at->replval(val); return at ? OK : NOTOK; }

inline char*  MO::getEventName (OID oid)
{ char* name = NULLCP; int i, j; findEvent(oid, &i, &j, &name); return name; }

#ifdef PERSISTENCY
inline DbmFile *MO::getDbmFile()
{ return(_db); }
#endif

inline void  MO::makePersistent ()
{ _persistent = True; }

inline void  MO::makeNonPersistent ()
{ _persistent = False; }

inline MO*  MO::getRoot ()
{ return _root; }

inline MO*  MO::getPeer ()
{ return _peer; }

inline MO*  MO::getSubordinate ()
{ return _subordinate; }

inline MO*  MO::getSuperior ()
{ return _superior; }

inline MOClassInfo*  MO::getClassInfo ()
{ return _classes[_nclasses-1]; }

inline MO*  MO::getClassNext ()
{ return _classnext; }

inline RDN  MO::getRDN ()
{ return _rdn; }

inline char*  MO::getRDNString ()
{ return rdn2str(_rdn); }

inline void MO::setSuperior(MO *superior)
{ _superior = superior; }


#endif
