//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// GenericAttr.cc - generic management attribute realisation

// By George Pavlou, December 1991


#include <stdlib.h>
#include "msap.h"
#include "GenericAttr.h"


Attr::~Attr ()
{
    clear();
}


void  Attr::clear ()
{
    if (_val) {
	_free();
	_val = NULLVD;
    }
    if (_pe) {
	pe_free(_pe);
	_pe = NULLPE;
    }
}


PE  Attr::encode ()
{
    if (_pe)
	return _pe;
    return _pe = _encode();
}


int  Attr::setstr (char* strval)
{
    void* val;

    if ((val = _parse(strval)) == NULLVD)
	return NOTOK;

    set(val);
    return OK;
}


CMISErrors  Attr::set (PE attrVal)
{
    void* newVal;
    if ((newVal = _decode(attrVal)) == NULL &&
		(!isMultiValued() || attrVal->pe_errno != PE_ERR_NONE))
	return m_invalidAttributeValue;
    if (set(newVal) == NOTOK)
	return m_invalidAttributeValue;
    if (_val != newVal)		// existing memory was re-used
	ffree(newVal);
    return m_noError;
}


void**  Attr::_find (void* elem, void** list)
{
    void** cur;
    if (!list || !*list)
	return NULL;

    for (cur = list; *cur; cur = _getnext(*cur))
	if (_compare(elem, *cur) == 0)
	    return cur;
    return NULL;
}


int  Attr::_compare (void* v1, void* v2)
{
    register void **nextv1p = NULL,
                  **nextv2p = NULL;
    register void *nextv1,
                  *nextv2;
    register PE   pe1, pe2;
    int  ret;
    Bool localpe = False;

    if (!v1 || ! v2)
	return v1 == v2 ? 0 : 2;

    if (nextv1p = (void**) _getnext(v1))
	nextv2p = (void**) _getnext(v2);

    if (nextv1p) {
	nextv1   = *nextv1p;
	nextv2   = *nextv2p;
	*nextv1p = NULL;
	*nextv2p = NULL;
    }

    pe1 = encode(v1);
    if (!nextv2p && v2 == _val){// non multi-valued and v2=the attr value
	pe2 = encode();		// to optimise encoding: pe stays in the attr
	localpe = True;
    }
    else
	pe2 = encode(v2);

    ret = pe_cmp(pe1, pe2);

    pe_free(pe1);
    if (!localpe)
	pe_free(pe2);

    if (nextv1p) {
	*nextv1p = nextv1;
	*nextv2p = nextv2;
    }

    return ret == 0 ? 0 : 2;		/* compare only for equality */
}


int  Attr::add (void* addVal)
{
    if (!addVal || _getnext(addVal) == NULL)
	return NOTOK;		// no value or no set-valued attribute

    void* val = get();		// may fetch value from the real resource

    replval(NULL);
    if (set(addVal) == NOTOK) {	// set may do additional checks
	replval(val);
	return NOTOK;
    }

    // check for duplicate elements and ignore them
    register void **cur;
    void **tmp, *duplElem;

    for (cur = &addVal; *cur; )
	if (_find(*cur, &val) != NULL) {
	    duplElem = *cur;
	    tmp = _getnext(*cur);
	    *cur = *tmp;
	    *tmp = NULL;	// not to free other elements
	    ffree(duplElem);
	}
	else
	    cur = _getnext(*cur);

    for (cur = &val; *cur; cur = _getnext(*cur));
    *cur = addVal;		// add at the end
    replval(NULL);
    set(val);

    return OK;
}


int  Attr::remove (void* rmVal)
{
    if (!rmVal || _getnext(rmVal) == NULL)
	return NOTOK;		// no value or no set-valued attribute

    void *val = get(),		// may fetch value from the real resource
         **cur, *rm, **rmElem[20];
    int i;

    // find first all supplied elements
    for (cur = &rmVal, i = 0; *cur; cur = _getnext(*cur), i++)
	if ((rmElem[i] = _find(*cur, &val)) == NULL)
	    return NOTOK;	// element not found

    // now remove them
    while (--i >= 0) {
	rm = *rmElem[i];
	cur = _getnext(*rmElem[i]);
	*rmElem[i] = *cur;
	*cur = NULL;		// not to free other elements
	ffree(rm);
    }

    replval(NULL);
    set(val);
    return OK;
}


Bool  Attr::filter (int type, void* val)
{
    Bool setvalued = (!val || _getnext(val)) ? True : False;
    void *value = get(),	// may fetch the real resource value
         **next;
    int  cmp;

    switch (type) {
    case FI_Equality:
	if (!setvalued)
	    return _compare(val, value) == 0 ? True : False;

	if (!value || !val) {
	    if (!value && !val)
		return True;
	    return False;
	}
	for (next = &val; *next; next = _getnext(*next))
	    if (_find(*next, &value) == NULL)
		return False;
	for (next = &value; *next; next = _getnext(*next))
	    if (_find(*next, &val) == NULL)
		return False;
	return True;

    case FI_GreaterOrEqual:
	if (!setvalued) {
	    cmp = _compare(val, value);
	    return (cmp == 1 || cmp == 0) ? True : False;
	}

	if (value == NULL || val == NULL || *(next = _getnext(val)) != NULL)
	    return False;	// asserted value should have exactly one elem
	for (next = &value; *next; next = _getnext(*next)) {
	    if ((cmp = _compare(val, *next)) == 2)
		return False;	// not comparable
	    if (cmp == 1 || cmp == 0)
		return True;
	}
	return False;

    case FI_LessOrEqual:
	if (!setvalued) {
	    cmp = _compare(val, value);
	    return (cmp == -1 || cmp == 0) ? True : False;
	}

	if (value == NULL || val == NULL || *(next = _getnext(val)) != NULL)
	    return False;	// asserted value should exactly only one elem
	for (next = &value; *next; next = _getnext(*next)) {
	    if ((cmp = _compare(val, *next)) == 2)
		return False;	// not comparable
	    if (cmp == -1 || cmp == 0)
		return True;
	}
	return False;

    case FI_SubsetOf:
	if (!setvalued)
	    return False;

	if (!value || !val) {
	    if (!val)
		return True;
	    return False;
	}

	for (next = &val; *next; next = _getnext(*next))
	    if (_find(*next, &value) == NULL)
		return False;
	return True;

    case FI_SupersetOf:
	if (!setvalued)
	    return False;

	if (!value || !val) {
	    if (!value)
		return True;
	    return False;
	}

	for (next = &value; *next; next = _getnext(*next))
	    if (_find(*next, &val) == NULL)
		return False;
	return True;

    case FI_NonNullSetIntersect:
	if (!setvalued)
	    return False;

	if (!value || !val)
	    return False;

	for (next = &val; *next; next = _getnext(*next))
	    if (_find(*next, &value) == NULL)
		return True;
	return False;


    default:
	return False;
    }
}


Bool  Attr::filter (CMISFilter* flt)
{
    switch (flt -> flt_type) {
    case Filter_Item:
    {
        Attr* attr = this;
        void* value;
        Bool  status;

        if (flt -> flt_item -> fi_type == FI_Substrings) {
            register CMISFilterSubstrings* substr =
                                        flt->flt_item->fi_substrings;

            for (; substr; substr = substr->fi_sub_next) {
                if ((value = attr -> decode(substr->fi_sub_string))
                                == NULLVD &&
                        substr->fi_sub_string->pe_errno
                                != PE_ERR_NONE)
                    return False;       // decoding failed

                Bool mode = (substr == flt->flt_item->fi_substrings) ?
                                        True : False;
                status = attr ->
                        filterSubstring(substr->fi_sub_type, value, mode);

                attr -> ffree(value);   // free decoded value
                if (status == False)
                    return False;
            }
            return True;        // logical ANDing of substring list
        }

        if (flt->flt_item->fi_type == FI_Present)
            return True;

        if ((value = attr -> decode(flt->flt_item->fi_attr.mp_val)) == NULLVD &&
		(!isMultiValued() ||
		flt->flt_item->fi_attr.mp_val->pe_errno != PE_ERR_NONE))
            return False;       // decoding failed

        status = attr -> filter(flt->flt_item->fi_type, value);

	attr -> ffree(value);   // free decoded value
        return status;
    }

    case Filter_Not:
        return (Bool) !filter(flt -> flt_filter);

    case Filter_And:
    {
        for (register CMISFilter* cur = flt -> flt_filter;
                                        cur; cur = cur -> flt_next)
            if (filter(cur) == False)
                return False;
    }
        return True;

    case Filter_Or:
    {
        register CMISFilter* cur = flt -> flt_filter;
        if (!cur)
            return True;
        for ( ; cur; cur = cur -> flt_next)
	    if (filter(cur) == True)
                return True;
    }
        return False;

    default:
        return False;   // never reached because of check_filter()
    }
}

