//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//


/******************************************************************/
/*

	File:	cmisclient.cc

	Notes:	CMIS Client Class Methods

	Author:
		James Cowan (J.Cowan@uk.ac.ucl.cs)
*/
/******************************************************************/

#include "cmis.h"
#include "cmisclient.h"
#include "cmisresult.h"

/******************************************************************/
/* Public methods */
/******************************************************************/
/* Initialise server */

Boolean	Cmis_client::initialise(AEI calling_title, 
				struct PSAPaddr *calling_address,
				AEI called_title, 
				struct PSAPaddr *called_address,
				External *access, External *info)
{
int	result;

	result = M_InitialiseReq(calling_title, calling_address,
				 called_title, called_address,
				 "management", CMIP_VERSION2, 0x17,
				 access, info, &_ms_connection,
				 &_ms_indication);

#ifdef	DEBUG_CLIENT
	CLIENT_TRACE("\nConnect status: %d, protocol: %d, funits: %d",
		    result, _ms_connection.mc_info.mai_protvrsn,
		    _ms_connection.mc_info.mai_funits);
#endif
	if (result == NOTOK)
		process_error("Initialise", &_ms_indication);

	mi_free(&_ms_indication);
	return(convert_ok(result));

} /* initialise */

/******************************************************************/
/* friendly initialise routine */

Boolean	Cmis_client::initialise(char *host, char *service)
{
AEI	aei;
struct	PSAPaddr *pa;

#ifdef USE_X500

  if ( strcmp (service, "@") == 0 ) {
    if ( (pa = lt_read(host)) == NULLPA ) {
        char hostpart[64],
             servicepart[64];

        /* Try to make a an host/agent pair from the DN (host) */
        host = rindex(host, '=');
        host++;
        sscanf(host, "%s-%s",hostpart, servicepart);
        aei = str2aei(hostpart, servicepart);
        pa = aei2addr(aei);
    } else
      if ( lt_oper_status == DOWN ) {
        fprintf(stderr,"The specified agent is not currently running.\n");
        return NOTOK;
      }
    aei = NULLAEI;
  } else {
    aei = str2aei(host, service);
    pa = aei2addr(aei);
  }

#else

        aei = str2aei(host, service);
        pa = aei2addr(aei);

#endif

	return(initialise(NULLAEI, NULLPA, aei, pa, NULLEXTERN, NULLEXTERN));

} /* initialise */

/******************************************************************/
/* Terminate */

Boolean	Cmis_client::terminate(PE info)
{
int	result;

	result = M_TerminateReq(_ms_connection.mc_sd, ACF_NORMAL, info,
				&_ms_indication);
	if (result == NOTOK)
		process_error("Terminate", &_ms_indication);
	else
	{
#ifdef	DEBUG_CLIENT
		CLIENT_TRACE("\nTerminate Ok");
#endif
	}
	mi_free(&_ms_indication);
	return(convert_ok(result));

} /* terminate */

/******************************************************************/
/* Terminate */

Boolean	Cmis_client::terminate()

{
	return(terminate(NULLPE));

} /* terminate */

/******************************************************************/
/* Terminate urgently */

Boolean	Cmis_client::urgent_terminate(PE info)
{
int	result;

	result = M_TerminateReq(_ms_connection.mc_sd, ACF_URGENT, info,
				&_ms_indication);
	if (result == NOTOK)
		process_error("Urgent Terminate", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nUrgent Terminate Ok");
#endif

	mi_free(&_ms_indication);
	return(convert_ok(result));

} /* urgent_terminate */

/******************************************************************/
/* Urgent Terminate */

Boolean	Cmis_client::urgent_terminate()

{
	return(urgent_terminate(NULLPE));

} /* urgent_terminate */

/******************************************************************/
/* Abort */

Boolean	Cmis_client::abort(External *info)
{
int	result;

	result = M_AbortReq(_ms_connection.mc_sd, info, &_ms_indication);

	if (result == NOTOK)
		process_error("Abort", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nAbort Ok");
#endif

	mi_free(&_ms_indication);
	return(convert_ok(result));

} /* Abort */

/******************************************************************/
/* Abort */

Boolean	Cmis_client::abort()

{
	return(abort(NULLEXTERN));

} /* abort */

/******************************************************************/
/* Create */


Boolean	Cmis_client::create(MID obj_class, MN obj_inst,
			    int  inst_type, MN ref_inst, External *access,
			    int nattrs, CMISParam *attrs) 
{
int		result;
int		invoke_id = get_invoke_count();
Create_result	create_result(invoke_id);

	result = M_Create(_ms_connection.mc_sd, invoke_id, obj_class,
			  obj_inst, inst_type, ref_inst, access, nattrs,
			  attrs, &_ms_indication);

	if (result == NOTOK)
		process_error("Create", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nCreate Ok");
#endif

	mi_free(&_ms_indication);
	if (result == OK)
		return(convert_result(wait_request(NOTOK, &create_result)));

	return(convert_ok(result));

} /* create */

/******************************************************************/
/* Delete */

Boolean	Cmis_client::remove(MID obj_class, MN obj_inst, CMISScope *scope,
			    CMISFilter *filter, External *access, CMISSync sync)
{
int		result;
int		invoke_id = get_invoke_count();
Delete_result	delete_result(invoke_id);

	result = M_Delete(_ms_connection.mc_sd, invoke_id, obj_class,
			  obj_inst, scope, filter, access, sync,
			  &_ms_indication);

	if (result == NOTOK)
		process_error("Delete", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nDelete Ok");
#endif

	mi_free(&_ms_indication);
	if (result == OK)
		return(convert_result(wait_request(NOTOK, &delete_result)));

	return(convert_ok(result));

} /* remove */

/******************************************************************/
/* Get */

Boolean	Cmis_client::get(MID obj_class, MN obj_inst, CMISScope *scope,
			 CMISFilter *filter, External *access, CMISSync sync,
			 int nattrs, MID attrs, Cmis_result *result_list)
{
int		result;
int		invoke_id = get_invoke_count();
Get_result	get_result(invoke_id, result_list);

	result = M_Get(_ms_connection.mc_sd, invoke_id, obj_class, obj_inst,
		       scope, filter, access, sync, nattrs, attrs,
		       &_ms_indication);

	if (result == NOTOK)
		process_error("Get", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nGet Ok");
#endif

	mi_free(&_ms_indication);
	if (result == OK)
	{
	Result	wait_result;

		do
		{
			wait_result = wait_request(NOTOK, &get_result);
#ifdef	DEBUG_CLIENT
			CLIENT_TRACE("\nWait_result %d", wait_result);
#endif
		}
		while (wait_result == Continue);

		return(convert_result(wait_result));
	}

	return(convert_ok(result));

} /* get */

/******************************************************************/
/* Get base and subordinates at next n levels */
/* Note that class name and instance are passed as string objects
   not addresses because ISODE overwrites them!! */

Boolean	Cmis_client::get(char *class_name, char *instance, int nattrs,
			 int level, Cmis_result *result_list)
{
MIDentifier	m_class;
MName           inst;
MIDentifier	attrs[1];
CMISScope       scope;
CMISSync        sync = s_bestEffort;
Boolean		result;

	encode_class_name(class_name, &m_class);
	encode_instance(instance, &inst);

        scope.sc_type = (level == 0 ? Sc_BaseObject : Sc_IndividualLevel);
        // scope.sc_type = (level == 0 ? Sc_BaseObject : Sc_BaseToNthLevel);
        scope.sc_level = level;

        result = get(&m_class, &inst, &scope, NULLMFILTER,
		     NULLEXTERN, sync, nattrs, attrs, result_list);

	mid_free(&m_class);
	mn_free(&inst);
	return(result);

} /* get */

/******************************************************************/
/* Set */

Boolean	Cmis_client::set(MID obj_class, MN obj_inst, CMISScope *scope,
			 CMISFilter *filter, External *access, CMISSync sync,
			 int nattrs, CMISSetAttr *attrs,
			 Cmis_result *result_list)
{
int		result;
int		invoke_id = get_invoke_count();
Set_result	set_result(invoke_id, result_list);

	result = M_SetConf(_ms_connection.mc_sd, invoke_id, obj_class, obj_inst,
		           scope, filter, access, sync, nattrs, attrs,
			   &_ms_indication);

	if (result == NOTOK)
		process_error("Set", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nSet Ok");
#endif

	mi_free(&_ms_indication);
	if (result == OK)
	{
	Result	wait_result;

		do
		{
			wait_result = wait_request(NOTOK, &set_result);
//			wait_result = wait_request(10, &set_result);
#ifdef	DEBUG_CLIENT
			CLIENT_TRACE("\nWait_result %d", wait_result);
#endif
		}
		while (wait_result == Continue);

		return(convert_result(wait_result));
	}

	return(convert_ok(result));

} /* set */

/******************************************************************/
/* set */

Boolean	Cmis_client::set(char *class_name, char *instance,
			 char *attribute, char *value,
			 Cmis_result *result_list)
{
Boolean		result = False;
X500_attribute	x500;
CMISSetAttr	attrs[1];

	if (x500.encode(attribute, value, &(attrs[0].sa_id.mid_global),
			&(attrs[0].sa_val)))
	{
	MIDentifier	m_class;
	MName           inst;
	CMISScope       scope;
	CMISSync        sync = s_bestEffort;

		attrs[0].sa_id.mid_type = MID_GLOBAL;
		attrs[0].sa_modify = m_replace;
		scope.sc_type = Sc_BaseObject;

		encode_class_name(class_name, &m_class);
		encode_instance(instance, &inst);

//        	result = set(&m_class, &inst, &scope, NULLMFILTER,
        	result = set(&m_class, &inst, NULLMSCOPE, NULLMFILTER,
			     NULLEXTERN, sync, 1, attrs, result_list);

		mid_free(&m_class);
		mn_free(&inst);
	}
	return(result);

} /* set */

/******************************************************************/
/* set */

Boolean	Cmis_client::set(char *class_name, char *instance,
			 char *attribute, Cmis_result *result_list)
{
Boolean		result = False;
CMISSetAttr	attrs[1];
MIDentifier	m_class;
MName           inst;
CMISScope       scope;
CMISSync        sync = s_bestEffort;

	encode_class_name(class_name, &m_class);
	encode_instance(instance, &inst);

	encode_class_name(attribute, &(attrs[0].sa_id));
	attrs[0].sa_modify = m_setToDefault;
	scope.sc_type = Sc_BaseObject;

	result = set(&m_class, &inst, NULLMSCOPE, NULLMFILTER,
		     NULLEXTERN, sync, 1, attrs, result_list);

	mid_free(&m_class);
	mn_free(&inst);

	return(result);

} /* set */

/******************************************************************/
/* Event report */

Boolean	Cmis_client::event_report(int invoke, MID obj_class, MN obj_inst,
				  MID type, char *time, PE info)
{
int	result;

	result = M_EventRep(_ms_connection.mc_sd, invoke, obj_class,
			    obj_inst, type, time, info, &_ms_indication);

	if (result == NOTOK)
		process_error("EventReport", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nEventRep Ok");
#endif

	mi_free(&_ms_indication);
	return(convert_ok(result));

} /* event_report */

/******************************************************************/

Boolean	Cmis_client::action(MID obj_class, MN obj_inst, CMISScope *scope,
			    CMISFilter *filter, External *access, CMISSync sync,
			    MID type, PE info)
{
int		result;
int		invoke_id = get_invoke_count();
Action_result	action_result(invoke_id);

	result = M_ActionConf(_ms_connection.mc_sd, invoke_id, obj_class,
			      obj_inst, scope, filter, access, sync,
			      type, info, &_ms_indication);

	if (result == NOTOK)
		process_error("Action", &_ms_indication);
#ifdef	DEBUG_CLIENT
	else
		CLIENT_TRACE("\nAction Ok");
#endif
	mi_free(&_ms_indication);
	if (result == OK)
		return(convert_result(wait_request(NOTOK, &action_result)));

	return(convert_ok(result));

} /* action */

/******************************************************************/
/* Private methods */
/******************************************************************/
/* Wait request */

Result	Cmis_client::wait_request(int secs, Cmis_client_result *client_result)
{
	/* If events in queue, check the queue */

	if (_wait_queue.count() > 0)
	{
	struct MSAPIndication *mi;
	int	count;

		for (count=0; mi=(struct MSAPIndication *) _wait_queue[count];
		     count++)
		{
		Result	result;

#ifdef	DEBUG_CLIENT
			CLIENT_TRACE("\nRead wait queue");
#endif
			if ((result = client_result->process(mi)) != Notfound)
			{
				_wait_queue.vdelete(count);
				return(result);
			}
		}
	}

	do
	{
	int	res;
	struct MSAPIndication *mi = (NEW struct MSAPIndication);
 
		res = M_WaitReq(_ms_connection.mc_sd, secs, mi);

		if (res == NOTOK)
		{
			process_error("WaitReq", mi);
			free_msap_indication(mi);
			return(Error);
		}
		else
		{
		Result	result;
#ifdef	DEBUG_CLIENT
			CLIENT_TRACE("\nWaitReq Ok");
#endif
			if ((result = client_result->process(mi)) != Notfound)
			{
				free_msap_indication(mi);
				return(result);
			}
		}
#ifdef	DEBUG_CLIENT
		CLIENT_TRACE("\nWaitReq loop");
#endif
		_wait_queue.add(mi);

		if (secs != NOTOK)
			return(Notfound);
	}
	while (1);

} /* wait_request */

/******************************************************************/

Boolean	Cmis_client::convert_ok(int type)
{
	switch (type)
	{
	default:
		printf("\nFATAL ERROR IN convert_ok %d", type);
		cmisbrowser_exit(1);

	case	OK:
		return(True);

	case	NOTOK:
		return(False);
	}

} /* convert_ok */

/******************************************************************/

Boolean	Cmis_client::convert_result(Result type)
{
	switch (type)
	{
	default:
		printf("\nFATAL ERROR IN convert_result %d", type);
		cmisbrowser_exit(1);

	case	Success:
	case	Finish:
		return(True);

	case	Error:
	case	Notfound:
		return(False);
	}

} /* convert_result */

/******************************************************************/

void	Cmis_client:: free_msap_indication(struct MSAPIndication *mi)
{
	mi_free(mi);
	delete mi;

} /* free_msap_indication */

/******************************************************************/

void	Cmis_client:: process_error(char *msg, struct MSAPIndication *mi)
{

	switch (mi->mi_type)
	{
	default:
#ifdef  DEBUG_CLIENT
		CLIENT_TRACE("\n%s error: unknown error", msg);
#endif
		break;

	case MI_ABORT:
#ifdef  DEBUG_CLIENT
		CLIENT_TRACE("\n%s: type %d, source: %d, reason: %d, data: %s",
			     msg, mi->mi_type, mi->mi_abort.ma_source, 
			     mi->mi_abort.ma_reason, mi->mi_abort.ma_data);
#endif
		break;

	case MI_PREJECT:
#ifdef  DEBUG_CLIENT
		CLIENT_TRACE("\n%s: type %d, reason: %d, data: %s",
			     msg, mi->mi_type, mi->mi_preject.mpr_reason,
			     mi->mi_preject.mpr_data);
#endif
		break;
	}

} /* process_error */

/******************************************************************/
/* End of module */
/******************************************************************/




