//
// Copyright (c) 1994 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// Set.cc - connect to an SMA agent on a given host (or otherwise using
//	the location transparency service), do a synchronous Set on the
//	{uxObjId=test} instance, print the result and disconnect.

// By Thurain Tin, July 1994


#include <stdlib.h>
#include "msap.h"
#include "Error.h"
#include "RMIBAgent.h"
#include "Syntaxes.h"
#include "ListIterator.h"

static void print_list (CMISObjectList* list);
static void print_attrs (AVAArray* attrs);


main (int argc, char** argv)
{
    if (argc < 3)
	error(NULLCP, "usage: mset <agent> <host>");

    RMIBAgent	agent;	// instantiate one Remote MIB agent
    // Note that the coordinator is only needed for asynchronous services,
    // this is why it is not instantiated here

    // load the syntaxes, see also initialiseSyntaxes(3n)
    if (initialiseSyntaxes(NULLCP, attrSyntaxes) == NOTOK)
	error(NULLCP, "unable to initialise syntaxes");

    // attempt connection to the remote agent
    if (agent.connect(argv[1], argv[2]) == NOTOK)
	error(NULLCP, "unable to establish association to %s at %s",
		argv[1], argv[2]);

    // declare an object list for the results
    CMISObjectList* objectList;

    AVAArray avaArray(1);
    AVA* ava = new AVA("wiseSaying", "Hello over there", m_replace);
    avaArray.add(ava);

    // perform the Set operation
    agent.Set(NULLCP, "uxObjId=test", Sc_BaseObject, 0, NULLCP,
		NULLMACCESS, s_bestEffort, &avaArray, objectList);

    // print the result
    if (objectList) {
	print_list(objectList);
	delete objectList;
    }

    delete ava;

    // disconnect from the remote agent
    agent.disconnect();
}


static void print_list (CMISObjectList* list)
{
    if (list->getError())
	fprintf(stdout, "\n%s\n", list->getErrorStr());

    // get the number of managed objects
    int n = list->getCount();

    if (n) {
	fprintf(stdout, "\nreply id: %d", list->getInvoke());
	fprintf(stdout, "\nnumber of managed objects = %d\n", n);

	// use the iterator to retrieve objects
	ListIterator iterator(list);
	CMISObject* obj;

	while (obj = (CMISObject*) iterator.getNext()) {
	    fprintf(stdout, "\nclass: %s\ninstance: {%s}",
		obj->getClass(), obj->getInstance());

	    // print the time of operation
	    fprintf(stdout, "\noperation time: %s", obj->getTime());

	    // print the attributes
	    print_attrs(obj->getAttrs());

	    // check for error
	    if (obj->getError())
		fprintf(stdout, "\nerror: %s", obj->getErrorStr());
	    printf("\n");
	}
    }
}


static void print_attrs (AVAArray* attrs)
{
    if (attrs) {
	// get the number of attributes
	int n = attrs->getCount();

	for (register int i = 0; i < n; i++) {
	    // get the ith entry
	    AVA* ava = (AVA*) attrs->get(i);

	    if (ava) {
		char* cp;
		Attr* value = ava->getValue();
		fprintf(stdout, "\n%s: %s%s", ava->getName(),
		    value ? "" : "error: ",
		    value ? (cp = value->print()) : ava->getErrorStr());
		if (cp) free(cp);
	    }
	}
    }
}
