//
// Copyright (c) 1990 University College London
// All rights reserved.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the Department of Computer Science, University College London.
// The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

// dbm-lock.cc - "nice" interface to gdbm

// Saleem N. Bhatti
// October 1992

/*****************************************************************************
To help with gdbm.

   dbmLock(), dbmUnlock(), dbmStore() and dbmDelete() return 0 on success,
   -1 on failure.
   dbmOpen() and dbmFetch() return NULL on failure.
   dbmGetKey() returns NULL if no keys exist or after it has returned the
   final key.

   NOTE :
   dbmFetch() and dbmGetKey() return a pointer to a static area that will be
   overwritten on the next call to those functions.

Now using fcntl(2) instead of flock(2). flock(2) is BSD only.
*****************************************************************************/

#if !defined(_dbm_lock_cc_)
#define _dbm_lock_cc_

#if defined(__GNUG__)
#include <std.h>
#elif defined(__cplusplus)
#include <stdlib.h>
#endif

#include <signal.h>

#ifdef AIXV3
#ifndef	ICR1
#include <isode/manifest.h>
#else
#include <isode/compat/manifest.h>
#endif
#endif

#include <unistd.h>
#include <fcntl.h>

#include "ucbext.h"

#include "gdbmdefs.h"
#include "extern.h"
#include "dbm-lock.h"

#if defined(DEBUG)
#if defined(__SUNPRO_CC)
#include <errno.h>
#else
#include <sys/errno.h>
#endif
#endif

#ifdef	__linux__
#include <linux/time.h>		/* for FD_SETSIZE */
#endif
static
datum prevKeys[FD_SETSIZE]; /* Shock horror! */

int
dbmLock(DbmFile *db, int lock)
{
    /*
    ** sort out retries later?
    */
    if (db == NULL) return(-1);

    if (lock == DBM_NoLock) return(0);

    struct flock fl;

    fl.l_type = lock == DBM_SharedLock ? F_RDLCK : (lock == DBM_ExclusiveLock ? F_WRLCK : -1);

    if (fl.l_type == -1) {
#if defined(DEBUG)
	printf("dbmLock() : bad lock type\n");
#endif
	return(-1);
    }
    fl.l_whence = 0;
    fl.l_start = 0;
    fl.l_len = 0;

    /* Blocking ?? */

#ifdef POSIX_SIGNALS
    sigset_t sig_set, old_sig_set;
    sigfillset(&(sig_set)); /* block all signals */
    sigprocmask (SIG_BLOCK,&sig_set,&old_sig_set);
#else
    register
    int mask = sigblock(0x7fffffff);	// block signals
#endif /* POSIX_SIGNALS */

    if (fcntl(db->desc, F_SETLK, (int) &fl) < 0) {
#if defined(DEBUG)
        printf("dbmLock() : flock() failed for %s\n", db->name);
#endif
#ifdef POSIX_SIGNALS
        sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
        sigsetmask(mask);	// restore signals
#endif /* POSIX_SIGNALS */
        return(-1);
    }

#ifdef POSIX_SIGNALS
    sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
    sigsetmask(mask);	// restore signals
#endif /* POSIX_SIGNALS */

    return(0);
}

int
dbmUnlock(DbmFile *db)
{
    /*
    ** sort out retries later?
    */
    if (db == NULL) return(-1);

    struct flock fl;

    fl.l_type = F_UNLCK;
    fl.l_whence = 0;
    fl.l_start = 0;
    fl.l_len = 0;

#ifdef POSIX_SIGNALS
    sigset_t sig_set, old_sig_set;
    sigfillset(&(sig_set)); /* block all signals */
    sigprocmask (SIG_BLOCK,&sig_set,&old_sig_set);
#else
    register
    int mask = sigblock(0x7fffffff);	// block signals
#endif /* POSIX_SIGNALS */

    if (fcntl(db->desc, F_SETLK, (int) &fl) < 0) {
#if defined(DEBUG)
        printf("dbmUnLock() : flock() failed for %s\n", db->name);
#endif
#ifdef POSIX_SIGNALS
        sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
        sigsetmask(mask);	// restore signals
#endif /* POSIX_SIGNALS */
        return(-1);
    }

#ifdef POSIX_SIGNALS
    sigprocmask (SIG_SETMASK, &old_sig_set, NULL);
#else
    sigsetmask(mask);	// restore signals
#endif /* POSIX_SIGNALS */

    return(0);
}

DbmFile *
dbmOpen(char *name, int flags, int mode)
{
    register
    int open_flags = flags == DBM_ReadOnly ? GDBM_READER : (flags == DBM_ReadWrite ? GDBM_WRCREAT : -1);

    if (open_flags == -1) {
#if defined(DEBUG)
	printf("dbmOpen() : bad flags\n");
#endif
	return(NULL);
    }

    register
    DbmFile *db = gdbm_open(name, 0, open_flags, mode, NULL);

#if defined(DEBUG)
    if (db == NULL) {
	printf("dbmOpen() : gdbm_open() failed for %s\n", name);
        printf("error : %d\n", errno);
    }
#endif

    if ((db != NULL) && (prevKeys[db->desc].dptr != NULL)) {
	free(prevKeys[db->desc].dptr);
	prevKeys[db->desc].dptr = NULL;
	prevKeys[db->desc].dsize = 0;
    }

    return(db);
}

void
dbmClose(DbmFile *db)
{
    if (db == NULL) return;

    if (prevKeys[db->desc].dptr != NULL) {
	free(prevKeys[db->desc].dptr);
	prevKeys[db->desc].dptr = NULL;
	prevKeys[db->desc].dsize = 0;
    }

    gdbm_close(db);
}

/*
** NOTE : Returns a pointer to a static area that will be overwritten on the
** next call to this function.
*/
datum *
dbmFetch(DbmFile *db, datum *key, int lock)
{
    static
    datum entry;

    if ((db == NULL) || (key == NULL)) return(NULL);
    if (entry.dptr != NULL) free(entry.dptr);

    if ((lock > 0) && (dbmLock(db, lock) < 0)) return(NULL);
    entry = gdbm_fetch(db, *key);
    if (lock > 0) dbmUnlock(db);

    return(entry.dptr != NULL ? &entry : NULL);
}

int
dbmStore(DbmFile *db, datum *key, datum *entry, int what, int lock)
{
    if ((db == NULL) || (key == NULL) || (entry == NULL)) return(-1);

    register
    int flags = what == DBM_Replace ? GDBM_REPLACE : (what == DBM_Insert ? GDBM_INSERT : -1);

    if (flags == -1) return(-1);

    if ((lock > 0) && (dbmLock(db, lock) < 0)) return(NULL);
    register
    int i = gdbm_store(db, *key, *entry, flags);
    if (lock > 0) dbmUnlock(db);

    return(i < 0 ? -1 : 0);
}

int
dbmDelete(DbmFile *db, datum *key, int lock)
{
    if ((db == NULL) || (key == NULL)) return(-1);

    if ((lock > 0) && (dbmLock(db, lock) < 0)) return(-1);
    register
    int i = gdbm_delete(db, *key);
    if (lock > 0) dbmUnlock(db);

    return(i < 0 ? -1 : 0);
}

/*
** NOTE : Returns a pointer to a static area that will be overwritten on the
** next call to this function.
*/
datum *
dbmGetKey(DbmFile *db, int what, int lock)
{
    static
    datum key;

    if ((db == NULL) || ((lock > 0) && (dbmLock(db, lock) < 0))) return(NULL);

    switch (what) {

    case DBM_GetNextKey :
        key = gdbm_nextkey(db, prevKeys[db->desc]);
	if (prevKeys[db->desc].dptr != NULL) free(prevKeys[db->desc].dptr);
	prevKeys[db->desc].dptr = key.dptr;
	prevKeys[db->desc].dsize = key.dsize;
        break;

    case DBM_GetFirstKey :
        key = gdbm_firstkey(db);
	if (prevKeys[db->desc].dptr != NULL) free(prevKeys[db->desc].dptr);
	prevKeys[db->desc].dptr = key.dptr;
	prevKeys[db->desc].dsize = key.dsize;
        break;

    default :
	key.dptr = NULL;
	key.dsize = 0;
	break;
    }

    if (lock > 0) dbmUnlock(db);

    return((key.dptr != NULL) && (*(key.dptr) != '\0') ? &key : NULL);
}

#endif /* !_dbm_lock_cc_ */
