/*
 * Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "xmp.h"
#include "mp_utilities.h"
#ifndef LPP
#define CMIP_ASN_OID    str2oid("2.9.1.1.4")            /* full CMIP stack */
#else
#define CMIP_ASN_OID    str2oid("1.3.6.1.2.1.9.1.2")    /* CMOT stack */
#endif

#if defined(SECURITY)
#include "mauth.h"
#include "mintegrity.h"
#include "minvoke.h"
#include "mp_integrity.h"
#endif

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

/************************************************************************/

static MIDAS_SESSION    default_session = {
                        NULLPA,
                        NULLAEI,
                        MP_T_MANAGING | MP_T_MONITORING | MP_T_PERFORMING |
                                                          MP_T_REPORTING,
                        MP_NO_VALID_FILE_DESCRIPTOR,
                        NULLPA,
                        NULLAEI,
                        (External *)NULL,       /*Access control*/
                        NULLOID,                /*Auth mech*/
                        NULL,                   /*Auth value*/
                        MP_T_FU_CANCEL_GET | MP_T_FU_FILTER |
                                MP_T_FU_MULTIPLE_OBJECT_SELECTION |
                                MP_T_FU_MULTIPLE_REPLY,
                        NULLOID,                /*Appl context*/
                        NULLOIDLIST,            /*Pres context list*/
                        (External *)NULL,       /*User info*/
                        NULL,                   /*SMASE user data*/
                        (MIDAS_SESSION *)NULL
};
/*************************************************************************/

static int workspace_initialised = FALSE;
static int workspace_secure = FALSE;
int session_initialised = FALSE;
static MIDAS_STATUS* build_session _a_((MIDAS_SESSION*, MIDAS_SESSION*));
int get_association_descriptor _a_((MIDAS_SESSION *, MIDAS_CONTEXT *));
MIDAS_STATUS* bad_session ();
void msp_Q_free _a_((MIDAS_SESSION*));
void oid_Q_free _a_((MIDAS_OID_LIST*));

static void free_oid_list _a_((MIDAS_OID_LIST *));

static char *context = "management";
static int protvrsn = CMIP_VERSION2;

/*
 * mp_bind_M - open a management session
 *
 * N.B. In "UCL mode" only one session is allowed per process 
 * but the session may support multiple associations
 *
 * In "GMD mode" multiple sessions are allowed but each may
 * support just one association
 *
 */

MIDAS_STATUS *
mp_bind_M(session, workspace, bound_session_return)
    MIDAS_SESSION *session, **bound_session_return;
    MIDAS_WORKSPACE workspace;
{
    static MIDAS_STATUS status, *statusp;
    register MIDAS_SESSION *new_session;
    static MSAPConnect mc;
    static MSAPIndication mi;
    int response;
#if defined(SECURITY)
    AuthAssocIntegrityInfo *aaii = (AuthAssocIntegrityInfo *) 0,
                           *aaii2 = (AuthAssocIntegrityInfo *) 0;
    int authFlag = FALSE;
    int e = 0;
#endif

    debug_test("In mp_bind_M @ 1");

    if (!workspace_initialised)
        return(MIDAS_NO_WORKSPACE);

    status.MIDAS_parameter = (MSAPIndication *)NULL;

    /*Allocate space for the session object we will return*/

    if ((new_session = *bound_session_return = 
			(MIDAS_SESSION *)malloc(sizeof(MIDAS_SESSION)))
			== (MIDAS_SESSION *)NULL)
	return(MIDAS_INSUFFICIENT_RESOURCES);

    /*Initialise the bound session with default values*/
    *new_session = default_session;    /*struct_copy*/

    if((!session->responder_title)&&(!session->responder_address)){
    /*The absence of naming/addressing information*/
    /*implies we are UCL mode of operation*/

        debug_test("In mp_bind_M @ 2 - in UCL mode");
        if (session_initialised){
	    msp_Q_free(new_session);
 	    *bound_session_return = (MIDAS_SESSION *)NULL;
            return(error_status(&status, (MSAPIndication *)NULL,
	    			midas_library_error,
			        MP_E_TOO_MANY_SESSIONS));
        }

        /*simply ignore the supplied session*/
        
	if((statusp = build_session(&default_session, new_session)) 
						!=  MIDAS_SUCCESS){
	    msp_Q_free(new_session);
 	    *bound_session_return = (MIDAS_SESSION *)NULL;
            return(statusp);
	}
        session_initialised = TRUE;
        debug_test("In mp_bind_M @ 3");
        return(MIDAS_SUCCESS);
    }
    else{
	/*We must be in GMD mode of operation*/

	debug_test("In mp_bind_M @ 4 - in GMD mode");
	if((statusp = build_session(session, new_session)) 
						!=  MIDAS_SUCCESS){
	    msp_Q_free(new_session);
 	    *bound_session_return = (MIDAS_SESSION *)NULL;
            return(statusp);
	}

        if(new_session->responder_address == NULLPA || 
			new_session->responder_title == NULLAEI){
	    msp_Q_free(new_session);
 	    *bound_session_return = (MIDAS_SESSION *)NULL;
	    return(bad_session());
        }
#if defined(SECURITY)

        /*
         * should check that new_session->authentication_mechanism
         * is set to request security, if it isn't then skip this
         * bit
         *
         */

        /* Is an authenticated association required? */
	if (workspace_secure &&
            new_session->authentication_mechanism != (OID) 0) {

            OID o = oid_cpy(str2oid(CREDENTIAL_OID));

	    debug_test("In mp_bind_M @ 5");

            if (!oid_equal(new_session->authentication_mechanism, o)) {
                /* unknown authentication_mechanism */

		fprintf(stderr, "mp_bind_M: bad auth mech\n");
                oid_free(o);
                msp_Q_free(new_session);
                *bound_session_return = (MIDAS_SESSION *)NULL;
                return(error_status(&status, (MSAPIndication *)NULL,
                                    midas_communication_error,
                                    AUTH_error));
            }

            oid_free(o);

	    /* authenticated association */

            authFlag = TRUE;
	
            aaii = (AuthAssocIntegrityInfo *)
                (new_session->authentication_value); 
	    
            if (aaii == (AuthAssocIntegrityInfo *) 0) {
                fprintf(stderr, "mp_bind_M: NULL authentication_value\n");
                msp_Q_free(new_session);
                *bound_session_return = (MIDAS_SESSION *)NULL;
                return(error_status(&status, (MSAPIndication *)NULL,
                                    midas_communication_error,
                                    AUTH_error));
            }

            if (new_session->user_info){
                /*Should be left blank as we want to use it!*/
                fprintf(stderr, "mp_bind_M: user_info in use\n");
                msp_Q_free(new_session);
                *bound_session_return = (MIDAS_SESSION *)NULL;
                return(error_status(&status, (MSAPIndication *)NULL,
                                    midas_communication_error,
                                    AUTH_error));
            }

            /* Make ACSE User Info */
            if ((e = makeAcseUserInfo(aaii, &new_session->user_info)) != 0) {
                fprintf(stderr, "mp_bind_M: makeAcseUserInfo() %s\n",
                        authErrorMessage(e));
                msp_Q_free(new_session);
                *bound_session_return = (MIDAS_SESSION *)NULL;
                return(error_status(&status, (MSAPIndication *)NULL,
                                    midas_communication_error,
                                    e));
            }
        }
#endif

        /*Establish a management association via MSAP.*/
        debug_test("In mp_bind_M @ 6");
        response = M_InitialiseReq (
		new_session->requestor_title,
		new_session->requestor_address,
		new_session->responder_title,
		new_session->responder_address,
			context, 
			protvrsn, 
		new_session->CMIS_functional_units,
		new_session->access_control,
		new_session->user_info,
		&mc, 
		&mi);

        debug_test("In mp_bind_M @ 5A");

#if defined(SECURITY)
 	/* In this case the user_info storage has been allocated by
	   XMP so free it*/

        if (new_session->user_info != (External *) 0) {
            external_free(new_session->user_info);
            new_session->user_info = (External *) 0;
        }
#endif

        /*Add the file descriptor to the session object*/
        if (response == OK){
	    debug_test("In mp_bind_M @ 6");
            session_initialised = TRUE;
            new_session->file_descriptor = mc.mc_sd;
	    new_session->user_info = mc.mc_info.mai_info;
        }
        else{
            debug_test("In mp_bind_M @ 7");
	    msp_Q_free(new_session);
	    *bound_session_return = (MIDAS_SESSION *)NULL;
	    return(error_status(&status, &mi,
		    midas_communication_error,
		    MP_E_COMMUNICATIONS_PROBLEM));
	}
        debug_test("In mp_bind_M @ END");

#if defined(SECURITY)

	if (authFlag) {
	    /*
	    ** Authenticate peer
	    */
	    e = getAcseUserInfo(mc.mc_info.mai_info, &aaii2);

    	    /* Sanity check */
	    if (e == 0 && memcmp(aaii->sessionKey, aaii2->sessionKey, 8) != 0) {
	        fprintf(stderr, "mp_bind_M: different sessionKey from peer\n");
	        e = AUTH_error;
	    }
	}

	AuthAssocIntegrityInfo_free(aaii2);

        if (authFlag && e == 0) {
	    /*SUCCESS!*/

	    /* Set up the session key */
	    setMd5Key(mc.mc_sd, aaii->sessionKey);

	    /* Clear previous Id state */
	    clearId(mc.mc_sd);

            /* Clear the invokeId manager */
            clearInvoke(mc.mc_sd);

            /* Set the IRStatus */
            setIRStatus(mc.mc_sd, IRStatus_initiator);
	}

	else if (authFlag) {
	    /*FAILURE!*/
	    fprintf(stderr, 
		"mp_bind_M: Could not athenticate %s: %s - aborting assoc\n",
		sprintaei(new_session->responder_title),
		authErrorMessage(e));
	    /*
	    ** If authentication fails, for now, drop the assoc. Should
	    ** actually give the user the choice, but that requires
	    ** changing the interface to mp_bind()
	    */
	    M_AbortReq(mc.mc_sd, (External *) 0, &mi); 

	    /*
	     * Give up the bind and return an error
	     *
	     */
	     msp_Q_free(new_session);
            *bound_session_return = (MIDAS_SESSION *)NULL;
             mc_free(&mc);
	     return(error_status(&status, (MSAPIndication *)NULL,
				    midas_communication_error,
				    e));
	}
#endif /* SECURITY */

	return(MIDAS_SUCCESS);
    } /*GMD mode*/
}

/*
 * mp_unbind_M - unbind from a management session
 *
 * N.B. In the MIDAS version only one session is allowed per process
 *
 */

MIDAS_STATUS *
mp_unbind_M(session)
    MIDAS_SESSION *session;{
    static MIDAS_STATUS status;
    static MSAPIndication mi;

    debug_test("In mp_unbind_M @ 1");	/*Debugging aid*/

    if (!session_initialised){
        return(bad_session());
    }

    /*Are we in GMD mode of operation ?*/
    if ((session->responder_title)||(session->responder_address)){
        debug_test("In mp_unbind_M @ 2 - GMD mode");	/*Debugging aid*/

        /*Close the MSAP association for this session*/
        if (M_TerminateReq (session->file_descriptor, ACF_NORMAL, NULLPE, &mi) 
		== NOTOK){
            status.MIDAS_problem.midas_problem_type = midas_service_error;
            status.MIDAS_problem.midas_problem_value.error_val 
						= MP_E_PROCESSING_FAILURE;
            status.MIDAS_parameter = &mi;
            debug_test("In mp_unbind_M @ 3");	/*Debugging aid*/
            return(&status);
        }

#if defined(SECURITY)
        /*
        ** In GMD mode the assoc is closed so clear any assoc related
        ** security info.
        */
        clearMd5Key(session->file_descriptor);
        clearId(session->file_descriptor);
        clearInvoke(session->file_descriptor);
        clearIRStatus(session->file_descriptor);
#endif
    
    }
    
    /*In UCL mode of operation it would be nice to call m_terminate for all 
      the open fd's at this point but we don't know which they are. This is 
      a drawback of not putting the file descriptor in the session object*/

    /*The problem above is easily fixed*/

    debug_test("In mp_unbind_M @ return SUCCESS");
    return(MIDAS_SUCCESS);
}


/*
 * mp_initialise_M - initialise the interface
 *
 * N.B. In the MIDAS version only one workspace is allowed. The call
 * simply checks whether it is being called twice.
 *
 */


MIDAS_WORKSPACE *
mp_initialise_M(){

    if (workspace_initialised) return((MIDAS_WORKSPACE *)NULL);

    workspace_initialised = TRUE;
    workspace_secure = FALSE;

    return((MIDAS_WORKSPACE *)1);
}


/*
 * mp_shutdown_M - delete a workspace and the associated resources
 *
 */
 
 
MIDAS_STATUS *
mp_shutdown_M(workspace)
    MIDAS_WORKSPACE workspace; {
 
    if (!workspace_initialised) return(MIDAS_NO_WORKSPACE);

    workspace_initialised = FALSE;
    workspace_secure = FALSE;

#if defined(SECURITY)
    /* Clear any assoc related security info */
{
    register
    int i;

    for(i = 0; i < FD_SETSIZE; ++i) {
        clearMd5Key(i); /* MD5 key store */
        clearId(i);     /* PDU id generator */
        clearInvoke(i); /* id/invokeId manager */
    }
}
#endif
 
    return(MIDAS_SUCCESS);
}
 
 

/*
 * mp_negotiate_M - negotiate features of the interface and service
 *
 * N.B. Just return success in the first implementation
 *
 */

MIDAS_STATUS *
mp_negotiate_M(feature_list, workspace)
    MIDAS_FEATURE *feature_list;
    MIDAS_WORKSPACE workspace;{
    register
    MIDAS_FEATURE *mfp;
#if defined(SECURITY)
    OID o = oid_cpy(str2oid(CREDENTIAL_OID));
#endif

    if (!workspace_initialised) return(MIDAS_NO_WORKSPACE);

    workspace_secure = FALSE;

    /* Check the feature list */
    for(mfp = feature_list; mfp != (MIDAS_FEATURE *) 0;
        mfp = mfp->MIDAS_feature_next) {
#if defined(SECURITY)
	debug_test("mp_negotiate @ 1\n");
        /* Only one feature known to us for the security stuff */
        if (oid_equal(&mfp->feature, o) && 
	    (mfp->request == MP_ACTIVATE)) {
	    debug_test("mp_negotiate @ 2\n");
            mfp->response = TRUE;
            workspace_secure = TRUE;

            /* Once only initialisations for security extensions */
            initialiseMd5KeySpace(); /* MD5 key store */
            initialiseIdSpace();     /* PDU id generator */
            initialiseInvoke();      /* id/invokeId manager */
            initialiseIRStatus();    /* initiator/responder status */
        }
        else
#endif
            /* No other features recognised */
            mfp->response = FALSE;
    }

#if defined(SECURITY)
    oid_free(o);
#endif

    return(MIDAS_SUCCESS);
}

/*
 * build_session - construct a bound-session object 
 *
 * N.B. This copies pointers to sub-structures it does not allocate
 * new storage
 *
 */

static MIDAS_STATUS*
build_session(session, new_session)
    MIDAS_SESSION *session;
    MIDAS_SESSION *new_session;
{
    MSAPIndication *mi;
    MIDAS_OID_LIST *pctx_list;
    OID ctx;

    /*Copy all the information to start with then overwrite as mecessary*/

    *new_session = *session; /*struct copy*/


    /*Ignore any supplied role*/
	
    new_session->role = MP_T_MANAGING | MP_T_MONITORING | MP_T_PERFORMING |
                            MP_T_REPORTING;

    /*set the access control attribute to NULL as we do not use it*/

    new_session->access_control = (External *)NULL;

    /*Ignore any supplied application context*/

    if ((new_session->application_context = ode2oid(context)) == NULLOID) {
        return (MIDAS_INSUFFICIENT_RESOURCES);
    }


    /*Ignore any supplied presentation contexts*/
    if ((pctx_list 
	 = (MIDAS_OID_LIST*)malloc(sizeof(MIDAS_OID_LIST))) == NULLOIDLIST){
	 return (MIDAS_INSUFFICIENT_RESOURCES);
    }

    pctx_list->MIDAS_OID_next = NULLOIDLIST;
    pctx_list->MIDAS_OID_this = *oid_cpy(CMIP_ASN_OID); /*struct copy*/
    new_session->presentation_context_list = pctx_list;

    /*Ignore any supplied SMASE_user_data*/

    new_session->SMASE_user_data = NULL;

    new_session->MIDAS_session_next == (MIDAS_SESSION *)NULL;

    return (MIDAS_SUCCESS);
}

/*
 *
 * free_session - free a session object and associated storage
 * 
 *
 */

static void
free_session(session)
    MIDAS_SESSION *session;{
    MIDAS_OID_LIST *pctx_list;

    if (session->application_context) 
        free((char *)session->application_context);

    free_oid_list(session->presentation_context_list);
}

/*
 *
 * free_oid_list - free an OID list
 *
 */

static void
free_oid_list(oidp)
    MIDAS_OID_LIST *oidp;{
    MIDAS_OID_LIST *next;

    if (oidp == (MIDAS_OID_LIST *)NULL) return;

    free_oid_list(oidp->MIDAS_OID_next);

    free(oidp);
}

int
get_association_descriptor(session, context)
MIDAS_SESSION *session;
MIDAS_CONTEXT *context;
{
    debug_test("In get_association_descriptor @ 1");
    /*Association descriptors should be positive integers*/
    if(session->file_descriptor < MP_NO_VALID_FILE_DESCRIPTOR)
        return (NOTOK);

    if(context->connection_id < MP_NO_VALID_FILE_DESCRIPTOR)
        return (NOTOK);

    if(session->file_descriptor == MP_NO_VALID_FILE_DESCRIPTOR)
        /*We must be in UCL mode, so use descriptor in context*/
        return (context->connection_id);
    else
    	/*We must be in GMD mode, so use descriptor in session*/
        return (session->file_descriptor);
}

MIDAS_STATUS *
bad_session()
{
    static MIDAS_STATUS status;

    status.MIDAS_problem.midas_problem_type = midas_library_error;
    status.MIDAS_problem.midas_problem_value.error_val = MP_E_BAD_SESSION;
    status.MIDAS_parameter = (MSAPIndication *)NULL;

    return (&status);
}


/*
 * msp_Q_free - free a list of MIDAS_SESSION structures
 *
 */

void
msp_Q_free(msp)
    MIDAS_SESSION* msp;
{
    register MIDAS_SESSION* next;

    if (msp == (MIDAS_SESSION *)NULL)
        return;

    do{
	next = msp->MIDAS_session_next;
        if (msp->application_context) 
            free(msp->application_context);
        oid_Q_free(msp->presentation_context_list);
        free(msp);
    }while (msp = next);
}

/*
 * oid_Q_free - free a  MIDAS_OID_LIST structure
 *
 */

void
oid_Q_free(list)
    MIDAS_OID_LIST *list;
{
    register MIDAS_OID_LIST *next;

    if (list == NULLOIDLIST)
         return;

    do{
        next = list->MIDAS_OID_next;
        free(list);
    }while (list = next);
}
