/*
 * Copyright (c) 1993 University College London
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Department of Computer Science, University College London.
 * The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/* Jonathan Ladipo, November 1993 */

#include "xmp.h"
#include "mp_utilities.h"

#if !defined(_a_)
#if defined(__STDC__) || defined(__cplusplus)
#define _a_(_args_) _args_
#else
#define _a_(_args_) ()
#endif
#endif

#if defined(SECURITY)
#include "mintegrity.h"
#include "minvoke.h"
#include "mp_integrity.h"
#endif

extern int mp_wait_pdu _a_((int, MSAPIndication *, MIDAS_STATUS *, int));
extern int debug_test _a_((char *));
extern int get_association_descriptor _a_((MIDAS_SESSION *, MIDAS_CONTEXT *));

/*
 * mp_set_req_M - retrieve management information
 *
 */

MIDAS_STATUS_P mp_set_req_M(session, context, argument, result_return, invoke_id_return)
     	MIDAS_SESSION *		session;
     	MIDAS_CONTEXT *		context;
     	CMISSetArg *		argument;		
     	MIDAS_RESULT **		result_return;	/*Set to NULLMIDASRESULT 
								for async mode*/
     	OM_sint32 *              invoke_id_return;	
{

	register MIDAS_RESULT *mrp_first, *mrp_last, *mrp_new;
	int msapres, secs, assoc_descriptor;
	static MIDAS_STATUS status; /*Will be returned if an error occurs*/
        static MSAPIndication indication;
#if defined(SECURITY)
        char *state;
#endif

	debug_test("In mp_set_req_M @ 1A");

        status.MIDAS_parameter = (MSAPIndication *)NULL;

	if (result_return) 
		*result_return = NULLMIDASRESULT;			

    	/*Do some sanity checks on the parameters*/

    	if (check_session(session, &status) != OK) {
		fprintf(stderr, "mp_set_req_M: bad session\n");
		return(&status);
    	}

    	if (check_context(context, &status) != OK) {
		fprintf(stderr, "mp_set_req_M: bad context\n");
		return(&status);
    	}

        /*Get the association descriptor*/
        if ((assoc_descriptor = get_association_descriptor(session, context))
        	== NOTOK){
                fprintf(stderr, 
			"mp_set_req_M: invalid association descriptor\n");
                return (error_status(&status, (MSAPIndication *)NULL,
			midas_library_error,
                        MP_E_INVALID_CONNECTION_ID));
        }

#if defined(SECURITY)
        /*
        ** Make theid value for the PDU. Save the sequence state also
        ** because we will need to rewind the sequence if the call fails.
        */
        if (getIRStatus(assoc_descriptor) == IRStatus_none) {
            /* should not happen */
            fprintf(stderr, "mp_set_req_M: IRStatus_none for fd=%d",
                    assoc_descriptor);
            return(error_status(&status, (MSAPIndication *) 0,
                                midas_library_error, 0));
        }

        saveState(assoc_descriptor, &state);
        *invoke_id_return = makeId(assoc_descriptor);
#endif

    	/*Now do the set*/

    	msapres = M_SetAux(
		assoc_descriptor,
		(context->mode == MP_T_CONFIRMED) ? M_SETC: M_SET,
		*invoke_id_return,
		&argument->sa_class,
		&argument->sa_inst,
		&argument->sa_scope,
		argument->sa_filter,
		argument->sa_access,
		argument->sa_sync,
		argument->sa_nattrs,
		argument->sa_attrs,
		&indication
            	);

#if defined(SECURITY)
	if (msapres == OK) {
	    /*
	    ** Add the {id, invokeId} pair to the list and return
	    ** invokeId to user.
	    */
	    addInvoke(assoc_descriptor, *invoke_id_return,
                              indication.mi_invoke);
	}
	else {
	    restoreState(assoc_descriptor, state);
	}
#endif

	*invoke_id_return = indication.mi_invoke;

	/*If this is an async call then tidy up and return*/

 	if (context->asynchronous){
	    if (msapres == OK) {
		debug_test("Leaving mp_set_req_M @ 1C");
#if defined(SECURITY)
		free(state);
#endif
		return(MIDAS_SUCCESS);
	    }

	    /*Async call has failed*/
	    debug_test("Leaving mp_set_req_M @ 2");
	    return(error_status(&status, &indication, midas_service_error, 
		    MP_E_COMMUNICATIONS_PROBLEM));  /*Not true!*/
    	}

	/*must be a sync call - wait for the reply*/

	/*calculate the timeout value - N.B. should really reduce this
	  each time a reply is received*/

	secs = (context->time_limit > 0)? context->time_limit: -1;

   	/*set up the queue pointers*/

	mrp_first = mrp_last = NULLMIDASRESULT;

   	for(;;) {
		/*Allocate storage for the result*/

        	if ((mrp_new = (MIDAS_RESULT *)malloc(sizeof(MIDAS_RESULT)))
                	== NULLMIDASRESULT){
            		mrp_Q_free(mrp_first);
            		return(MIDAS_INSUFFICIENT_RESOURCES);
        	}

        /*Add result structure to Q*/  

        	if (mrp_first == NULLMIDASRESULT)mrp_last = mrp_first = mrp_new;
          	mrp_last->MIDAS_result_next = mrp_new; 
		mrp_last = mrp_new;
        	mrp_new->MIDAS_result_next = NULLMIDASRESULT;

		switch (mp_wait_pdu(assoc_descriptor, 
				&mrp_new->MIDAS_result_info,
				&status, secs))

                {
            	case DONE:
                        *result_return = mrp_first;
                        debug_test("Leaving mp_set_req_M @ 3");
                        return(MIDAS_SUCCESS);

            	case OK:
                	/*OK so far - more linked replies may be forthcoming*/
                	/*N.B. SHOULD REALLY ADJUST secs!*/
                	break;
            	case NOTOK:

/*
  NOTE
  Maybe case NOTOK should just do the same as case DONE. This will allow all
  outstanding linked replies to be returned. At present all outstandding replies
  received before a NOTOK are freed.
 */

                	/*Error prevents further processing*/
                	/*N.B. I THINK THAT IF SOME LINKED REPLIES HAVE
                	SUCCEEDED THESE SHOULD BE RETURNED IN THE PARAMETER
                	PART OF THE STATUS MESSAGE. I CAN'T FIND THE REFERENCE
                	FOR THIS AT PRESENT SO I HAVE IGNORED THE PROBLEM*/
                	mrp_Q_free(mrp_first);
                	return(&status);
            	default:
                	fprintf(stderr, "bad return from mp_mait_pdu\n");
                	return(MIDAS_INSUFFICIENT_RESOURCES);
        }

    }
}


